Package: PEAXAI
Title: Probabilistic Efficiency Analysis Using Explainable Artificial
        Intelligence
Version: 1.0.3
Authors@R: 
  c(
    person(given = "Ricardo",
           family = "González Moyano",
           role = c("cre", "aut"),
           email = "ricardo.gonzalezm@umh.es",
           comment = c(ORCID = "0009-0002-8608-5545")),
    person(given = "Juan",
           family = "Aparicio",
           role = "aut",
           comment = c(ORCID = "0000-0002-0867-0004")),
    person(given = "José Luis",
           family = "Zofío",
           role = "aut",
           comment = c(ORCID = "0000-0003-1170-9501")),
    person(given = "Víctor",
           family = "España",
           role = "aut",
           comment = c(ORCID = "0000-0002-1807-6180"))
  )
Description: Provides a probabilistic framework that integrates Data Envelopment
  Analysis (DEA) (Banker et al., 1984) <doi:10.1287/mnsc.30.9.1078> with machine
  learning classifiers (Kuhn, 2008) <doi:10.18637/jss.v028.i05> to estimate both the
  (in)efficiency status and the probability of efficiency for decision-making
  units. The approach trains predictive models on DEA-derived efficiency labels
  (Charnes et al., 1985) <doi:10.1016/0304-4076(85)90133-2>, enabling explainable
  artificial intelligence (XAI) workflows with global and local interpretability
  tools, including permutation importance (Molnar et al., 2018) <doi:10.21105/joss.00786>,
  Shapley value explanations (Strumbelj & Kononenko, 2014) <doi:10.1007/s10115-013-0679-x>,
  and sensitivity analysis (Cortez, 2011) <https://CRAN.R-project.org/package=rminer>.
  The framework also supports probability-threshold peer selection and counterfactual
  improvement recommendations for benchmarking and policy evaluation. The probabilistic
  efficiency framework is detailed in González-Moyano et al. (2025)
  "Probability-based Technical Efficiency Analysis through Machine Learning",
  in review for publication.
License: GPL-3
URL: https://github.com/rgonzalezmoyano/PEAXAI
BugReports: https://github.com/rgonzalezmoyano/PEAXAI/issues
Encoding: UTF-8
Language: en
RoxygenNote: 7.3.2
Depends: R (>= 4.4.0)
Imports: Benchmarking, caret, deaR, dplyr, kernelshap, iml, lime, np,
        PRROC, pROC, rminer, rms, stats, peakRAM
Suggests: ggplot2, knitr, rmarkdown
VignetteBuilder: knitr
LazyData: false
ByteCompile: true
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-23 10:08:15 UTC; Ricardo
Author: Ricardo González Moyano [cre, aut] (ORCID:
    <https://orcid.org/0009-0002-8608-5545>),
  Juan Aparicio [aut] (ORCID: <https://orcid.org/0000-0002-0867-0004>),
  José Luis Zofío [aut] (ORCID: <https://orcid.org/0000-0003-1170-9501>),
  Víctor España [aut] (ORCID: <https://orcid.org/0000-0002-1807-6180>)
Maintainer: Ricardo González Moyano <ricardo.gonzalezm@umh.es>
Repository: CRAN
Date/Publication: 2026-08-23 10:41:14 UTC
Built: R 4.5.2; ; 2026-08-23 13:38:46 UTC; unix

# Generated by roxygen2: do not edit by hand

export(PEAXAI_counterfactuals)
export(PEAXAI_fitting)
export(PEAXAI_global_importance)
export(PEAXAI_local_importance)
export(PEAXAI_peer)
export(PEAXAI_predict)
export(PEAXAI_ranking)
export(reffcy)
importFrom(Benchmarking,dea.add)
importFrom(PRROC,pr.curve)
importFrom(caret,confusionMatrix)
importFrom(caret,createDataPartition)
importFrom(caret,createFolds)
importFrom(caret,train)
importFrom(caret,trainControl)
importFrom(deaR,make_deadata)
importFrom(deaR,maximal_friends)
importFrom(deaR,model_additive)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,select)
importFrom(dplyr,semi_join)
importFrom(dplyr,summarise)
importFrom(iml,FeatureImp)
importFrom(iml,Predictor)
importFrom(kernelshap,kernelshap)
importFrom(lime,lime)
importFrom(np,npudens)
importFrom(np,npudensbw)
importFrom(pROC,auc)
importFrom(pROC,roc)
importFrom(peakRAM,peakRAM)
importFrom(rminer,Importance)
importFrom(rms,val.prob)
importFrom(stats,approxfun)
importFrom(stats,binomial)
importFrom(stats,family)
importFrom(stats,glm)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,combn)
importFrom(utils,head)
importFrom(utils,tail)

# PEAXAI

`PEAXAI` (**Probabilistic Efficiency Analysis Using Explainable Artificial Intelligence**) is an R package for estimating technical efficiency through a probabilistic framework that integrates Data Envelopment Analysis (DEA), machine learning classifiers, and explainable artificial intelligence (XAI) tools.

The package uses DEA-derived efficiency labels to train supervised machine learning models that estimate the probability that each decision-making unit (DMU) is efficient. This probability-based approach allows users to move beyond a purely deterministic efficient/inefficient classification and to analyse efficiency under uncertainty.

In addition, `PEAXAI` provides post-hoc interpretability tools to explain the estimated efficiency probabilities. These tools include global and local feature importance, Shapley value explanations, permutation importance, sensitivity analysis, counterfactual improvement recommendations, efficiency rankings, and peer identification based on user-defined probability thresholds.

The package is designed for researchers and practitioners interested in productivity and efficiency analysis, benchmarking, policy evaluation, operational research, and explainable machine learning applications.

Please check the vignette for more details.

## Metadata Table

| Metadata item | Description |
|---|---|
| Software name | `PEAXAI` |
| Full name | Probabilistic Efficiency Analysis Using Explainable Artificial Intelligence |
| Current version | 1.0.2 |
| Programming language | R |
| Minimum R version | R >= 3.5 |
| Repository | GitHub |
| Repository URL | https://github.com/rgonzalezmoyano/PEAXAI |
| Issue tracker | https://github.com/rgonzalezmoyano/PEAXAI/issues |
| License | GPL-3 |
| Main purpose | Probability-based technical efficiency analysis using DEA, machine learning and explainable artificial intelligence |
| Main methodology | DEA-based efficiency labelling, supervised machine learning classification, probability-based efficiency assessment, XAI, counterfactual analysis and peer identification |
| Main outputs | Efficiency probabilities, efficiency rankings, global and local explanations, counterfactual targets and benchmark peers |
| Example data | Spanish food industry firms from the SABI database |
| Documentation | Function documentation and package vignette |
| Associated manuscript | *Probability-based Technical Efficiency Analysis through Machine Learning* |
| Authors | Ricardo González Moyano, Juan Aparicio, José Luis Zofío and Víctor España |
| Maintainer | Ricardo González Moyano |
| Contact | ricardo.gonzalezm@umh.es |

## Installation

You can install the released version of the package from CRAN with:

```r
install.packages("PEAXAI")
```

And the development version from GitHub with:

```r
install.packages("devtools")
devtools::install_github("rgonzalezmoyano/PEAXAI")
```

## Dependencies

`PEAXAI` depends on R and imports the following packages:

```text
Benchmarking
caret
deaR
dplyr
kernelshap
iml
isotone
lime
np
PRROC
pROC
rminer
rms
stats
```

The following packages are suggested for examples, documentation and vignettes:

```text
ggplot2
knitr
rmarkdown
```

## Main functions

The main functions exported by `PEAXAI` are:

| Function | Description |
|---|---|
| `PEAXAI_fitting()` | Fits machine learning classifiers using DEA-derived efficiency labels and selects the best model according to user-defined performance metrics. |
| `PEAXAI_predict()` | Predicts the probability that each DMU is efficient. |
| `PEAXAI_global_importance()` | Computes global feature importance using XAI methods such as sensitivity analysis, SHAP or permutation importance. |
| `PEAXAI_local_importance()` | Computes local explanations for individual DMUs. |
| `PEAXAI_counterfactuals()` | Computes counterfactual input-output targets required for a DMU to reach a given efficiency probability threshold. |
| `PEAXAI_ranking()` | Produces efficiency rankings based on predicted probabilities or attainable counterfactual improvements. |
| `PEAXAI_peer()` | Identifies benchmark peers according to estimated efficiency probabilities and user-defined thresholds. |
| `label_efficiency()` | Generates DEA-based efficiency labels for the training process. |
| `SMOTE_data()` | Generates synthetic observations to address class imbalance. |
| `SMOTE_Z_data()` | Generates synthetic observations based on frontier-related combinations. |

## Basic example

The following example illustrates the main workflow of the package using the example dataset included in `PEAXAI`.

```r
library(PEAXAI)

data("firms", package = "PEAXAI")

# Select firms from the Autonomous Community of Valencia
data <- subset(firms, autonomous_community == "Comunidad Valenciana")
data <- data[, -ncol(data)]

# Define inputs and output
x <- c(1:4)
y <- c(5)

# Define returns to scale
RTS <- "vrs"

# Define machine learning method
methods <- list(
  "nnet" = list(
    tuneGrid = expand.grid(
      size = c(1, 5, 10, 20),
      decay = 10^seq(-5, -1, by = 1)
    ),
    preProcess = c("center", "scale"),
    entropy = TRUE,
    skip = TRUE,
    maxit = 1000,
    MaxNWts = 100000,
    trace = FALSE,
    weights = NULL
  )
)

# Cross-validation settings
trControl <- list(
  method = "cv",
  number = 5
)

# Metrics used to select the best model
metric_priority <- c("Balanced_Accuracy", "F1", "ROC_AUC")

# Define imbalance treatment
imbalance_rate <- seq(0.05, 0.4, 0.05)

# Set seed for reproducibility
seed <- 1

# Fit PEAXAI model
models <- PEAXAI_fitting(
  data = data,
  x = x,
  y = y,
  RTS = RTS,
  imbalance_rate = imbalance_rate,
  methods = methods,
  trControl = trControl,
  metric_priority = metric_priority,
  hold_out = NULL,
  verbose = TRUE,
  seed = seed
)

# Extract final model
final_model <- models[["best_model_fit"]][["nnet"]]

# Estimate efficiency probabilities
probabilities <- PEAXAI_predict(
  data = data,
  x = x,
  y = y,
  final_model = final_model
)

head(probabilities)
```

## Explainable AI analysis

`PEAXAI` allows users to compute global and local explanations for the estimated efficiency probabilities.

For example, global feature importance can be obtained as follows:

```r
importance_method <- list(
  name = "SHAP",
  nsim = 200
)

relative_importance <- PEAXAI_global_importance(
  final_model = final_model,
  x = x,
  y = y,
  explain_data = data,
  reference_data = data,
  importance_method = importance_method
)

relative_importance
```

## Counterfactual efficiency targets

Given a set of probability thresholds, `PEAXAI` can estimate the input-output changes required for a DMU to reach a desired probability of being efficient.

```r
efficiency_thresholds <- seq(0.75, 0.95, 0.1)

directional_vector <- list(
  relative_importance = relative_importance,
  baseline = "mean"
)

targets <- PEAXAI_counterfactuals(
  data = data,
  x = x,
  y = y,
  final_model = final_model,
  efficiency_thresholds = efficiency_thresholds,
  directional_vector = directional_vector,
  n_expand = 0.25,
  n_grid = 50,
  max_y = 1,
  min_x = 1
)

head(targets[["0.85"]][["counterfactual_dataset"]])
head(targets[["0.85"]][["inefficiencies"]])
```

## Efficiency rankings

Efficiency rankings can be computed either from the predicted probabilities or from attainable counterfactual improvements.

```r
ranking_predicted <- PEAXAI_ranking(
  data = data,
  x = x,
  y = y,
  final_model = final_model,
  rank_basis = "predicted"
)

head(ranking_predicted)
```

If counterfactual targets have already been computed, an attainable-efficiency ranking can be obtained as follows:

```r
ranking_attainable <- PEAXAI_ranking(
  data = data,
  x = x,
  y = y,
  final_model = final_model,
  efficiency_thresholds = efficiency_thresholds,
  targets = targets,
  rank_basis = "attainable"
)

head(ranking_attainable[["0.85"]])
```

## Peer identification

`PEAXAI` also identifies benchmark peers for each DMU according to estimated efficiency probabilities and selected probability thresholds.

```r
peers <- PEAXAI_peer(
  data = data,
  x = x,
  y = y,
  final_model = final_model,
  targets = targets,
  efficiency_thresholds = efficiency_thresholds,
  weighted = FALSE,
  relative_importance = relative_importance
)

head(peers)
```

## Repository structure

The repository is organised as follows:

```text
PEAXAI/
├── R/
│   └── R source files containing the main functions of the package
├── data/
│   └── Example datasets included in the package
├── man/
│   └── Function documentation generated with roxygen2
├── vignettes/
│   └── Package vignette with a complete worked example
├── DESCRIPTION
│   └── R package metadata, dependencies, authors and license information
├── NAMESPACE
│   └── Exported functions and imported methods
├── README.md
│   └── General description, installation instructions and usage examples
└── LICENCE.txt
    └── Full license text
```

## Data

The package includes example data that can be loaded directly in R.

The vignette illustrates the use of `PEAXAI` with a dataset of Spanish food industry firms obtained from the SABI database for the year 2023. The dataset includes firms with more than 50 employees and contains financial and operational variables relevant for productivity and efficiency analysis.

The output variable used in the vignette is:

```text
operating_income
```

The input variables used in the vignette are:

```text
total_assets
employees
fixed_assets
personnel_expenses
```

The variable `autonomous_community` indicates the geographical location of each firm within one of Spain's autonomous communities.

## Documentation

After installing the package, the documentation of each function can be accessed in R using:

```r
?PEAXAI_fitting
?PEAXAI_predict
?PEAXAI_global_importance
?PEAXAI_local_importance
?PEAXAI_counterfactuals
?PEAXAI_ranking
?PEAXAI_peer
```

The package vignette can be opened with:

```r
browseVignettes("PEAXAI")
```

## License

This software is distributed under the GPL-3 license.

Please see the `LICENCE.txt` file for the full license text.

## Citation

If you use `PEAXAI`, please cite the associated manuscript:

```bibtex
@article{gonzalez_moyano_peaxai,
  title   = {Probability-based Technical Efficiency Analysis through Machine Learning},
  author  = {González Moyano, Ricardo and Aparicio, Juan and Zofío, José Luis and España, Víctor},
  journal = {In review},
  year    = {2025}
}
```

## Contact

For questions, bug reports or suggestions, please use the GitHub Issues page:

```text
https://github.com/rgonzalezmoyano/PEAXAI/issues
```

or contact the maintainer:

```text
Ricardo González Moyano
ricardo.gonzalezm@umh.es
```

#' @title Training a Classification Machine Learning Model
#'
#' @description This function trains a set of models and selects best hyperparameters for each of them.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables in the model.
#' @param method Parameters for controlling the training process (from the \code{'caret'} package).
#' @param parameters A \code{list} of selected machine learning models and their hyperparameters.
#' @param trControl A \code{list} of selected machine learning learning.
#' @param metric_priority A \code{string} specifying the summary metric for classification to select the optimal model. Default includes \code{"Balanced_Accuracy"} due to (normally) unbalanced data.
#' @param seed  Integer. Seed for reproducibility.

#'
#' @importFrom caret train trainControl
#' @importFrom pROC roc
#' @importFrom PRROC pr.curve
#'
#' @return It returns a \code{list} with the chosen model.

train_PEAXAI <- function (
    data, method, parameters, trControl, metric_priority, seed
    ) {

  # general function to get performance
  PEAXAIsummaryFunction <- function (data, lev = NULL, model = NULL) {

    if(any(is.na(data$pred))) {

      out_names <- c("Accuracy", "Kappa",
                     "Recall", "Specificity", "Precision", "F1", "Balanced_Accuracy", "G_mean",
                     "ROC-AUC", "PR-AUC",
                     "Cross_Entropy", "Cross_Entropy_Efficient")

      performance <- setNames(rep(NA, length(out_names)), out_names)

      return(performance)

      }

    # levels
    levls <- c("efficient", "not_efficient")

    # predictions
    y_hat <- data$pred
    y_hat <- factor(y_hat, levels = levls)
    if(levels(y_hat) != c("efficient", "not_efficient")) browser()
    y_hat_prob <- data$efficient

    # reference
    y_obs <- data$obs
    y_obs <- factor(y_obs, levels = levls)

    # confusion matrix
    cm <- confusionMatrix(
      data = y_hat,
      reference = y_obs,
      mode = "everything",
      positive = "efficient"
    )

    # caret problem when all units are in the same class
    # not NAs
    if (is.na(cm$byClass[["F1"]])) {
      cm$byClass[["F1"]] <- 0
    }
    if (is.na(cm$byClass[["Precision"]])) {
      cm$byClass[["Precision"]] <- 0
    }

    # G-mean
    gmean <- sqrt(cm$byClass["Recall"]*cm$byClass["Specificity"])

    # ROC-AUC
    roc_obj <- roc(
      response = y_obs,
      predictor = y_hat_prob,
      levels = rev(levls),
      direction = "<",
      quiet = TRUE)

    # PR-AUC
    pr_obj <- pr.curve(
      scores.class0 = y_hat_prob[y_obs == "not_efficient"],
      scores.class1 = y_hat_prob[y_obs == "efficient"],
      curve = TRUE
    )

    # cross_entropy
    y_bin <- ifelse(y_obs == "efficient", 1, 0)
    p <- y_hat_prob

    # no log(0)
    eps <- 1e-15
    p_clipped <- pmin(pmax(p, eps), 1 - eps)

    cross_entropy <- -mean(
      y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
    )

    # cross entropy efficient class
    p <- y_hat_prob[y_bin == 1]
    y_bin <- y_bin[y_bin == 1]

    # no log(0)
    eps <- 1e-15
    p_clipped <- pmin(pmax(p, eps), 1 - eps)

    cross_entropy_efficient <- -mean(
      y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
    )

    performance <- c(
      cm$overall[c("Accuracy", "Kappa")],
      cm$byClass[c("Recall", "Specificity",
                   "Precision", "F1")],
      "Balanced_Accuracy" = cm$byClass[["Balanced Accuracy"]],
      "G_mean" = unname(gmean),
      "ROC_AUC" = roc_obj$auc,
      "PR_AUC" = unname(pr_obj$auc.integral),
      "Cross_Entropy" = cross_entropy,
      "Cross_Entropy_Efficient" = cross_entropy_efficient
    )

    return(performance)

  }

  # trControl actualize
  # if (trControl[["method"]] == "cv") {
  #   trControl <- trainControl(
  #     method = trControl[["method"]],
  #     number = trControl[["number"]],
  #     summaryFunction = PEAXAIsummaryFunction,
  #     classProbs = TRUE,
  #     savePredictions = "all"
  #   )
  # } else if (trControl[["method"]] == "repeatedcv") {
  #   trControl <- trainControl(
  #     method = trControl[["method"]],
  #     number = trControl[["number"]],
  #     repeats = trControl[["repeats"]],
  #     summaryFunction = PEAXAIsummaryFunction,
  #     classProbs = TRUE,
  #     savePredictions = "all"
  #   )
  # }
  trControl <- trainControl(method = "none", classProbs = TRUE)

  # ----------------------------------------------------------------------------
  # Neural Network -------------------------------------------------------------
  # ----------------------------------------------------------------------------

  set.seed(seed)
  if (method == "nnet") {

    model_fit <- train(
      class_efficiency ~ .,
      data = data,
      method = "nnet",
      preProcess = parameters[["preProcess"]],
      tuneGrid = parameters[["tuneGrid"]],
      trControl = trControl,
      metric = "Accuracy",

      # nnet (no fine-tuning)
      skip = parameters[["skip"]],
      maxit = parameters[["maxit"]],
      MaxNWts = parameters[["MaxNWts"]],
      trace = parameters[["trace"]]
    )

  } else if (method == "rf") {

    mtry <- as.data.frame(parameters[["tuneGrid"]])
    names(mtry) <- "mtry"

    model_fit <- train(
      class_efficiency ~ .,
      data = data,
      method = "rf",
      tuneGrid = mtry,
      trControl = trControl,
      metric = "Accuracy",

      # rf (no fine-tuning)
      ntree = parameters[["ntree"]]
    )

  } else if (method == "svmPoly") {

    model_fit <- train(
      class_efficiency ~ .,
      data = data,
      method = "svmPoly",
      preProcess = parameters[["preProcess"]],
      tuneGrid = parameters[["tuneGrid"]],
      trControl = trControl,
      metric = "Accuracy",
      prob.model = TRUE
    )

  } else if (method == "svmRadial") {

    model_fit <- train(
      class_efficiency ~ .,
      data = data,
      method = "svmRadial",
      preProcess = parameters[["preProcess"]],
      tuneGrid = parameters[["tuneGrid"]],
      trControl = trControl,
      metric = "Accuracy",
      prob.model = TRUE
    )

  } else if (method == "glm") {

    if (is.null(parameters[["weights"]])) {

      w0 <- 1
      w1 <- 1

    } else {

      if (parameters[["weights"]][1] == "dinamic") {
        w0 <- nrow(data) / (2 * length(which(data$class_efficiency == "not_efficient")))
        w1 <- nrow(data) / (2 * length(which(data$class_efficiency == "efficient")))
      } else if (is.data.frame(parameters[["weights"]])) {
        w0 <- parameters[["weights"]][["w0"]]
        w1 <- parameters[["weights"]][["w1"]]
      }

    }

    model_fit <- train(
      class_efficiency ~ .,
      data = data,
      method = "glm",
      family = binomial(),
      trControl = trControl,
      weights  = ifelse(data$class_efficiency == "efficient", w1, w0),
      metric = "Accuracy"
    )

    if (trControl[["method"]] != "none") {
      model_fit[["results"]] <- model_fit[["results"]][-1]

      model_fit[["results"]] <- cbind(w0, w1, model_fit[["results"]])
    }

  }

  return(model_fit)

}

#' @title Create New SMOTE Units to Balance Data combinations of m + s
#'
#' @description This function creates new DMUs to address data imbalances. If the majority class is efficient, it generates new inefficient DMUs by worsering the observed units. Conversely, if the majority class is inefficient, it projects inefficient DMUs to the frontier. Finally, a random selection if performed to keep a proportion of 0.65 for the majority class and 0.35 for the minority class.
#'
#' @param data A \code{data.frame} containing the variables used in the model.
#' @param x Column indexes of the input variables in the \code{data}.
#' @param y Column indexes of the output variables in the \code{data}.
#' @param z_numeric Integer vector with column indices of numeric environment variables in \code{data}. By default is \code{NULL}.
#' @param z_factor Integer vector with column indices of factor environment variables in \code{data}. By default is \code{NULL}.
#' @param B number of bootstrap replicates in Conditional DEA.
#' @param m number of units to be included in the reference set.
#' @param alpha This allow to choose the size of the Confidence Intervals computed. By defaulta alpha = FALSE. In this case no confidence interval are computed.
#' @param RTS Text string or number defining the underlying DEA technology /
#'   returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
#'     \item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
#'   }
#' @param balance_data Indicate level of efficient units to achive and the number of efficient and not efficient units.
#' @param bandwidth the bandwidth parameters for the unconditional kernel density estimator used in the conditional DEA framework. It is typically obtained using \code{\link[np]{npudensbw}} and supports mixed data types, including continuous variables and discrete unordered or ordered factors. Bandwidths can be selected using normal reference rules, likelihood cross-validation, or least-squares cross-validation following Li and Racine (2003). If \code{NULL}, the bandwidth is estimated internally.
#' @param seed  Integer. Seed for reproducibility.
#'
#'
#' @importFrom dplyr %>% semi_join
#'
#' @return It returns a \code{data.frame} with the newly created set of DMUs incorporated.
#' @export

SMOTE_Z_data <- function (
    data, x, y, z_numeric, z_factor, balance_data, RTS = "vrs", B, m, alpha,
    bandwidth = NULL, seed
) {

  n <- nrow(data)
  exogenous <- c(z_numeric, z_factor)

  # expand grid of regions
  if(!is.null(z_factor)) {

    data_factor <- data[,z_factor]
    grid_factor <- unique(data_factor)

  } else {

    grid_factor <- data.frame(1)

  }

  # number of unique factor combiantions
  n_grid <- NROW(grid_factor)

  # create structure to save datasets balanced
  balance_datasets <- vector("list", length(balance_data))
  names(balance_datasets) <- balance_data
  n_balanced <- length(balance_data)

  # original similarity matrix
  s <- matrix(NA, ncol = n, nrow = n)

  # calculate similarity matrix
  if (is.null(bandwidth)) {
    bw <- npudensbw(dat = data[,exogenous])
    bandwidth <- bw

    bw[["bw"]][(length(z_numeric)+1):length(bw[["bw"]])] <- rep(0, length(z_factor))

  } else {
    bw <- bandwidth

    bw[["bw"]][(length(z_numeric)+1):length(bw[["bw"]])] <- rep(0, length(z_factor))
  }

  for (i in 1:n) {
    kerz <- npudens(
      bws = bw,
      cykertype = "epanechnikov",
      cxkertype = "epanechnikov",
      tdat = data[i,exogenous, drop = FALSE],
      edat = data[,exogenous, drop = FALSE])

    s[i, ] <- cbind(kerz$dens)
  }
  s <- as.data.frame(s)

  # ----------------------------------------------------------------------------
  # Create similarity subgroups ------------------------------------------------
  # ----------------------------------------------------------------------------
  # minimum size of each sub-group
  min_group_size <- max(length(x) * length(y), 3 * (length(x) + length(y)))

  # Inicialize
  data$group_similarity <- NA
  group <- 0
  assigned <- rep(FALSE, nrow(data))

  # Group by z_factor
  for (z_i in 1:n_grid) { # sig = z_i

    name_z_i <- as.data.frame(grid_factor[z_i,])
    names(name_z_i) <- names(grid_factor)

    # filter by z_factor_i
    data_z_i <- semi_join(data, data_factor[z_i,], by = names(name_z_i))

    # which are potential candidates
    idxs <- which(!assigned)

    if (length(idxs) < min_group_size) next

    while (any(!assigned[idxs])) {
      candidates <- idxs[!assigned[idxs]]
      random_DMU <- sample(candidates, 1)

      sim_vals <- s[random_DMU, idxs]
      similar <- idxs[which(sim_vals > 0.01 & !assigned[idxs])]

      if (length(similar) >= min_group_size) {
        group <- group + 1
        data$group_similarity[similar] <- group
        assigned[similar] <- TRUE
      } else {
        assigned[random_DMU] <- TRUE
      }
    }
  }

  # Add no subgroup
  data_no_subgroup <- which(is.na(data$group_similarity))

  for (current_id in data_no_subgroup) {

    # filter by z_factor_i
    current_z_factor <- data[current_id, z_factor]

    ### new
    idx_env_match <- function(data, current_z_factor, z_factor = NULL) {

      env_cols <- names(current_z_factor)

      # Comprobar que existen en data
      missing_cols <- setdiff(env_cols, names(data))
      if (length(missing_cols) > 0) {
        stop("These columns are not in `data`: ", paste(missing_cols, collapse = ", "))
      }

      # 2) Extract the "target" (the specific combination)
      if (!is.data.frame(current_z_factor)) {
        stop("`current_z_factor` must be a 1-row data.frame/tibble with columns = environmental variables.")
      }
      if (nrow(current_z_factor) != 1) {
        stop("`current_z_factor` must have exactly 1 row.")
      }

      target <- current_z_factor[1, env_cols, drop = FALSE]

      # 3) Comparar fila a fila en TODAS las variables de entorno (robusto con factors)
      mat <- sapply(env_cols, function(cl) {
        as.character(data[[cl]]) == as.character(target[[cl]])
      })

      idx <- which(rowSums(mat) == length(env_cols))
      idx
    }

    idx <- idx_env_match(data = data, current_z_factor = current_z_factor, z_factor = z_factor)

    ### end new

    # DMUs which have their group
    candidates <- which(!is.na(data$group_similarity))

    # DMUs which hace the same z_factor
    candidates <- candidates[candidates %in% idx]

    if (length(candidates) == 0) {
      group <- group + 1
      data$group_similarity[current_id] <- group
    } else {

      sim_vals <- s[current_id, candidates]
      ref_unit <- candidates[which.max(sim_vals)]
      data$group_similarity[current_id] <- data$group_similarity[ref_unit]
    }
  }
  #
  # resumen_final <- data %>%
  #   group_by(group_similarity) %>%
  #   summarise(
  #     n = dplyr::n(),
  #     prop_efficient = mean(class_efficiency == "efficient")
  #   )

  # Step 2: determine the efficient facets
  for (group_i in unique(data[["group_similarity"]])) {

    data_i <- data[data[["group_similarity"]] == group_i,]
    data_i <- data_i[, setdiff(names(data_i), "group_similarity")]

    # second, populate the efficient facets
    if (nrow(data_i) >= min_group_size) {

      facets <- convex_facets(
        data = data_i,
        x = x,
        y = y,
        RTS = RTS
      )

      if (nrow(facets) == 0) {
        balance_datasets_i <- vector("list", length = length(balance_data))
        names(balance_datasets_i) <- balance_data

        balance_datasets_i[] <- list(data_i)

      } else {

        balance_datasets_i <- get_SMOTE_DMUs(
          data = data_i,
          facets = facets,
          x = x,
          y = y,
          z_numeric = z_numeric,
          z_factor = z_factor,
          alpha = alpha,
          balance_data = balance_data,
          bandwidth = bandwidth,
          seed = seed
        )

      }

    } else {

      balance_datasets_i <- vector("list", length = length(balance_data))
      names(balance_datasets_i) <- balance_data

      balance_datasets_i[] <- list(data_i)

    }

    # merge new datasets
    balance_datasets <- Map(function(base_df, add_df) {
      if (is.null(base_df)) return(add_df)
      if (is.null(add_df))  return(base_df)
      rbind(base_df, add_df)
    }, balance_datasets, balance_datasets_i)

  }

  return(balance_datasets)
}

#' Spanish Food Industry Firms Dataset
#'
#' Dataset containing information on food industry companies located in Spain,
#' used to illustrate efficiency analysis within the \pkg{PEAXAI} package.
#' The dataset reflects the institutional and market heterogeneity that shapes firm-level efficiency
#' across Spain’s 17 autonomous communities.
#'
#' @details
#' The dataset includes 917 food industry firms with more than 50 employees, collected from the
#' \emph{SABI} database for the year 2023. Each observation corresponds to a single company.
#' Variables reflect both operational and financial dimensions relevant for productivity and efficiency assessment.
#'
#' The output variable is:
#' \itemize{
#'   \item \code{operating_income} — Operating income (in millions of euros), measuring revenues generated
#'   from core business activities.
#' }
#'
#' The input variables are:
#' \itemize{
#'   \item \code{total_assets} — Total assets (millions of euros), representing resources employed.
#'   \item \code{employees} — Number of employees, indicating workforce size (only firms with more than 50 workers are included).
#'   \item \code{fixed_assets} — Tangible fixed assets (millions of euros), such as buildings and machinery.
#'   \item \code{personnel_expenses} — Personnel expenses (millions of euros), including salaries, benefits, and training.
#' }
#'
#' The variable \code{autonomous_community} identifies the territorial location of each firm within Spain.
#'
#' The sample displays substantial dispersion across variables, encompassing both small and large firms.
#' This heterogeneity affects measures of central tendency—mean and median values differ considerably—thus providing a realistic challenge for efficiency and explainability analyses.
#'
#' @format
#' A \code{data.frame} with 917 rows and 6 columns:
#' \describe{
#'   \item{total_assets}{Total assets (millions of euros).}
#'   \item{employees}{Number of employees.}
#'   \item{fixed_assets}{Tangible fixed assets (millions of euros).}
#'   \item{personnel_expenses}{Personnel expenses (millions of euros).}
#'   \item{operating_income}{Operating income (millions of euros).}
#'   \item{autonomous_community}{Autonomous community where the firm operates.}
#' }
#'
#' @source
#' SABI (Sistema de Análisis de Balances Ibéricos) database, 2023.
#' Firms with more than 50 employees in the Spanish food industry.
#'
#' @examples
#' data(firms)
#' str(firms)
#' summary(firms)
#'
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   ggplot2::ggplot(firms, ggplot2::aes(x = employees, y = operating_income)) +
#'     ggplot2::geom_point(alpha = 0.6) +
#'     ggplot2::labs(
#'       x = "Number of employees",
#'       y = "Operating income (millions of euros)",
#'       title = "Spanish Food Industry Firms (2023)"
#'     ) +
#'     ggplot2::theme_minimal() +
#'     ggplot2::theme(
#'       plot.title = ggplot2::element_text(face = "bold"),
#'       axis.line = ggplot2::element_line(color = "black"),
#'       axis.ticks = ggplot2::element_line(color = "black"),
#'       panel.grid.minor = ggplot2::element_blank()
#'     )
#' }
#'
#' @docType data
#' @name firms
#' @usage data(firms)
#' @keywords datasets
"firms"

#' @title Prepare Training and Target Datasets from a caret Model
#'
#' @description
#' Extracts and formats the training and/or target datasets from a machine learning model trained with \code{caret::train},
#' allowing for distinction between using the full training data or only the original subset used for modeling.
#' It standardizes the class column to be named \code{"class_efficiency"} and positions it as the last column.
#'
#' @param data A \code{data.frame} containing the original dataset used to train the model. Only needed when using \code{"real"} as background or target.
#' @param x Not currently used. Reserved for future input variable selection.
#' @param y Not currently used. Reserved for future output variable specification.
#' @param final_model A trained model object of class \code{"train"} from the \pkg{caret} package.
#' @param background A character string, either \code{"train"} or \code{"real"}, specifying the background dataset used for explainability.
#' @param target A character string, either \code{"train"} or \code{"real"}, specifying the target dataset to be explained.
#' @param type Not currently used. Reserved for future prediction types.
#' @param threshold Not currently used. Reserved for future thresholding logic.
#' @param levels_order A character vector specifying the levels of the response factor, typically \code{c("not_efficient", "efficient")}. Not currently used, but can help in reordering or relabeling.
#'
#' @return A \code{list} with two elements:
#' \describe{
#'   \item{\code{train_data}}{A \code{data.frame} representing the background dataset, with the class column renamed to \code{"class_efficiency"} and positioned last.}
#'   \item{\code{target_data}}{A \code{data.frame} representing the target dataset, formatted in the same way.}
#' }
#'

xai_prepare_sets <- function (
    data, x, y, final_model, background, target,
    type, threshold, levels_order
    ) {

  # domain set; by default is "train"
  if (background == "train") {

    # a ML model train by caret
    # save train data, change name and position
    train_data <- final_model[["trainingData"]]
    names(train_data)[1] <- "class_efficiency"

    train_data <- train_data[,c(2:length(train_data),1)]

  } else if (background == "real") {

    n <- nrow(data)

    # a ML model train by caret
    # save train data, change name and position
    train_data <- final_model[["trainingData"]][1:n,]
    names(train_data)[1] <- "class_efficiency"

    train_data <- train_data[,c(2:length(train_data),1)]


  }

  # target data set; by default is "train"
  if (target == "train") {

    # a ML model train by caret
    # save train data, change name and position
    target_data <- final_model[["trainingData"]]
    names(target_data)[1] <- "class_efficiency"

    target_data <- target_data[,c(2:length(target_data),1)]

  } else if (target == "real") {

    n <- nrow(data)

    # a ML model train by caret
    # save train data, change name and position
    target_data <- final_model[["trainingData"]][1:n,]
    names(target_data)[1] <- "class_efficiency"

    target_data <- target_data[,c(2:length(target_data),1)]

  }

  output <- list(
    train_data = train_data,
    target_data = target_data
  )

  return(output)

}

#' @title Identify Benchmark Peers Based on Estimated Efficiency Probabilities
#'
#' @description
#' Identifies peer units (i.e., reference benchmarks) for each decision-making unit (DMU) based on predicted probabilities of technical efficiency.
#' Given a fitted classification model that estimates the probability of being efficient, the function selects, for each DMU, its nearest efficient peer
#' according to Euclidean or weighted distances. Multiple efficiency thresholds can be specified to assess different levels of benchmarking stringency.
#'
#' @param data A \code{data.frame} or \code{matrix} containing input and output variables used in the efficiency model.
#' @param x Integer vector indicating the column indices of input variables in \code{data}.
#' @param y Integer vector indicating the column indices of output variables in \code{data}.
#' @param final_model A fitted classification model used to estimate efficiency probabilities. Supported classes: \code{"train"} (from \pkg{caret}) or \code{"glm"} (binomial).
#' @param calibration_model Optional probability-calibration model applied to the raw
#' predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
#' If provided, calibrated probabilities are used for ranking and threshold-based decisions.
#' Set to \code{NULL} to use uncalibrated predictions.
#' @param efficiency_thresholds Numeric vector indicating the minimum probability values required to consider a DMU as efficient.
#' @param targets A named list containing, for each efficiency threshold, the corresponding

#' @param weighted Logical. If \code{TRUE}, peers are selected using weighted Euclidean distances based on variable importance. If \code{FALSE} (default), unweighted distances are used.
#' @param relative_importance Optional named numeric vector indicating the relative importance of each input/output variable (used when \code{weighted = TRUE}).
#'
#' @return A named list of matrices. Each element corresponds to an efficiency threshold and contains, for each DMU, the index of the closest efficient peer.
#' If \code{weighted = FALSE}, the list contains unweighted peers. If \code{weighted = TRUE}, the list contains weighted peers.
#'
#' @details
#' This function enables probabilistic peer identification under uncertainty, supporting flexible definitions of efficiency based on thresholds over estimated probabilities.
#' When \code{weighted = TRUE}, variable weights (e.g., derived from feature importance) modulate the peer selection process, allowing for context-aware benchmarking.
#'
#' @importFrom stats predict
#' @importFrom utils head
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   x <- 1:4
#'   y <- 5
#'   RTS <- "vrs"
#'   imbalance_rate <- NULL
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'       weights = "dinamic"
#'      )
#'    )
#'
#'   metric_priority <- c("Balanced_Accuracy", "ROC_AUC")
#'
#'   models <- PEAXAI_fitting(
#'     data = data, x = x, y = y, RTS = RTS,
#'     imbalance_rate = imbalance_rate,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = metric_priority,
#'     verbose = FALSE,
#'     seed = 1
#'   )
#'
#'   final_model <- models[["best_model_fit"]][["glm"]]
#'
#'   importance_method <- list(name = "PI", n.repetitions = 5)

#'
#'   relative_importance <- PEAXAI_global_importance(
#'     final_model = final_model,
#'     x = x,
#'     y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   efficiency_thresholds <- seq(0.75, 0.95, 0.1)
#'
#'   directional_vector <- list(
#'    relative_importance = relative_importance,
#'    baseline  = "mean"
#'   )
#'
#'   targets <- PEAXAI_counterfactuals(
#'     data = data,
#'     x = x, y = y,
#'     final_model = final_model,
#'     efficiency_thresholds = efficiency_thresholds,
#'     directional_vector = directional_vector,
#'     n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
#'   )
#'
#'   peers <- PEAXAI_peer(data = data, x = x, y = y, final_model = final_model,
#'   efficiency_thresholds = efficiency_thresholds, targets = targets, weighted = FALSE)
#' }
#'
#' @export
#'

PEAXAI_peer <- function(
    data, x, y, final_model, calibration_model = NULL,
    efficiency_thresholds, targets, weighted = FALSE,
    relative_importance = NULL
    ) {

  validate_parametes_PEAXAI_peer <- validate_parametes_PEAXAI_peer(
    data, x, y, final_model, efficiency_thresholds,
    weighted, relative_importance
  )

  # reorder index 'x' and 'y' in data
  data <- data[, c(x,y)]
  x <- 1:(ncol(data) - length(y))
  y <- (length(x) + 1):ncol(data)

  variables <- c(x,y)

  # ----------------------------------------------------------------------------
  # predict probability --------------------------------------------------------
  # ----------------------------------------------------------------------------
  data <- as.data.frame(data)

  if (inherits(final_model, "train")) {
    # caret::train
    # prob_vector <- predict(final_model, newdata = data, type = "prob")["efficient"]
    prob_vector <- PEAXAI_predict(
      data = data,
      x = x,
      y = y,
      final_model = final_model,
      calibration_model = calibration_model
    )

  } else {
    stop("Unsupported model type.")
  }

  # ----------------------------------------------------------------------------
  # get peers ------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  # data_scenario_list <- list()
  # metrics_list <- list()
  peer_list <- list()
  peer_weight_list <- list()
  na_count_list <- list()
  n_not_prob_list <- list()

  for (thr in efficiency_thresholds) {
    message(paste0("Efficiency threshold: ", thr))

    # check if there are some DMUs which have the threshold desired
    if (max(prob_vector) < thr) {

      peer_list[[as.character(thr)]] <-NA
      peer_weight_list[[as.character(thr)]] <- NA
      next
    }

    # first, determinate efficient units
    if (inherits(final_model, "train")) {
      idx_eff <- which(prob_vector > thr)
    }
    # else if (inherits(final_model, "glm")) {
    #   idx_eff <- which(prob_vector > thr)
    # }

    if (weighted == FALSE) {

      # ------------------------------------------------------------------------
      # normal peers -----------------------------------------------------------
      # ------------------------------------------------------------------------

      # save distances structure
      save_dist <- matrix(
        data = NA,
        ncol = length(idx_eff),
        nrow = nrow(data)
      )

      # calculate distances
      for (unit_eff in idx_eff) {

        # set reference
        reference <- data[unit_eff, variables]
        # reference <- targets[[as.character(thr)]][["counterfactual_dataset"]][unit_eff, variables]

        distance <- unname(apply(targets[[as.character(thr)]][["counterfactual_dataset"]][, variables], 1, function(x) {
          sqrt(sum((x - reference)^2))
        }
        ))

        # get position in save results
        idx_dis <- which(idx_eff == unit_eff)

        save_dist[,idx_dis] <- as.matrix(distance)
      }

      # near DMU efficient for projection
      near_idx_eff <- apply(save_dist, 1, function(row) {

        which.min(abs(row))

      })

      peer_restult <- matrix(
        data = NA,
        ncol = 1,
        nrow = nrow(data)
      )

      peer_restult[, 1] <- idx_eff[near_idx_eff]

      # save_peer
      peer_list[[as.character(thr)]] <- peer_restult

    } else {

      # ------------------------------------------------------------------------
      # weighted peers ---------------------------------------------------------
      # ------------------------------------------------------------------------

      # save weighted distances structure
      save_dist_weight <- matrix(
        data = NA,
        ncol = length(idx_eff),
        nrow = nrow(data)
      )

      # # calculate weighted distances
      # result_importance_matrix <- as.data.frame(matrix(
      #   data = rep(unlist(relative_importance[variables]), each = nrow(data)),
      #   nrow = nrow(data),
      #   ncol = ncol(data[,variables]),
      #   byrow = FALSE
      # ))
      # names(result_importance_matrix) <- names(data)[variables]

      w <- as.numeric(relative_importance[1, variables])
      names(w) <- variables

      w_eval_data <- sweep(data[, variables, drop = FALSE], 2, w, `*`)

      for (unit_eff in idx_eff) {

        # set reference
        # reference <- data[unit_eff, variables]
        reference <- sweep(targets[[as.character(thr)]][["counterfactual_dataset"]][unit_eff, variables, drop = FALSE], 2, w, `*`)

        distance <- unname(apply(w_eval_data[, variables], 1, function(row) {
          sqrt(sum(((row - reference)^2)))
        }))

        # get position in save results
        idx_dis <- which(idx_eff == unit_eff)
        save_dist_weight[,idx_dis] <- as.matrix(distance)

      }

      near_idx_eff_weight <- apply(save_dist_weight, 1, function(row) {

        which.min(abs(row))

      })

      peer_restult_weight <- matrix(
        data = NA,
        ncol = 1,
        nrow = nrow(save_dist_weight)
      )

      peer_restult_weight[, 1] <- idx_eff[near_idx_eff_weight]

      # save_peer
      peer_weight_list[[as.character(thr)]] <- peer_restult_weight

    }

  } # end loop thresholds

  if (weighted == FALSE) {
    peer <- as.data.frame(peer_list, check.names = FALSE)
    peer <- cbind(DMU = seq_len(nrow(peer)), peer)
    return(peer)
  } else {
    peer_weight <- as.data.frame(peer_weight_list, check.names = FALSE)
    peer_weight <- cbind(DMU = seq_len(nrow(peer_weight)), peer_weight)
    return(peer_weight)
  }

}

#' @title Identify Maximal Facets of the Convex Frontier
#'
#' @description
#' Identifies the maximal subsets of efficient Decision-Making Units (DMUs)
#' that lie on the same face (facet) of the efficient frontier. It relies on
#' the concept of "maximal friends" to determine which efficient units can
#' be grouped together to form convex combinations under the chosen returns-to-scale assumption.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables used in the model.
#' @param x Integer vector indicating the column indexes of the input variables in \code{data}.
#' @param y Integer vector indicating the column indexes of the output variables in \code{data}.
#' @param RTS Text string or number defining the underlying DEA technology /
#'   returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
#'     \item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
#'   }
#'
#' @importFrom deaR make_deadata maximal_friends
#'
#' @return A \code{matrix} where each row represents a maximal facet, and the values are the row indices of the efficient DMUs in \code{data} that form that facet. Returns an empty \code{matrix} if there are no efficient units.

convex_facets <- function (
    data, x, y, RTS = "vrs"
) {

  # ----------------------------------------------------------------------------
  # Determine the efficient combinations by Additive DEA -----------------------
  # ----------------------------------------------------------------------------

  if (length(which(data$class_efficiency == "efficient")) == 0) {

    results_convx <- matrix(nrow = 0, ncol = 0)

  } else {

    # Obtain all possible combinations of maximal friends (facets)
    datadea <- make_deadata(
      data,
      inputs = x,
      outputs = y
    )

    eff_convex_list <- maximal_friends(
      datadea = datadea,
      rts = RTS,
      dmu_ref = which(data$class_efficiency == "efficient"))

    # Methodological choice: We only retain the facets with the maximum
    # number of vertices (i.e. full-dimensional facets in the dataset).
    long <- lengths(eff_convex_list)
    max_long <- max(long)

    list_max_facets <- eff_convex_list[long == max_long]

    # Combine into a matrix where each row represents a facet and
    # columns represent the indices of the DMUs forming it
    results_convx <- do.call(rbind, list_max_facets)

  }

  return(results_convx)

}

#' @title Generate Efficiency Rankings Based on Probabilistic Classification
#'
#' @description
#' Produces efficiency rankings of decision-making units (DMUs) according to the probabilities
#' estimated by a fitted classification model. Two ranking modes are supported:
#' \itemize{
#'   \item \code{"predicted"}: ranks DMUs solely by their predicted probability of being efficient.
#'   \item \code{"attainable"}: ranks DMUs hierarchically according to:
#'         (1) the attainable (target) efficiency probability,
#'         (2) the size of the improvement parameter \eqn{\beta} (smaller is better),
#'         and (3) the predicted efficiency probability (higher is better).
#' }
#' This allows to integrate both predictive and counterfactual (attainable) information
#' into the efficiency ranking.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the input and output variables.
#' @param x Integer vector specifying the column indices of input variables in \code{data}.
#' @param y Integer vector specifying the column indices of output variables in \code{data}.
#' @param final_model A fitted classification model used to estimate efficiency probabilities.
#' Supported types are:
#' \itemize{
#'   \item \code{"train"}: an object fitted with \pkg{caret}.
#'   \item \code{"glm"}: a binomial logistic regression model.
#' }
#' @param calibration_model Optional probability-calibration model applied to the raw
#' predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
#' If provided, calibrated probabilities are used for ranking and threshold-based decisions.
#' Set to \code{NULL} to use uncalibrated predictions.
#' @param efficiency_thresholds Numeric vector defining one or more efficiency probability
#' thresholds to determine the attainable frontier or peer set.
#' @param targets A named list containing, for each efficiency threshold, the corresponding
#' attainable targets and estimated \eqn{\beta} values (e.g., obtained from counterfactual analysis).
#' Each element should be a list with a component named \code{"beta"}.
#' @param rank_basis Character string specifying the ranking criterion. Options are:
#' \itemize{
#'   \item \code{"predicted"}: order units by predicted efficiency probability.
#'   \item \code{"attainable"}: order by attainable probability, then by \eqn{\beta},
#'   and finally by predicted probability (see Details).
#' }
#'
#' @details
#' The attainable-based ranking combines predictive efficiency with the modeled potential
#' for improvement (\eqn{\beta}) and the probability of reaching a target frontier level.
#' This approach yields a more nuanced and interpretable prioritization of DMUs, reflecting
#' both their current and achievable performance under the estimated model.
#'
#' When \code{rank_basis = "attainable"}, ties in attainable probability are broken first
#' by the magnitude of \eqn{\beta} (ascending), and then by the predicted probability
#' (descending).
#'
#' @return
#' \itemize{
#'   \item If \code{rank_basis = "predicted"}: a \code{data.frame} sorted by predicted efficiency probability.
#'   \item If \code{rank_basis = "attainable"}: a named list of \code{data.frame}s, one per efficiency threshold,
#'   each sorted according to the hierarchical ranking scheme described above.
#' }
#'
#' @importFrom stats predict
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   x <- 1:4
#'   y <- 5
#'   RTS <- "vrs"
#'   imbalance_rate <- NULL
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'       weights = "dinamic"
#'      )
#'   )
#'
#'   metric_priority <- c("Balanced_Accuracy", "ROC_AUC")
#'
#'   models <- PEAXAI_fitting(
#'     data = data, x = x, y = y, RTS = RTS,
#'     imbalance_rate = imbalance_rate,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = metric_priority,
#'     verbose = FALSE,
#'     seed = 1
#'   )
#'
#'   final_model <- models[["best_model_fit"]][["glm"]]
#'
#'   importance_method <- list(name = "PI", n.repetitions = 5)
#'
#'   relative_importance <- PEAXAI_global_importance(
#'     final_model = final_model,
#'     x = x,
#'     y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   efficiency_thresholds <- seq(0.75, 0.95, 0.1)
#'
#'   directional_vector <- list(
#'     relative_importance = relative_importance,
#'     baseline  = "mean"
#'   )
#'
#'   targets <- PEAXAI_counterfactuals(
#'     data = data,
#'     x = x, y = y,
#'     final_model = final_model,
#'     efficiency_thresholds = efficiency_thresholds,
#'     directional_vector = directional_vector,
#'     n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
#'   )
#'
#'   ranking <- PEAXAI_ranking(data = data, x = x, y = y,
#'   final_model = final_model, rank_basis = "predicted")
#' }
#'
#' @export

PEAXAI_ranking <- function(
    data, x, y, final_model, calibration_model = NULL,
    efficiency_thresholds, targets = NULL, rank_basis
    ) {

  validate_parametes_PEAXAI_ranking(
    data, x, y, final_model,
    efficiency_thresholds, targets, rank_basis
  )

  # reorder index 'x' and 'y' in data
  data <- data[, c(x,y)]
  x <- 1:(ncol(data) - length(y))
  y <- (length(x) + 1):ncol(data)

  # ----------------------------------------------------------------------------
  # predict probability --------------------------------------------------------
  # ----------------------------------------------------------------------------
  data <- as.data.frame(data)

  if (inherits(final_model, "train")) {
    # caret::train
    prob_vector <- PEAXAI_predict(
      data = data,
      x = x,
      y = y,
      final_model = final_model,
      calibration_model = calibration_model
    )
    # prob_vector <- predict(final_model, newdata = data, type = "prob")["efficient"]
  } else {
    stop("Unsupported model type.")
  }

  DMU <- as.data.frame(c(1:nrow(data)))
  names(DMU) <- "DMU"
  prob_vector <- cbind(DMU, prob_vector)
  names(prob_vector)[2] <- "Probability_predicted"

  # ----------------------------------------------------------------------------
  # get ranking ----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (rank_basis == "predicted") {

    position <- as.data.frame(1:nrow(prob_vector))
    names(position) <- "Ranking"
    ranking_order <- prob_vector[order(prob_vector[["Probability_predicted"]], decreasing = TRUE), ]

    ranking_order <- cbind(position, ranking_order)
    row.names(ranking_order) <- NULL

    return(ranking_order)

  } else if (rank_basis == "attainable") {

    list_ranking <- vector("list", length = length(efficiency_thresholds))
    names(list_ranking) <- efficiency_thresholds

    for(thr in as.character(efficiency_thresholds)) {

      position <- as.data.frame(1:nrow(prob_vector))
      names(position) <- "Ranking"

      ranking_order <- cbind(prob_vector, targets[[thr]][["inefficiencies"]])
      names(ranking_order)[4] <- "Probability_target"

      ranking_order <- ranking_order[order(
        -ranking_order$Probability_target,
        ranking_order$betas,
        -ranking_order$Probability_predicted),]

      ranking_order <- cbind(position, ranking_order)
      row.names(ranking_order) <- NULL

      list_ranking[[thr]] <- ranking_order

    }

    return(list_ranking)

  }

}

#' @title Data preprocessing and efficiency labeling with Additive DEA
#'
#' @description
#' Labels each DMU (Decision Making Unit) as efficient or not using the
#' Additive DEA model, optionally after basic data preprocessing. It supports both
#' standard unconditional DEA and conditional DEA frameworks (when exogenous variables
#' are provided). The resulting factor \code{class_efficiency} has levels
#' \code{c("not_efficient","efficient")}, where \code{"efficient"} is the positive
#' class for downstream modeling.
#'
#' @param data A \code{data.frame} or \code{matrix} containing all variables.
#' @param REF Optional reference set of inputs that defines the technology
#'   (defaults to the columns indicated by \code{x} in \code{data}). Must have
#'   the same number of rows as \code{data}.
#' @param x Integer vector with column indices of input variables in \code{data}.
#' @param y Integer vector with column indices of output variables in \code{data}.
#' @param z_numeric Integer vector with column indices of continuous/numeric exogenous variables in \code{data}. By default is \code{NULL}.
#' @param z_factor Integer vector with column indices of discrete/factor exogenous variables in \code{data}. By default is \code{NULL}.
#' @param RTS Character or integer specifying the DEA technology / returns-to-scale
#'   assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull (no convexity).}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale (convexity + free disposability).}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale (convexity, down-scaling, free disposability).}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale (convexity + free disposability).}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling only, convexity + free disposability).}
#'     \item{\code{5} / \code{"add"}}{Additivity (integer up/down scaling) with free disposability.}
#'   }
#' @param B Integer. The number of bootstrap replicates (iterations) in the conditional DEA framework.
#' @param alpha Numeric. The nominal size of the confidence interval (e.g., \code{0.05}). If \code{FALSE}, no confidence intervals are computed.
#' @param m Integer. The number of units to be drawn to form the reference set in each bootstrap replicate of the conditional DEA.
#' @param bandwidth Optionally, the bandwidth parameters for the unconditional kernel density estimator used in the conditional DEA framework. Typical estimation is done using \code{\link[np]{npudensbw}} which supports mixed continuous and discrete data types.
#' @param seed Integer or \code{NULL}. Seed for pseudo-random number generator ensuring reproducibility in the bootstrap sampling process.
#'
#' @details
#' Internally relies on \code{\link[Benchmarking]{dea.add}} to compute Additive DEA
#' scores and derive the binary efficiency label. When exogenous variables \code{z_numeric}
#' or \code{z_factor} are provided, a conditional DEA is performed by sampling a reference
#' set of size \code{m} internally \code{B} times, based on distances computed via kernel
#' density estimation.
#'
#' @importFrom Benchmarking dea.add
#' @importFrom np npudensbw npudens
#'
#' @return
#' A \code{data.frame} equal to \code{data} (retaining all input \code{x} and
#' output \code{y} columns) plus a new factor column \code{class_efficiency}
#' with levels \code{c("not_efficient","efficient")}.
#'
#' @seealso \code{\link[Benchmarking]{dea.add}}, \code{\link[rcDEA]{conditional_DEA}}
#'
#' @examples
#' \donttest{
#' # Example (assuming columns 1:2 are inputs and 3 is output):
#' # out <- label_efficiency(data = df, x = 1:2, y = 3, RTS = "vrs")
#' # table(out$class_efficiency)
#' }
#'
#' @export

label_efficiency <- function (
    data, REF = data, x, y, z_numeric = NULL, z_factor = NULL, RTS = "vrs",
    B = 1, alpha = FALSE, m = NULL, bandwidth = NULL, seed
  ) {

  # check if parameters are well introduced
  validate_parametes_label_efficiency(
    data = data,
    x = x,
    y = y,
    RTS = RTS
  )

  if (is.null(z_numeric) & is.null(z_factor)) {

    #
    tol = 1e-9
    X <- as.matrix(data[,x])
    Y <- as.matrix(data[,y])

    Z <- cbind(-X, Y)
    n <- nrow(Z)

    pareto <- rep(TRUE, n)

    for (k in seq_len(n)) {
      # En Z = [-X, Y], dominar significa ser >= componente a componente
      weak_dom <- rowSums(Z >= matrix(Z[k, ], n, ncol(Z), byrow = TRUE) - tol) == ncol(Z)
      strict_dom <- rowSums(Z > matrix(Z[k, ], n, ncol(Z), byrow = TRUE) + tol) > 0

      pareto[k] <- !any(weak_dom & strict_dom)
    }

    pareto
    #
    # the additive DEA

    # benchmarking to calculate additive-DEA
    add_scores <- dea.add(
      X = as.matrix(data[,x]),
      Y = as.matrix(data[,y]),
      XREF = as.matrix(REF[pareto,x]),
      YREF = as.matrix(REF[pareto,y]),
      RTS = RTS
    )[["sum"]]

    # determine efficient and inefficient DMUs
    class_efficiency <- ifelse(add_scores <= 0.0001, 1, 0)

    # add new varaible
    data <- as.data.frame (
      cbind(data, class_efficiency)
    )

    # labels
    data$class_efficiency <- ifelse(
      data$class_efficiency == 1, "efficient", "not_efficient")

    # create the 'class_efficiency' in a factor
    # order of levels
    data$class_efficiency <- factor(
      data$class_efficiency,
      levels = c("efficient", "not_efficient"))

  } else {

    # conditional DEA

    # define preliminary variables:
    n <- nrow(data) # number of observations
    k <- NCOL(data[,c(y)]) # number of outputs
    s <- NCOL(data[,c(x)]) # number of inputs

    # initialize
    eff <- rep(0, n)
    DEA_B <- rep(0, B) # bootstrap

    if (alpha != FALSE) {
      ci_low <- rep(0, n)
      ci_up <- rep(0,n)
    }

    # 1 determine bandwidths
    message(c("R is now computing the bandwidth using the function npudensbw in the package np"))

    exogenous <- c(z_numeric, z_factor)

    # calculate similarity matrix
    if (is.null(bandwidth)) {
      bw <- npudensbw(dat = data[,exogenous])
    } else {
      bw <- bandwidth
    }

    # 2 similarity by unit
    message(c("R is now computing the conditional DEA"))

    set.seed(seed)
    for (i in 1:n) {

      # compute the similarity for each obs i
      kerz <- npudens(
        bws = bw,
        cykertype = "epanechnikov",
        cxkertype = "epanechnikov",
        tdat = data[i,exogenous, drop = FALSE],
        edat = data[,exogenous, drop = FALSE])

      similarity_i <- as.numeric(kerz$dens)
      sim_target <- similarity_i[i] # Similitud de la unidad i consigo misma

      # consider only the units that perform at least as good as unit i
      y_i <- data[i, y]
      x_i <- data[i, x]
      Y_Rob <- data[, y]
      X_Rob <- data[, x]

      # Filtro 1: Similitud (al menos 0.01 de similitud respecto a la unidad i)
      cond_sim <- (similarity_i / sim_target) >= 0.01

      # Filtro 2: Dominancia fuerte (menor o igual input, mayor o igual output)
      cond_output <- matrix(TRUE, nrow = nrow(Y_Rob), ncol = k)
      for (l in 1:k) {
        cond_output[, l] <- Y_Rob[, l] >= as.numeric(y_i[l])
      }

      cond_input <- matrix(TRUE, nrow = nrow(X_Rob), ncol = s)
      for (l in 1:s) {
        cond_input[, l] <- X_Rob[, l] <= as.numeric(x_i[l])
      }

      cond_dom <- apply(cond_output, 1, all) & apply(cond_input, 1, all)

      # Intersección lógica: cumplen ambas condiciones a la vez
      combined_filter <- cond_sim & cond_dom

      # Aplicar ambos filtros a la vez para construir las matrices de referencia
      Y_ref <- as.data.frame(Y_Rob[combined_filter, , drop = FALSE])
      Y_ref <- unique(Y_ref)

      X_ref <- as.data.frame(X_Rob[combined_filter, , drop = FALSE])
      X_ref <- unique(X_ref)

      n_sample <- nrow(X_ref) #(equivalent to nrow(Y_Rob))

      #pick a sample of random unit in the reference set if there are at least 2 units in the ref
      if (n_sample < 2) {
        eff[i] <- 0
      }
      else {
        for (j in 1:B) {

          # REF <- cbind(X_ref, Y_ref)

          # compute the DEA for unit i
          # DEA_B[j] <- Benchmarking::dea(
          #   X = data[i,x],
          #   Y = data[i,y],
          #   RTS = RTS,
          #   ORIENTATION = "in-out",
          #   DIRECT = TRUE,
          #   XREF = X_ref,
          #   YREF = Y_ref)$eff

          DEA_B[j] <- dea.add(
            X = as.matrix(data[i,x, drop = FALSE]),
            Y = as.matrix(data[i,y, drop = FALSE]),
            XREF = as.matrix(X_ref, drop = FALSE),
            YREF = as.matrix(Y_ref, drop = FALSE),
            RTS = RTS
          )[["sum"]]

          # DEA_B[j] <- tryCatch({
          #   # Intenta ejecutar la función normalmente
          #   dea.add(
          #     X = as.matrix(data[i, x, drop = FALSE]),
          #     Y = as.matrix(data[i, y, drop = FALSE]),
          #     XREF = as.matrix(X_ref, drop = FALSE),
          #     YREF = as.matrix(Y_ref, drop = FALSE),
          #     RTS = RTS
          #   )[["sum"]]
          # }, error = function(e) {
          #   # Si hay error, imprime el mensaje y lanza el browser
          #   message("Se detectó un error en la iteración: ", j)
          #   print(e)
          #   browser()
          #   browser()
          # })
        }

        eff[i] <- mean(DEA_B[DEA_B != -Inf & DEA_B != Inf])
        if (alpha != FALSE) {
          ci_low[i] <- stats::quantile(DEA_B[DEA_B != -Inf & DEA_B != Inf], alpha/2)
          ci_up[i] <- stats::quantile(DEA_B[DEA_B != -Inf & DEA_B != Inf], 1-alpha/2)
        }

      }
    }

    if(alpha != FALSE) {
      save <- data.frame(eff, ci_low, ci_up)
    }
    else {
      save <- data.frame(eff)
    }

    # assing efficiency
    if (alpha != FALSE) {
      labels <- ifelse(round(save$ci_low, 4) == 0, "efficient", "not_efficient")
    } else {
      labels <- ifelse(round(save$eff, 4) == 0, "efficient", "not_efficient")
    }

    data$class_efficiency <- factor(
      labels,
      levels = c("efficient", "not_efficient"))

  }

  return(data)

}

#' @title Global feature importance for efficiency classifiers
#'
#' @description
#' Computes **global feature importance** for a fitted classification model that
#' separates Pareto-efficient DMUs, using one of three XAI backends:
#' \itemize{
#'   \item \code{"SA"} — Sensitivity Analysis via \pkg{rminer}.
#'   \item \code{"SHAP"} — Model-agnostic SHAP approximations via \pkg{kernelshap}.
#'   \item \code{"PI"} — Permutation Importance via \pkg{iml}.
#' }
#' You can evaluate the model on either the training domain (\code{background = "train"})
#' or the real-world domain (\code{background = "real"}) and compute importance on a
#' chosen \code{target} set (\code{"train"} or \code{"real"}). Importances are
#' returned normalized to sum to 1.
#'
#' @param final_model A fitted model. If it is a base-\code{glm} binomial, probabilities
#'   are obtained with \code{type = "response"}; otherwise the function expects
#'   \code{predict(type = "prob")} with a column named \code{"efficient"}.
#' @param x Integer or character vector with the columns used as **inputs** (predictors).
#' @param y Integer or character vector with the columns used as **outputs** (targets used
#'   to define \code{class_efficiency} in training; not included in \code{X} when explaining).
#' @param explain_data A \code{data.frame} with the original observed DMUs passed
#'   by the user to \code{PEAXAI_fitting()}. This dataset represents the real,
#'   non-augmented observations on which explanations can be computed.
#' @param reference_data A \code{data.frame} with the training dataset used to fit
#'   the optimal model returned by \code{PEAXAI_fitting()}. Depending on the class-
#'   balancing procedure, this dataset may contain only real observed DMUs or both
#'   real and synthetic DMUs.
#' @param importance_method A named list (or data.frame-like) with the backend and its args:
#'   \describe{
#'     \item{\code{name}}{One of \code{"SA"}, \code{"SHAP"}, \code{"PI"}.}
#'     \item{\code{method}}{(SA) One of \code{"1D-SA"}, \code{"sens"}, \code{"DSA"}, \code{"MSA"}, \code{"CSA"}, \code{"GSA"}.}
#'     \item{\code{measures}}{(SA) e.g. \code{"AAD"}, \code{"gradient"}, \code{"variance"}, \code{"range"}.}
#'     \item{\code{levels}}{(SA) Discretization levels used by \code{rminer::Importance}.}
#'     \item{\code{baseline}}{(SA) Baseline value for SA, if applicable.}
#'     \item{\code{bg_n}}{(SHAP) Background sample size for \code{kernelshap::kernelshap} (default: 200).}
#'     \item{\code{n.repetitions}}{(PI) Number of permutations per feature for \code{iml::FeatureImp}.}
#'   }
#'
#' @param seed  Integer. Seed for reproducibility.
#'
#' @details
#' Internally, the function builds background/target sets with \code{xai_prepare_sets()}.
#' For \code{glm} models, the positive class is assumed to be the **second level**
#' (\code{"efficient"}) and probabilities are extracted with \code{type = "response"}.
#' For other models (e.g., \pkg{caret}), \code{predict(type = "prob")[, "efficient"]} is used.
#'
#'
#' @return A named numeric vector (or 1-row data.frame) of normalized importances,
#'   with names matching the predictor columns; the values sum to 1.
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   x <- 1:4
#'   y <- 5
#'   RTS <- "vrs"
#'   imbalance_rate <- NULL
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'       weights = "dinamic"
#'      )
#'    )
#'
#'   metric_priority <- c("Balanced_Accuracy", "ROC_AUC")
#'
#'   models <- PEAXAI_fitting(
#'     data = data, x = x, y = y, RTS = RTS,
#'     imbalance_rate = imbalance_rate,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = metric_priority,
#'     seed = 1,
#'     verbose = FALSE
#'   )
#'
#'   final_model <- models[["best_model_fit"]][["glm"]]
#'
#'   importance_method <- list(name = "PI", n.repetitions = 5)
#'
#'   relative_importance <- PEAXAI_global_importance(
#'     final_model = final_model,
#'     x = x,
#'     y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   imp <- PEAXAI_global_importance(
#'     final_model = final_model,
#'     x = x,
#'     y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   head(imp)
#' }
#'
#' @seealso \code{\link[kernelshap]{kernelshap}}, \code{\link[iml]{FeatureImp}},
#'   \code{\link[rminer]{Importance}}
#'
#' @importFrom stats predict glm
#' @importFrom kernelshap kernelshap
#' @importFrom iml FeatureImp Predictor
#' @importFrom rminer Importance
#' @importFrom pROC auc
#'
#' @export

PEAXAI_global_importance <- function(
    final_model, x, y, explain_data, reference_data,
    importance_method, seed = 314
    ) {

  # reproducibility
  set.seed(seed)

  # as data.frame
  explain_data <- as.data.frame(explain_data)
  reference_data <- as.data.frame(reference_data)

  # as data.frame
  explain_data <- as.data.frame(explain_data)
  reference_data <- as.data.frame(reference_data)

  # --------------------------------------------------------

  # ----------------------------------------------------------------------------
  # detecting importance variables ---------------------------------------------
  # ----------------------------------------------------------------------------
  # type of prediction
  if (inherits(final_model, "glm")) {
    # type if is a glm model
    type <- "response"
  } else {
    # type if is a ML model
    type <- "prob"
  }

  # format datasets
  format_dataset <- function(d, x_cols, y_cols) {
    if (".outcome" %in% names(d)) {
      names(d)[names(d) == ".outcome"] <- "class_efficiency"
      d <- d[, c(setdiff(names(d), "class_efficiency"), "class_efficiency")]
      return(d)
    } else {
      d <- d[, c(x_cols, y_cols)]
      return(d)
    }
  }

  target_data <- format_dataset(explain_data, x, y)
  train_data <- format_dataset(reference_data, x, y)

  x_len <- length(x)
  y_len <- length(y)

  # --- SECURITY CHECK AGAINST EXTERNAL DATA ---
  n_explain <- nrow(target_data)
  if (n_explain > nrow(train_data)) {
    stop("Error: 'explain_data' has more rows than 'reference_data'. They must be the original DMUs passed to PEAXAI_fitting().")
  }
  # Extract the features from train_data (dropping class_efficiency) to compare
  ref_subset <- train_data[1:n_explain, setdiff(names(train_data), "class_efficiency"), drop = FALSE]
  # Compare formatted datasets
  if (!isTRUE(all.equal(target_data, ref_subset, check.attributes = FALSE))) {
    stop("Error: 'explain_data' features do not match the original training observations. You must provide the same DMUs used in PEAXAI_fitting().")
  }

  # methods XAI
  if (importance_method[["name"]] == "SA") {

    # importance with our model of Caret
    f_pred <- function(object, newdata) {
      if (inherits(object, "glm")) {
        predict(object, newdata = newdata, type = type)
      } else {
        predict(object, newdata = newdata, type = type)[, "efficient"]
      }
    }

    # Define methods and measures
    method <- if (!is.null(importance_method[["method"]])) importance_method[["method"]] else "1D-SA"
    measure <- if (!is.null(importance_method[["measures"]])) importance_method[["measures"]] else "AAD"
    levels <- if (!is.null(importance_method[["levels"]])) importance_method[["levels"]] else 7

    # matrix of data without label
    dataset_chosen <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]

    dataset_chosen <- as.data.frame(dataset_chosen)

    imp <- rminer::Importance(
      M = final_model,
      data = dataset_chosen,
      RealL = levels,
      method = method,
      measure = measure,
      baseline = importance_method[["baseline"]],
      responses = TRUE,
      PRED = f_pred)[["imp"]]

    imp <- as.data.frame(t(imp))

    names(imp) <- names(dataset_chosen)

    # Normalize
    importance <- imp

  } else if (importance_method[["name"]] == "SHAP") {

    # matrix of data without label
    target_data <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]
    train_data <- train_data[, setdiff(names(train_data), "class_efficiency"), drop = FALSE]

    # predict efficiency
    f_pred <- function(object, newdata) {

      if (inherits(object, "glm")) {
        predict(object, newdata = newdata, type = type)
      } else {
        predict(object, newdata = newdata, type = type)[, "efficient"]
      }

    }

    train_data  <- as.data.frame(train_data)
    target_data <- as.data.frame(target_data)

    bg_n <- if (!is.null(importance_method[["bg_n"]])) importance_method[["bg_n"]] else 200

    #
    shap_model <- kernelshap::kernelshap(
      object = final_model,
      X = target_data,
      bg_X = train_data,
      pred_fun = f_pred,
      bg_n = bg_n
    )

    # global importance = mean |SHAP| per variable
    imp <- t(data.frame(
      importance = colMeans(abs(shap_model$S), na.rm = TRUE)
    ))

    # Normalize for importance relative
    sum_imp <- sum(imp)
    imp_norm <- imp / sum_imp
    importance <- imp_norm

  } else if (importance_method[["name"]] == "PI") {

    labels_DEA <- reference_data$.outcome[1:nrow(target_data)]
    dataset_chosen <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]

    # predict efficiency
    f_pred <- function(object, newdata) {

      if (inherits(object, "glm")) {
        predict(object, newdata = newdata, type = type)
      } else {
        predict(object, newdata = newdata, type = type)[, "efficient"]
      }

    }

    # 4) Construir el Predictor de iml
    pred_obj <- iml::Predictor$new(
      model = final_model,
      data = dataset_chosen,
      y = labels_DEA,
      predict.function = f_pred,
      type  = "classification"
    )

    # Loss Function by AUC
    if (inherits(final_model, "glm")) {

      loss_auc <- function(truth, estimate) {

        # change to 0/1
        y_bin <- as.numeric(truth == levels(truth)[2])

        if (length(unique(y_bin)) < 2) return(NA_real_)

        1 - as.numeric(pROC::auc(response = y_bin, predictor = estimate))

      }

    } else {

      loss_auc <- function(truth, estimate) {

        # change to 0/1
        y_bin <- as.numeric(truth == levels(truth)[1])

        if (length(unique(y_bin)) < 2) return(NA_real_)

        1 - as.numeric(pROC::auc(response = y_bin, predictor = estimate))

      }

    }

    # 6) Permutation Importance (repite permutaciones para estabilidad)
    n_rep <- if (!is.null(importance_method[["n.repetitions"]])) importance_method[["n.repetitions"]] else 5

    fi <- iml::FeatureImp$new(
      predictor = pred_obj,
      loss = loss_auc,
      compare = "difference",      # caída de rendimiento vs. modelo completo
      n.repetitions = n_rep # sube si quieres más estabilidad
    )

    order_names <- names(dataset_chosen)
    idx <- match(order_names, fi$results$feature)

    imp <- fi$results[idx, "importance"]

    imp <- as.data.frame(t(imp))

    names(imp) <- names(dataset_chosen)

    # Normalize
    sum_imp <- sum(imp)
    imp_norm <- imp / sum_imp
    importance <- imp_norm
  }

  result_importance <- importance

  # messages
  if(inherits(final_model, "glm")) {
    name_ML_method <- "glm"
  } else {
    name_ML_method <- final_model[["method"]]
  }

  message(paste0("XAI method: ", importance_method[["name"]], " | Classification model: '", name_ML_method,"'"))
  message(paste("Inputs importance: ", round(sum(result_importance[1:x_len]), 5)))
  message(paste("Outputs importance: ", round(sum(result_importance[(x_len+1):(x_len + y_len)]),5)))

  return(result_importance)
}

#' @title Local feature importance for efficiency classifiers
#'
#' @description
#' Computes **local feature importance** (i.e., per-observation explanations) for a
#' fitted classification model that separates Pareto-efficient DMUs, using one of
#' three XAI backends:
#' \itemize{
#'   \item \code{"SA"} — Local Sensitivity Analysis via \pkg{rminer}.
#'   \item \code{"SHAP"} — Model-agnostic SHAP approximations via \pkg{kernelshap}.
#'   \item \code{"LIME"} — Model-agnostic LIME approximations via \pkg{lime}
#' }
#' You can compute local importance on specific DMUs provided in \code{explain_data},
#' using \code{reference_data} as the reference distribution or training background.
#' For each target DMU, importances are returned normalized to sum to 1 **within that DMU**.
#'
#' @param final_model A fitted model. If it is a base-\code{glm} binomial, probabilities
#'   are obtained with \code{type = "response"}; otherwise the function expects
#'   \code{predict(type = "prob")} with a column named \code{"efficient"}.
#' @param x Integer or character vector with the columns used as **inputs** (predictors).
#' @param y Integer or character vector with the columns used as **outputs** (targets used
#'   to define \code{class_efficiency} in training; not included in \code{X} when explaining).
#' @param explain_data A \code{data.frame} with the original observed DMUs (or new DMUs)
#'   on which **local** explanations are computed.
#' @param reference_data A \code{data.frame} with the training dataset used to fit
#'   the optimal model returned by \code{PEAXAI_fitting()}. Depending on the class-
#'   balancing procedure, this dataset may contain only real observed DMUs or both
#'   real and synthetic DMUs. It defines the reference distribution for perturbing features.
#' @param importance_method A named list (or data.frame-like) with the backend and its args:
#'   \describe{
#'     \item{\code{name}}{One of \code{"SA"}, \code{"SHAP"}, \code{"LIME"}.}
#'     \item{\code{method}}{(SA) One of \code{"1D-SA"}, \code{"sens"}, \code{"DSA"}, \code{"MSA"}, \code{"CSA"}, \code{"GSA"}.}
#'     \item{\code{measures}}{(SA) e.g. \code{"AAD"}, \code{"gradient"}, \code{"variance"}, \code{"range"}.}
#'     \item{\code{levels}}{(SA) Discretization levels used by \code{rminer::Importance}.}
#'     \item{\code{baseline}}{(SA) Baseline value for SA, if applicable. If omitted, the DMU itself is used as baseline.}
#'     \item{\code{bg_n}}{(SHAP) Background sample size for \code{kernelshap::kernelshap} (default: 200).}
#'     \item{\code{n_permutations}}{(LIME) Number of permutations per observation (default: 5000).}
#'     \item{\code{feature_select}}{(LIME) Feature selection method, e.g., \code{"auto"}, \code{"none"}, \code{"forward_selection"} (default: \code{"auto"}).}
#'     \item{\code{bin_continuous}}{(LIME) Logical indicating if continuous features should be binned (default: \code{TRUE}).}
#'     \item{\code{n_bins}}{(LIME) Number of bins for continuous features (default: 4).}
#'   }
#'
#' @param seed  Integer. Seed for reproducibility.
#'
#' @details
#' Local explanations are produced for the predicted probability \eqn{\hat{p}_i = P(\text{efficient}\mid x_i)}.
#' For \code{glm} models, the positive class is assumed to be the **second level**
#' (\code{"efficient"}) and probabilities are extracted with \code{type = "response"}.
#' For other models (e.g., \pkg{caret}), \code{predict(type = "prob")[, "efficient"]} is used.
#'
#' The meaning of “local importance” depends on the backend:
#' \itemize{
#'   \item \code{"SHAP"} returns per-observation SHAP attributions; importances are computed
#'   from attribution magnitudes (e.g., \code{abs(SHAP)}) and normalized to sum to 1 per DMU.
#'   \item \code{"LIME"} fits a local surrogate linear model around the DMU using
#'   permutations from the reference distribution; the absolute weights of the local
#'   model are used as importance scores, normalized per DMU.
#'   \item \code{"SA"} computes local sensitivity of \eqn{\hat{p}_i} to each feature around
#'   the DMU (using the chosen SA method/measure) and normalizes sensitivities per DMU.
#' }
#'
#' @return A numeric matrix (or data.frame) with one row per target observation and one
#'   column per predictor. Each row is a normalized importance profile whose values sum to 1.
#'   Column names match the predictor columns in \code{x}.
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   x <- 1:4
#'   y <- 5
#'   RTS <- "vrs"
#'   imbalance_rate <- NULL
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'       weights = "dinamic"
#'     )
#'   )
#'
#'   metric_priority <- c("Balanced_Accuracy", "ROC_AUC")
#'
#'   models <- PEAXAI_fitting(
#'     data = data, x = x, y = y, RTS = RTS,
#'     imbalance_rate = imbalance_rate,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = metric_priority,
#'     seed = 1,
#'     verbose = FALSE
#'   )
#'
#'   final_model <- models[["best_model_fit"]][["glm"]]
#'
#'   importance_method <- list(
#'     name = "SHAP",
#'     nsim = 200
#'    )
#'
#'   imp_local <- PEAXAI_local_importance(
#'     final_model = final_model,
#'     x = x, y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   head(imp_local)
#' }
#'
#' @seealso \code{\link[kernelshap]{kernelshap}},
#'   \code{\link[rminer]{Importance}},
#'   \code{\link[lime]{lime}}
#'
#' @importFrom stats predict glm
#' @importFrom kernelshap kernelshap
#' @importFrom rminer Importance
#' @importFrom pROC auc
#' @importFrom lime lime
#'
#' @export

PEAXAI_local_importance <- function(
    final_model, x, y, explain_data, reference_data,
    importance_method, seed = 314
) {

  # reproducibility
  set.seed(seed)

  # as data.frame
  explain_data <- as.data.frame(explain_data)
  reference_data <- as.data.frame(reference_data)

  # ----------------------------------------------------------------------------
  # detecting importance variables ---------------------------------------------
  # ----------------------------------------------------------------------------
  # type of prediction
  if (inherits(final_model, "glm")) {
    # type if is a glm model
    type <- "response"
  } else {
    # type if is a ML model
    type <- "prob"
  }

  # format datasets
  format_dataset <- function(d, x_cols, y_cols) {
    if (".outcome" %in% names(d)) {
      names(d)[names(d) == ".outcome"] <- "class_efficiency"
      d <- d[, c(setdiff(names(d), "class_efficiency"), "class_efficiency")]
      return(d)
    } else {
      d <- d[, c(x_cols, y_cols)]
      return(d)
    }
  }

  target_data <- format_dataset(explain_data, x, y)
  train_data <- format_dataset(reference_data, x, y)

  x_len <- length(x)
  y_len <- length(y)

  # methods XAI
  if (importance_method[["name"]] == "SA") {

    # falta comprobar
    # importance with our model of Caret
    f_pred <- function(object, newdata) {

      if (inherits(object, "glm")) {
        predict(object, newdata = newdata, type = type)
      } else {
        predict(object, newdata = newdata, type = type)[, "efficient"]
      }

    }

    # Define methods and measures
    method <- if (!is.null(importance_method[["method"]])) importance_method[["method"]] else "1D-SA"
    measure <- if (!is.null(importance_method[["measures"]])) importance_method[["measures"]] else "variance"
    levels <- if (!is.null(importance_method[["levels"]])) importance_method[["levels"]] else 5

    # matrix of data without label
    dataset_chosen <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]

    imp_list <- lapply(seq_len(nrow(dataset_chosen)), function(i) {
      b_i <- dataset_chosen[i, , drop = FALSE]

      rminer::Importance(
        M = final_model,
        data = dataset_chosen,
        RealL = levels,
        method = method,
        measure = measure,
        baseline = b_i,
        responses = TRUE,
        PRED = f_pred
      )[["imp"]]
    })

    imp_mat <- do.call(rbind, imp_list)

    imp <- as.data.frame(imp_mat)

    names(imp) <- names(dataset_chosen)

    # Normalize
    importance <- imp

  } else if (importance_method[["name"]] == "SHAP") {

    # matrix of data without label
    target_data <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]
    train_data <- train_data[, setdiff(names(train_data), "class_efficiency"), drop = FALSE]

    # predict efficiency
    f_pred <- function(object, newdata) {

      if (inherits(object, "glm")) {
        predict(object, newdata = newdata, type = type)
      } else {
        predict(object, newdata = newdata, type = type)[, "efficient"]
      }

    }

    bg_n <- if (!is.null(importance_method[["bg_n"]])) importance_method[["bg_n"]] else 200

    #
    shap_model <- kernelshap::kernelshap(
      object = final_model,
      X = target_data,
      bg_X = train_data,
      pred_fun = f_pred,
      bg_n = bg_n
    )

    # local importance = SHAP per variable
    imp <- data.frame(
      importance = shap_model$S
    )
    names(imp) <- names(train_data)

    # Normalize for relative importance per observation
    imp_abs <- abs(imp)
    row_totals <- rowSums(imp_abs, na.rm = TRUE)
    row_totals[row_totals == 0] <- 1 # prevent division by zero

    importance_abs <- imp_abs / row_totals

    # Reimpose original sign
    importance <- importance_abs * sign(imp)

  } else if (importance_method[["name"]] == "LIME") {

    # matrix of data without label
    target_data <- target_data[, setdiff(names(target_data), "class_efficiency"), drop = FALSE]
    train_data <- train_data[, setdiff(names(train_data), "class_efficiency"), drop = FALSE]

    train_data  <- as.data.frame(train_data)
    target_data <- as.data.frame(target_data)

    # Determine the correct label name expected by lime
    if (inherits(final_model, "glm")) {
      # lime's default method for glm binomial often names the target "1"
      explain_label <- "1"
    } else {
      explain_label <- "efficient"
    }

    # LIME parameters from importance_method (with defaults)
    bin_continuous <- if (!is.null(importance_method[["bin_continuous"]])) importance_method[["bin_continuous"]] else TRUE
    n_bins         <- if (!is.null(importance_method[["n_bins"]]))         importance_method[["n_bins"]]         else 4
    n_permutations <- if (!is.null(importance_method[["n_permutations"]])) importance_method[["n_permutations"]] else 5000
    feature_select <- if (!is.null(importance_method[["feature_select"]])) importance_method[["feature_select"]] else "auto"

    # create explainer
    explainer <- lime::lime(
      x = train_data,
      model = final_model,
      bin_continuous = bin_continuous,
      n_bins = n_bins
    )

    # run LIME explanation
    lime_model <- lime::explain(
      x = target_data,
      explainer = explainer,
      n_features = ncol(target_data),
      labels = explain_label,
      n_permutations = n_permutations,
      feature_select = feature_select
    )

    # Convert long format to wide matrix (rows = cases, cols = features)
    imp_matrix <- tapply(
      # abs(lime_model$feature_weight),
      lime_model$feature_weight,
      list(as.character(lime_model$case), as.character(lime_model$feature)),
      FUN = sum, na.rm = TRUE
    )

    # Ensure cases and features are ordered correctly to match target_data
    cases_order <- as.character(rownames(target_data))
    if (is.null(cases_order)) cases_order <- as.character(1:nrow(target_data))

    imp <- as.data.frame(imp_matrix[cases_order, names(target_data), drop = FALSE])

    # Replace NAs with 0 (for features not selected by LIME, though n_features = ncol)
    imp[is.na(imp)] <- 0

    # Normalize for relative importance per observation
    imp_abs <- abs(imp)
    row_totals <- rowSums(imp_abs, na.rm = TRUE)
    row_totals[row_totals == 0] <- 1 # prevent division by zero

    importance_abs <- imp_abs / row_totals

    # Reimpose original sign
    importance <- importance_abs * sign(imp)

  }

  result_importance <- importance

  # # messages
  # if(inherits(final_model, "glm")) {
  #   name_ML_method <- "glm"
  # } else {
  #   name_ML_method <- final_model[["method"]]
  # }

  # message(paste0("XAI method: ", importance_method[["name"]], " | Classification model: '", name_ML_method,"'"))
  # message(paste("Inputs importance: ", round(sum(colMeans(result_importance)[1:x_len]), 5)))
  # message(paste("Outputs importance: ", round(sum(colMeans(result_importance)[(x_len+1):(x_len + y_len)]),5)))

  return(result_importance)
}

#' @title Create New SMOTE Units to Balance Data combinations of m + s
#'
#' @description This function creates new DMUs to address data imbalances.
#' If the majority class is efficient, it generates new inefficient DMUs by worsering the observed units.
#' Conversely, if the majority class is inefficient, it projects inefficient DMUs to the frontier.
#' Finally, a random selection if performed to keep a proportion of 0.65 for the majority class and 0.35 for the minority class.
#'
#' @param data A \code{list} of \code{data.frames}, where each element represents a dataset with labeled data.
#' @param facets A \code{list} where each element represents a subgroup containing index combinations that generate efficient units.
#' @param x Column indexes of the input variables in the \code{data}.
#' @param y Column indexes of the output variables in the \code{data}.
#' @param z_numeric Column indexes of the continuous environment variables (z) in the \code{data}.
#' @param z_factor Column indexes of the factor environment variables (z) in the \code{data}.
#' @param RTS Text string or number defining the underlying DEA technology /
#'   returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
#'     \item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
#'   }
#' @param balance_data A numeric vector indicating the different levels of balance required (e.g., c(0.1, 0.45, 0.6)).
#' @param bandwidth the bandwidth parameters for the unconditional kernel density estimator used in the conditional DEA framework. It is typically obtained using \code{\link[np]{npudensbw}} and supports mixed data types, including continuous variables and discrete unordered or ordered factors. Bandwidths can be selected using normal reference rules, likelihood cross-validation, or least-squares cross-validation following Li and Racine (2003). If \code{NULL}, the bandwidth is estimated internally.
#' @param seed  Integer. Seed for reproducibility.
#' @param alpha This allow to choose the size of the Confidence Intervals computed. By defaulta alpha = FALSE. In this case no confidence interval are computed.

#'
#' @importFrom dplyr anti_join
#' @importFrom stats quantile
#'
#' @return A \code{list} where each element corresponds to a balance level, containing a single \code{data.frame}
#' with the real and synthetic DMUs, correctly labeled.

get_SMOTE_DMUs <- function(
    data, facets, x, y, z_numeric = NULL, z_factor = NULL,
    RTS = "vrs", alpha = FALSE, balance_data = NULL, bandwidth = NULL, seed) {
  # save a copy
  copy_data <- data

  # # Ensure column indices are character names to avoid indexing issues in new subsets
  # if (is.numeric(x)) x <- names(data)[x]
  # if (is.numeric(y)) y <- names(data)[y]
  # if (!is.null(z_numeric) && is.numeric(z_numeric)) z_numeric <- names(data)[z_numeric]
  # if (!is.null(z_factor) && is.numeric(z_factor)) z_factor <- names(data)[z_factor]

  save_all_datasets_balanced <- vector("list", length(balance_data))
  names(save_all_datasets_balanced) <- as.character(balance_data)

  # reproducibility.
  set.seed(seed)

  # case 1 DMU is a facet
  facet_length <- length(facets)

  # we need to determine, for each balance level, the number of synthetic DMUs to create
  for (balance in balance_data) {
    save_dataset <- as.data.frame(matrix(
      data = NA,
      ncol = ncol(copy_data),
      nrow = 0
    ))

    names(save_dataset) <- names(copy_data)

    # check if it is possible to balance
    if (nrow(facets) == 0 | facet_length == 1) {
      warning("No facets found; could not apply class balancing.")

      save_dataset <- rbind(save_dataset, data)

      # join real data + SMOTE
      new_data_completed <- save_dataset

      # save
      save_all_datasets_balanced[[as.character(balance)]] <- new_data_completed
      next
    }

    # --------------------------------------------------------------------------
    # determinate number of efficient and not efficient to create --------------
    # --------------------------------------------------------------------------
    # information balancing process
    message(paste0("Balancing: ", balance))

    # determinate numbre of efficient and ineeficient units
    n_real_eff <- nrow(data[data$class_efficiency == "efficient", ])
    n_real_ineff <- nrow(data[data$class_efficiency == "not_efficient", ])

    prop_real <- n_real_eff / nrow(data)

    n_new_eff <- 0

    n_new_ineff <- 0

    # proportion of efficients
    prop <- prop_real

    sense_balance <- NULL

    # determinate the way to balance, create efficient or not efficient
    if (prop < balance) {
      # need to create efficient units
      sense_balance <- "efficient"

      # in each itaretion we create these DMUs
      add_eff <- 1
      add_not_eff <- 0
    } else {
      # need to create not efficient units
      sense_balance <- "not_efficient"

      # in each itaretion we create these DMUs
      add_eff <- 0
      add_not_eff <- 1
    }

    # determinate how many DMUs create PROPORTION
    eff_level <- balance

    test_n_eff <- n_real_eff
    test_n_ineff <- n_real_ineff

    if (sense_balance == "not_efficient") {
      while (prop > eff_level) {
        test_n_ineff <- test_n_ineff + add_not_eff

        prop <- test_n_eff / (test_n_eff + test_n_ineff)
      }
    } else {
      while (prop < eff_level) {
        test_n_eff <- test_n_eff + add_eff

        prop <- test_n_eff / (test_n_eff + test_n_ineff)
      }
    }

    # it is necessary to create create_eff units
    create_eff <- test_n_eff - n_real_eff

    # it is necessary to create create_ineff units
    create_ineff <- test_n_ineff - n_real_ineff

    # balance perfect, next
    if (create_eff == 0 & create_ineff == 0) {
      message(paste("Balance perfect"))

      save_dataset <- rbind(save_dataset, data)

      # join real data + SMOTE
      new_data_completed <- save_dataset

      # save
      save_all_datasets_balanced[[as.character(balance)]] <- new_data_completed

      next
    }

    # ============================================ #
    # get index to create efficient synthetic DMUs #
    # ============================================ #
    data_eff <- data[data$class_efficiency == "efficient", ]

    # real efficient combination
    idx <- facets
    n_idx <- 1:nrow(idx)

    # number of efficient DMUs
    n_eff <- nrow(data_eff)

    # create units
    # lambda

    # proportion importance
    len <- ncol(facets)

    prop_imp <- 1 / len

    lambda <- rep(prop_imp, ncol(facets))

    n_comb <- nrow(data_eff)

    iter <- 0
    # print(sense_balance)
    if (sense_balance == "not_efficient") {
      # proportion importance
      len <- ncol(facets)

      prop_imp <- 1 / len

      lambda <- rep(prop_imp, ncol(facets))

      # number of not efficient units to create, more than it is necessary
      new_create_ineff <- 4 * create_ineff
      iter_ineff <- 0
      while (nrow(save_dataset) < new_create_ineff) {
        iter_ineff <- iter_ineff + 1
        # Ruptura de seguridad para evitar bucles infinitos
        if (iter_ineff > new_create_ineff * 10) {
          warning("Could not generate enough inefficient DMUs. Stopping early.")
          break
        }

        idx_eff <- which(data$class_efficiency == "efficient")

        # first, select a random index and combination
        # number of dimensions
        n_combinations <- NROW(facets)
        length_facets <- NCOL(facets)

        ctrl_facet <- TRUE

        iter_0 <- 0
        less_facet <- 0
        while (ctrl_facet == TRUE) {
          iter_0 <- iter_0 + 1
          # FIX: Evitar bucle infinito cuando es imposible encontrar una combinación que no sea una faceta.
          # Se sustituyó la llamada a browser() por una ruptura segura tras 30 iteraciones.
          if (iter_0 >= 30) {
            ctrl_facet <- FALSE
          }

          random_convex <- sample(idx_eff, size = length_facets, replace = FALSE)

          check <- random_convex[order(random_convex)]
          ref <- facets[1, ]
          ref <- ref[order(ref)]

          if (nrow(facets) == 1 & all(random_convex[order(random_convex)] == ref[order(facets[1, ])]) & length_facets - less_facet > 2) {
            less_facet <- less_facet + 1

            random_convex <- sample(idx_eff, size = length_facets - less_facet, replace = FALSE)

            # new
            # proportion importance
            len <- length_facets - less_facet

            prop_imp <- 1 / len

            lambda <- rep(prop_imp, length_facets - less_facet)

            # brake search
            ctrl_facet <- FALSE
          }

          if (nrow(facets) == 1 & all(random_convex[order(random_convex)] == ref[order(facets[1, 1:(length_facets - less_facet)])]) & length_facets - less_facet == 2) {
            ctrl_facet <- FALSE
          }

          # random_convex <- facets[random_convex,]
          next_sample <- FALSE

          if (length(random_convex) >= length(facets)) {
            # check not on facet
            for (facet_i in 1:nrow(facets)) {
              check <- random_convex[order(random_convex)]
              ref <- facets[facet_i, ]
              ref <- ref[order(ref)]

              if (all(check == ref)) {
                next_sample <- TRUE
                break
              }
            }
          }

          if (next_sample == TRUE) {
            next
          } else {
            ctrl_facet <- FALSE
          }
        } # end while generate random sample

        if (is.null(z_numeric)) {
          selection <- data[unlist(as.vector(random_convex)), c(x, y)]
        } else {
          selection <- data[unlist(as.vector(random_convex)), c(x, y, z_numeric)]
        }

        # second, determine random weights by DMU
        # process to generate lambda
        generate_lambda <- runif(nrow(selection), min = 0.01, max = 0.99)

        normalize_lambda <- generate_lambda / sum(generate_lambda)

        # third, generate the synthetic unit
        new_unit <- colSums(selection * normalize_lambda)
        new_unit <- as.data.frame(t(new_unit))

        # paste z_variables
        if (!is.null(z_factor)) {
          # FIX: Forzar que current_z_factor sea siempre de una sola fila mediante [1, , drop = FALSE].
          # Esto previene que cbind recicle la unidad y genere filas duplicadas si data tiene varios niveles en z_factor.
          current_z_factor <- unique(data[, z_factor, drop = FALSE])[1, , drop = FALSE]
          new_unit <- cbind(new_unit, current_z_factor)
        }

        # check new_unit
        DEA_B <- dea.add(
          X = as.matrix(new_unit[, x]),
          Y = as.matrix(new_unit[, y]),
          XREF = as.matrix(data[data$class_efficiency == "efficient", x]),
          YREF = as.matrix(data[data$class_efficiency == "efficient", y]),
          RTS = RTS
        )[["sum"]]

        # assing efficiency
        label <- ifelse(round(DEA_B, 4) <= 0.0001, "efficient", "not_efficient")

        # FIX: Cuando se requieren DMUs ineficientes en modelos de una sola faceta (single-facet),
        # las combinaciones convexas siempre resultan en unidades eficientes (causando un bucle infinito).
        # Solución: "worsening" (empeorar) artificialmente la unidad multiplicando sus inputs (x)
        # por un factor > 1 y sus outputs (y) por un factor < 1, proyectándola al interior ineficiente.
        # if (label == "efficient") {
        #   new_unit[, x] <- new_unit[, x] * runif(length(x), 1.05, 1.15)
        #   new_unit[, y] <- new_unit[, y] * runif(length(y), 0.85, 0.95)
        #
        #   DEA_B <- dea.add(
        #     X = as.matrix(new_unit[, x]),
        #     Y = as.matrix(new_unit[, y]),
        #     XREF = as.matrix(data[data$class_efficiency == "efficient", x]),
        #     YREF = as.matrix(data[data$class_efficiency == "efficient", y]),
        #     RTS = RTS
        #   )[["sum"]]
        #
        #   label <- ifelse(round(DEA_B, 4) <= 0.0001, "efficient", "not_efficient")
        # }

        # save if the DMU is not_efficient
        if (label == "not_efficient") {
          new_unit$class_efficiency <- "not_efficient"
          new_unit$score <- round(DEA_B, 4)
          save_dataset <- rbind(save_dataset, new_unit)
        } # end check
      } # end while

      # order by score in innefficieny DEA
      # make quiantiles
      q_innef <- quantile(save_dataset$score)

      # group_by
      quantiles <- cut(
        save_dataset$score,
        breaks = q_innef,
        include.lowest = TRUE,
        labels = c("Q1", "Q2", "Q3", "Q4")
      )

      # choose the same sample by quantile
      need_by_quantile <- create_ineff / 4

      copy_save_dataset <- save_dataset[, -length(save_dataset)]
      save_dataset <- save_dataset[0, ]

      # sample of each quantile
      for (quantile_i in levels(quantiles)) {
        q_i <- copy_save_dataset[quantiles == quantile_i, ]

        q_sample <- sample(1:nrow(q_i), size = need_by_quantile, replace = FALSE)

        save_dataset <- rbind(save_dataset, copy_save_dataset[q_sample, ])
      }
    } else {
      # first, populate the middle point to ensure that all facets are populate
      results_convx <- t(apply(facets, 1, function(indices) {
        # select row
        if (is.null(z_numeric)) {
          seleccion <- data[unlist(as.vector(indices)), c(x, y)]
        } else {
          seleccion <- data[unlist(as.vector(indices)), c(x, y, z_numeric)]
        }

        # calculate
        colSums(seleccion * lambda)
      }))

      results_convx <- as.data.frame(results_convx)

      # add factor
      if (!is.null(z_factor)) {
        # FIX: Selección segura de la primera fila única para prevenir duplicación accidental de DMUs con cbind.
        current_z_factor <- unique(data[, z_factor, drop = FALSE])[1, , drop = FALSE]
        results_convx <- cbind(results_convx, current_z_factor)
      }

      # too much efficient
      if (nrow(facets) > create_eff) {
        # select the index to save
        idx_save <- sample(nrow(facets), size = create_eff, replace = FALSE)

        new_unit <- results_convx[idx_save, ]

        # save DMUs selected
        new_unit$class_efficiency <- "efficient"
        save_dataset <- rbind(save_dataset, new_unit)
      } else {
        # save the previous SMOTE units generated
        results_convx$class_efficiency <- "efficient"
        save_dataset <- rbind(save_dataset, results_convx)
        bro_stop <- 0
        # if not too much, is it necessary to create more SMOTE DMUs
        iter_eff <- 0
        while (nrow(save_dataset) < create_eff) {

          iter_eff <- iter_eff + 1

          # FIX: Ruptura de seguridad si el modelo no puede alcanzar la proporción de unidades eficientes
          # debido a problemas de tolerancia y decimales en dea.add (para evitar bucle infinito).
          if (iter_eff > create_eff * 10) {

            warning("Could not generate enough efficient DMUs. Stopping early.")
            break
          }

          # first, select a random index and combination
          idx_save <- sample(nrow(facets), size = 1)
          dmus_by_facet <- facets[idx_save, ]

          selection <- data[unlist(as.vector(dmus_by_facet)), c(x, y, z_numeric)]

          # second, determine random weights by DMU
          # process to generate lambda
          generate_lambda <- runif(ncol(facets), min = 0.01, max = 0.99)
          normalize_lambda <- generate_lambda / sum(generate_lambda)

          # third, generate the synthetic unit
          new_unit <- colSums(selection * normalize_lambda)
          new_unit <- as.data.frame(t(new_unit))

          # paste z_factor
          if (!is.null(z_factor)) {
            # FIX: Selección segura de una única fila para variables categóricas (evita errores cbind de filas múltiples)
            current_z_factor_inner <- unique(data[, z_factor, drop = FALSE])[1, , drop = FALSE]
            new_unit <- cbind(new_unit, current_z_factor_inner)
          }

#           # check new_unit
#           DEA_B <- dea.add(
#             X = as.matrix(new_unit[, x]),
#             Y = as.matrix(new_unit[, y]),
#             XREF = as.matrix(data[data$class_efficiency == "efficient", x]),
#             YREF = as.matrix(data[data$class_efficiency == "efficient", y]),
#             RTS = RTS
#           )[["sum"]]
# #
#           if (DEA_B > 0.0001) {
#             print("DEA_B solution")
#             print(DEA_B)
#             browser()
#           }
#
#           # print(paste("Iter:", iter_eff, " | DEA_B:", DEA_B))
# #
#           # assing efficiency
#           label <- ifelse(round(DEA_B, 4) <= 0.0001, "efficient", "not_efficient")
          # print(paste("Label assigned:", label))
#
          # save if the DMU is not_efficient
          # if (label == "efficient") {
            new_unit$class_efficiency <- "efficient"
            save_dataset <- rbind(save_dataset, new_unit)
          # } # end check

        } # end while search
      } # end necessary to populate more
    } # end generating SMOTE

    # join real data + SMOTE
    new_data_completed <- rbind(data, save_dataset)

    # save
    save_all_datasets_balanced[[as.character(balance)]] <- new_data_completed
  } # end balance_data loop

  return(save_all_datasets_balanced)
}

#' @title Validate PEAXAI_fitting input parameters
#'
#' @description
#' Validates arguments for PEAXAI fitting/projection routines. It checks container
#' types, index ranges, non-overlapping input/output indices, admissible returns-to-scale
#' codes, resampling configuration, and basic probability ranges.
#'
#' @param data A \code{data.frame} or \code{matrix} with the variables used in the model.
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param RTS Returns-to-scale specification. Either a numeric code in
#'   \code{c(0,1,2,3,4,5)} or a character in
#'   \code{c("fdh","vrs","drs","crs","irs","add")}, matching \code{Benchmarking::dea.add()}.
#' @param trControl A list with resampling/control settings, e.g.
#'   \code{list(method = "cv", number = 5)}. Currently, \code{method} must be
#'   \code{"cv"} and \code{number} must be numeric.
#' @param methods A \strong{named list} of learners and their options (e.g., tuning grids),
#'   e.g. \code{list("nnet" = list(tuneGrid = ...), "svmPoly" = list(tuneGrid = ...))}.
#'   Allowed learner keys are \code{c("nnet","svmPoly","glm")}.
#' @param metric_priority Character vector with one or more primary metrics
#'   (e.g., \code{"ROC-AUC"}, \code{"Accuracy"}). Must belong to the supported set.
#' @param hold_out Numeric proportion in (0,1) (used if applicable by downstream code).
#' @param imbalance_rate Optional numeric vector with class proportions or target imbalance
#'   rates. All values must lie strictly in \code{(0,1)} if provided.
#' @param verbose Logical; whether to print additional diagnostics (validated as scalar logical).
#'
#' @details
#' Checks performed include:
#' \itemize{
#'   \item \strong{Data container}: \code{data} must be a non-empty \code{data.frame} or \code{matrix}.
#'   \item \strong{Index vectors}: \code{x} and \code{y} are numeric indices within \code{ncol(data)} and must not overlap.
#'   \item \strong{RTS}: must be one of \code{0:5} or \code{"fdh","vrs","drs","crs","irs","add"}.
#'   \item \strong{Resampling}: \code{trControl$method == "cv"} with numeric \code{trControl$number}.
#'   \item \strong{Methods}: names must include at least one of \code{"nnet","svmPoly","glm"}.
#'   \item \strong{Imbalance}: if provided, all entries of \code{imbalance_rate} are in \code{(0,1)}.
#' }
#'
#' @return Invisibly returns \code{NULL} on success; otherwise throws an error describing the first issue found.
#'
#' @seealso \code{\link[Benchmarking]{dea.add}} for RTS codes; \code{\link[caret]{trainControl}} for CV specs.
#'
#' @keywords internal
#' @noRd

validate_parametes_PEAXAI_fitting <- function(
    data, x, y, RTS, trControl, methods, metric_priority,
    hold_out, imbalance_rate, verbose
) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }
  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # RTS ------------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  RTS_available_num  <- 0:5
  RTS_available_char <- c("fdh","vrs","drs","crs","irs","add")

  if (is.numeric(RTS) && !RTS %in% RTS_available_num)
    stop("'RTS' not recognized.", call. = FALSE)
  if (is.character(RTS) && !RTS %in% RTS_available_char)
    stop("'RTS' not recognized.", call. = FALSE)

  # ----------------------------------------------------------------------------
  # trControl ------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (is.null(trControl) || (is.list(trControl) && length(trControl) == 0L)) {
    stop("'trControl' must be specified (e.g., list(method = 'cv', number = 5)).", call. = FALSE)
  }

  if (!is.character(trControl[["method"]]) ||
      length(trControl[["method"]]) != 1L ||
      !(trControl[["method"]] %in% c("cv", "repeatedcv"))) {
    stop("`trControl$method` must be 'cv' or 'repeatedcv'.", call. = FALSE)
  }

  if (trControl[["method"]] == "cv") {
    if (is.null(trControl[["number"]]) || !is.numeric(trControl[["number"]])) {
      stop("`trControl$number` must be numeric.", call. = FALSE)
    }
  } else if (trControl[["method"]] == "repeatedcv") {

    if (is.null(trControl[["number"]]) || is.null(trControl[["repeats"]]) || !is.numeric(trControl[["number"]]) || !is.numeric(trControl[["repeats"]])) {
      stop("`trControl$number` and `trControl$repeats` must be numeric.", call. = FALSE)
    }
  }


  # ----------------------------------------------------------------------------
  # Methods --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (is.null(methods) || !is.list(methods)) {
    stop("`methods` must be a named list.", call. = FALSE)
  }
  if (!any(names(methods) %in% c("nnet","svmPoly","glm", "rf", "svmRadial"))) {
    stop("`methods` must include at least one of: nnet, svmPoly, glm, rf or svmRadial.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Metric priority ------------------------------------------------------------
  # ----------------------------------------------------------------------------
  metric_priority_available <- c("Accuracy", "Kappa",
                                 "Recall", "Specificity",
                                 "Precision", "F1", "Balanced_Accuracy", "G_mean",
                                 "ROC_AUC", "PR_AUC")
  if (!any(metric_priority %in% metric_priority_available)) {
    stop("`metric_priority` must be in the supported set.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Hold out -------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.null(hold_out)) {
    if (!is.numeric(hold_out) || hold_out <= 0 || hold_out >= 1) {
      stop("`hold_out` must be numeric in (0,1).", call. = FALSE)
    }
  }

  # ----------------------------------------------------------------------------
  # Imbalance rate -------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.null(imbalance_rate)) {
    if (!is.numeric(imbalance_rate) || any(imbalance_rate <= 0 | imbalance_rate >= 1)) {
      stop("'imbalance_rate' values must be in (0,1).", call. = FALSE)
    }
  }

  # ----------------------------------------------------------------------------
  # verbose --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.logical(verbose) || length(verbose) != 1L) {
    stop("'verbose' must be TRUE or FALSE.", call. = FALSE)
  }

  invisible(NULL)
}

#' @title Validate input parameters for efficiency labeling (Additive DEA)
#'
#' @description
#' Checks that the arguments supplied to the efficiency labeling routine are
#' correctly specified. The function validates container types, index ranges,
#' non-overlapping input/output indices, returns-to-scale codes, and basic
#' dimensional consistency between \code{data} and \code{REF}.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables used
#'   for labeling. Rows correspond to DMUs.
#' @param REF Optional reference set with the same number of rows as \code{data}
#'   used to define the technology. Defaults to \code{data}. Must be a
#'   \code{data.frame} or \code{matrix}.
#' @param x Numeric vector of column indices in \code{data}/\code{REF}
#'   corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data}/\code{REF}
#'   corresponding to output variables.
#' @param RTS Returns-to-scale specification. Either a numeric code in
#'   \code{c(0,1,2,3,4,5)} or a character in
#'   \code{c("fdh","vrs","drs","crs","irs","add")}, matching the options in
#'   \code{Benchmarking::dea.add()}.
#'
#' @details
#' The following checks are performed:
#' \itemize{
#'   \item \strong{Containers}: \code{data} (and \code{REF}) must be non-empty
#'         \code{data.frame} or \code{matrix} objects with the same number of rows.
#'   \item \strong{Index vectors}: \code{x} and \code{y} must be numeric index
#'         vectors within \code{ncol(data)} and must not overlap.
#'   \item \strong{RTS}: must be either one of the numeric codes \code{0:5} or one of
#'         \code{"fdh","vrs","drs","crs","irs","add"} (as in \code{Benchmarking::dea.add}).
#' }
#' Any violation triggers \code{stop()} with a descriptive message.
#'
#' @return
#' Invisibly returns \code{NULL} on success. Otherwise, throws an error describing
#' the first problem encountered.
#'
#' @seealso
#' \code{\link[Benchmarking]{dea.add}} for RTS codes and the Additive DEA model.
#'
#' @keywords internal
#' @noRd

# check if parameters are well introduced
validate_parametes_label_efficiency <- function(data, REF = NULL, x, y, RTS) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # REF ------------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  # REF: si es NULL, usa data; si no, valida
  if (is.null(REF)) {
    REF <- data
  } else {
    if (!is.data.frame(REF) && !is.matrix(REF)) {
      stop("'REF' must be a data.frame or matrix.", call. = FALSE)
    }
    if (nrow(REF) == 0L || ncol(REF) == 0L) {
      stop("'REF' has zero rows or zero columns.", call. = FALSE)
    }
    if (nrow(REF) != nrow(data)) {
      stop("'REF' must have the same number of rows as 'data'.", call. = FALSE)
    }
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }
  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # RTS ------------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  RTS_available_num  <- 0:5
  RTS_available_char <- c("fdh","vrs","drs","crs","irs","add")
  if (is.numeric(RTS) && !RTS %in% RTS_available_num)
    stop("'RTS' not recognized.", call. = FALSE)
  if (is.character(RTS) && !RTS %in% RTS_available_char)
    stop("'RTS' not recognized.", call. = FALSE)

  invisible(NULL)
}

#' @title Validate input parameters for SMOTE preprocessing
#'
#' @description
#' Validates arguments used before applying SMOTE (or class-balancing) in the
#' PEAXAI pipeline. It checks container types, index ranges, non-overlapping
#' input/output indices, admissible returns-to-scale (RTS) codes (for downstream
#' DEA-related steps), and class-balance proportions.
#'
#' @param data A \code{data.frame} or \code{matrix} with the variables used in the model.
#'   Rows correspond to DMUs (observations).
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param RTS Returns-to-scale specification. Either a numeric code in
#'   \code{c(0,1,2,3,4,5)} or a character in
#'   \code{c("fdh","vrs","drs","crs","irs","add")}, matching \code{Benchmarking::dea.add()}.
#'   (Kept for consistency with downstream DEA steps; not specific to SMOTE.)
#' @param balance_data Optional numeric value or vector with target class balance
#'   proportions in the open interval \code{(0,1)}. For a single binary target, a
#'   scalar can represent the desired positive-class proportion (e.g., \code{0.5}).
#'
#' @details
#' Checks performed include:
#' \itemize{
#'   \item \strong{Data container}: \code{data} must be a non-empty \code{data.frame} or \code{matrix}.
#'   \item \strong{Index vectors}: \code{x} and \code{y} are numeric indices within \code{ncol(data)} and must not overlap.
#'   \item \strong{RTS}: must be one of \code{0:5} or \code{"fdh","vrs","drs","crs","irs","add"}.
#'   \item \strong{Balance}: if \code{balance_data} is provided, all entries must lie strictly in \code{(0,1)}.
#' }
#'
#' @return Invisibly returns \code{NULL} on success; otherwise throws an error describing the first issue found.
#'
#' @seealso \code{\link[Benchmarking]{dea.add}} for RTS codes (DEA step).
#'
#' @keywords internal
#' @noRd
#'

# check if parameters are well introduced
validate_parametes_SMOTE_data <- function(
  # SMOTE_data
  data, x, y, RTS, balance_data

) {

  # ----------------------------------------------------------------------------
  # Data --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }
  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # RTS ------------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  RTS_available_num  <- 1:5
  RTS_available_char <- c("vrs","drs","crs","irs","add")

  if (is.numeric(RTS) && !RTS %in% RTS_available_num) {
    stop("'RTS' not recognized.", call. = FALSE)
  }

  if (is.character(RTS) && !RTS %in% RTS_available_char) {
    stop("'RTS' not recognized.", call. = FALSE)
  }

  if (any(RTS == "0" || RTS == 0)) {
    stop("'fdh' in RTS' is not possible in SMOTE data", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Imbalance rate -------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.null(balance_data)) {
    if (!is.numeric(balance_data) || any(balance_data <= 0 | balance_data >= 1)) {
      stop("'balance_data' values must be in (0,1).", call. = FALSE)
    }
  }

  invisible(NULL)
}

#' @title Validate input parameters for PEAXAI global importance
#'
#' @description
#' Validates arguments prior to computing global feature importance in the PEAXAI
#' pipeline. It checks data containers, input/output index ranges, returns-to-scale
#' (RTS) codes used in downstream DEA steps, the trained model object, the origin
#' of background/target data, and the global importance method specification
#' using packages: \strong{rminer} (SA), \strong{fastshap} (SHAP), and \strong{iml} (PI).
#'
#' @param data A \code{data.frame} or \code{matrix} with the variables used in the model
#'   (rows = DMUs/observations).
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param final_model A trained \code{caret} model, i.e., an object of class \code{"train"}.
#' @param background Background data origin for global explanations: \code{"train"} or \code{"real"}.
#' @param target Target data origin to be explained: \code{"train"} or \code{"real"}.
#' @param importance_method A \code{list} specifying the global importance approach.
#'   Must include \code{name} in \code{c("SA","SHAP","PI")}. Accepted fields by method:
#'   \itemize{
#'     \item \strong{SA} (via \pkg{rminer}): \code{method} in
#'           \code{c("1D-SA","sens","DSA","MSA","CSA")};
#'           \code{measures} in \code{c("AAD","gradient","variance","range")};
#'           \code{levels} single integer >= 2;
#'           \code{baseline} either \code{"mean"}, \code{"media"} or a \code{data.frame}.
#'     \item \strong{SHAP} (via \pkg{fastshap}): \code{nsim} single positive integer.
#'     \item \strong{PI} (via \pkg{iml}): \code{n.repetitions} single positive integer.
#'   }
#'
#' @details
#' The function throws an informative error on the first detected violation. On success,
#' it returns \code{NULL} invisibly.
#'
#' @return Invisibly returns \code{NULL} on success.
#'
#' @seealso \code{\link[caret]{train}} for the \code{"train"} object;
#'   \code{\link[Benchmarking]{dea.add}} for RTS codes;
#'   \code{\link[rminer]{Importance}} for SA settings;
#'   \code{\link[fastshap]{explain}} for SHAP;
#'   \code{\link[iml]{FeatureImp}} for permutation importance (PI).
#'
#' @note For any ambiguity on argument semantics, please refer to the official
#'   documentation of \pkg{rminer} (SA), \pkg{fastshap} (SHAP), and \pkg{iml} (PI).
#'
#' @keywords internal
#' @noRd
#'

validate_parametes_PEAXAI_global_importance <- function(
    data, x, y, final_model, background, target, importance_method
) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }

  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Final model ----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(final_model) || is.null(final_model) || !inherits(final_model, "train")) {
    stop("'final_model' must be a caret::train() object (class 'train').", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # background and target ------------------------------------------------------
  # ----------------------------------------------------------------------------
  allowed_bt <- c("train","real")
  if (!is.character(background) || length(background) != 1L || !(background %in% allowed_bt)) {
    stop("'background' must be 'train' or 'real'.", call. = FALSE)
  }
  if (!is.character(target) || length(target) != 1L || !(target %in% allowed_bt)) {
    stop("'target' must be 'train' or 'real'.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Importance method ----------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(importance_method) || !is.list(importance_method)) {
    stop("'importance_method' must be a list.", call. = FALSE)
  }

  name <- importance_method[["name"]]

  if (!is.character(name) || length(name) != 1L || !(name %in% c("SA","SHAP","PI"))) {
    stop("'importance_method$name' must be one of: 'SA', 'SHAP', 'PI'.", call. = FALSE)
  }

  if (identical(name, "SA")) {
    # rminer
    if (!is.null(importance_method[["method"]]) &&
        !importance_method[["method"]] %in% c("1D-SA","sens","DSA","MSA","CSA")) {
      stop("'importance_method$method' (SA/rminer) must be one of: '1D-SA','sens','DSA','MSA','CSA'.", call. = FALSE)
    }
    if (!is.null(importance_method[["measures"]]) &&
        !all(importance_method[["measures"]] %in% c("AAD","gradient","variance","range"))) {
      stop("'importance_method$measures' (SA/rminer) must be in: 'AAD','gradient','variance','range'.", call. = FALSE)
    }
    if (!is.null(importance_method[["levels"]])) {
      lv <- importance_method[["levels"]]
      if (!is.numeric(lv) || length(lv) != 1L || lv < 2) {
        stop("'importance_method$levels' (SA/rminer) must be a single numeric >= 2.", call. = FALSE)
      }
    }
    if (!is.null(importance_method[["baseline"]])) {
      base <- importance_method[["baseline"]]
      base_ok <- (is.character(base) && base %in% c("mean","media")) || is.data.frame(base)
      if (!base_ok) {
        stop("'importance_method$baseline' (SA/rminer) must be 'mean', 'media', or a data.frame.", call. = FALSE)
      }
    }
  } else if (identical(name, "SHAP")) {
    # fastshap
    nsim <- importance_method[["nsim"]]
    if (is.null(nsim) || !is.numeric(nsim) || length(nsim) != 1L || nsim <= 0) {
      stop("'importance_method$nsim' (SHAP/fastshap) must be a positive integer.", call. = FALSE)
    }
  } else if (identical(name, "PI")) {
    # iml
    nrep <- importance_method[["n.repetitions"]]
    if (is.null(nrep) || !is.numeric(nrep) || length(nrep) != 1L || nrep <= 0) {
      stop("'importance_method$n.repetitions' (PI/iml) must be a positive integer.", call. = FALSE)
    }
  }

  invisible(NULL)
}

#' @title Validate input parameters for PEAXAI targets
#'
#' @description
#' Validates arguments prior to constructing target sets with \code{PEAXAI_targets}.
#' It checks data containers, input/output index ranges, the trained model object,
#' efficiency thresholds, and the directional vector specification (scope, baseline,
#' and relative importance), as well as grid/expansion and bound parameters.
#'
#' @param data A \code{data.frame} or \code{matrix} containing all variables (rows = DMUs).
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param final_model A trained \code{caret} model, i.e., an object of class \code{"train"}.
#' @param efficiency_thresholds Numeric vector of efficiency thresholds (each strictly in \code{(0,1)}).
#' @param directional_vector A \code{list} with fields:
#'   \itemize{
#'     \item \code{relative_importance}: numeric, non-negative weights for inputs; length \code{= length(x)}; not all zero.
#'     \item \code{scope}: \code{"global"} or \code{"local"} (currently only \code{"global"} supported).
#'     \item \code{baseline}: one of \code{"mean"}, \code{"median"}, \code{"self"}, \code{"ones"}.
#'   }
#' @param n_expand Numeric expansion factor (e.g., \code{0.5}); must be a single non-negative number.
#' @param n_grid Integer number of grid points used in search/refinement; must be a single integer >= 2.
#' @param max_y Numeric non-negative upper bound multiplier for outputs (domain-specific).
#' @param min_x Numeric non-negative lower bound multiplier for inputs (domain-specific).
#'
#' @details
#' The function throws an informative error on the first detected violation. On success,
#' it returns \code{NULL} invisibly. If in doubt about modelling choices behind
#' \code{relative_importance}/\code{scope}/\code{baseline}, consult your method section
#' and keep them consistent across experiments.
#'
#' @return Invisibly returns \code{NULL} on success.
#'
#' @seealso \code{\link[caret]{train}} for the \code{"train"} object.
#'
#' @keywords internal
#' @noRd
#'

validate_parametes_PEAXAI_targets <- function(
    data, x, y, final_model,
    efficiency_thresholds, directional_vector,
    n_expand, n_grid, max_y, min_x
) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x e y ----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }

  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # final model ----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(final_model) || is.null(final_model) || !inherits(final_model, "train")) {
    stop("'final_model' must be a caret::train() object (class 'train').", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Efficiency thresholds ------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(efficiency_thresholds) || is.null(efficiency_thresholds) ||
      !is.numeric(efficiency_thresholds) || length(efficiency_thresholds) < 1L) {
    stop("'efficiency_thresholds' must be a non-empty numeric vector.", call. = FALSE)
  }
  if (any(efficiency_thresholds <= 0 | efficiency_thresholds >= 1)) {
    stop("'efficiency_thresholds' values must lie strictly in (0,1).", call. = FALSE)
  }
  # # Optional monotonicity check:
  # if (is.unsorted(efficiency_thresholds, strictly = TRUE)) {
  #   stop("'efficiency_thresholds' must be strictly increasing.", call. = FALSE)
  # }

  # ----------------------------------------------------------------------------
  # Directional vector ---------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(directional_vector) || !is.list(directional_vector)) {
    stop("'directional_vector' must be a list.", call. = FALSE)
  }
  # relative_importance
  ri <- directional_vector[["relative_importance"]]

  if (!is.data.frame(ri)) {
    ri <- as.data.frame(t(ri))

    if (ncol(ri) == 1) {
      ri <- as.data.frame(t(ri))
    }
  }

  if (is.null(ri) || !is.data.frame(ri) || any(ri < 0) || length(ri) != length(c(x, y))) {

    stop("'directional_vector$relative_importance' must be a non-negative numeric vector of length length(c(x,y)).",
         call. = FALSE)
  }
  if (sum(ri) == 0) {
    stop("'directional_vector$relative_importance' cannot be the all-zero vector.", call. = FALSE)
  }
  if (abs(sum(ri) - 1) > 1e-8) {
    warning("'directional_vector$relative_importance' does not sum to 1; proceeding without normalization.", call. = FALSE)
  }
  # scope
  scope <- directional_vector[["scope"]]
  if (is.null(scope) || !is.character(scope) || length(scope) != 1L ||
      !(scope %in% c("global","local"))) {
    stop("'directional_vector$scope' must be 'global' or 'local'.", call. = FALSE)
  }
  if (identical(scope, "local")) {
    warning("Support for 'local' scope is not yet implemented; using 'global' semantics.", call. = FALSE)
  }
  # baseline
  baseline <- directional_vector[["baseline"]]
  if (is.null(baseline) || !is.character(baseline) || length(baseline) != 1L ||
      !(baseline %in% c("mean","median","self","ones"))) {
    stop("'directional_vector$baseline' must be one of: 'mean','median','self','ones'.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # n_expand and n_grid --------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.numeric(n_expand) || length(n_expand) != 1L || n_expand < 0) {
    stop("'n_expand' must be a single non-negative numeric value.", call. = FALSE)
  }
  if (!is.numeric(n_grid) || length(n_grid) != 1L || n_grid < 2 || n_grid != as.integer(n_grid)) {
    stop("'n_grid' must be a single integer >= 2.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # max_y and min_x ------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.numeric(max_y) || length(max_y) != 1L || max_y < 0) {
    stop("'max_y' must be a single non-negative numeric value.", call. = FALSE)
  }
  if (!is.numeric(min_x) || length(min_x) != 1L || min_x < 0) {
    stop("'min_x' must be a single non-negative numeric value.", call. = FALSE)
  }

  invisible(NULL)
}

#' @title Validate input parameters for PEAXAI ranking
#'
#' @description
#' Validates arguments prior to computing a PEAXAI-based ranking. It checks data
#' containers, input/output index ranges, the trained model object, efficiency
#' thresholds, the targets object consistency, and the ranking basis.
#'
#' @param data A \code{data.frame} or \code{matrix} with all variables (rows = DMUs).
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param final_model A trained \code{caret} model, i.e., an object of class \code{"train"}.
#' @param efficiency_thresholds Numeric vector of efficiency thresholds (each strictly in \code{(0,1)}),
#'   usually strictly increasing (e.g., \code{seq(0.75, 0.95, 0.1)}).
#' @param targets Target specification produced by \code{PEAXAI_targets()} (either a \code{data.frame}
#'   or a \code{list}). Minimal consistency checks are performed (see Details).
#' @param rank_basis Character scalar indicating the basis for ranking. Must be one of
#'   \code{"predicted"} or \code{"attainable"}.
#'
#' @details
#' \strong{Targets consistency:}
#' \itemize{
#'   \item If \code{targets} is a \code{data.frame}, it must be non-empty. If it has a column
#'         named \code{efficiency_threshold} (or \code{threshold}), all values must be a subset
#'         of \code{efficiency_thresholds}.
#'   \item If \code{targets} is a \code{list}, it must be non-empty. If it is named, and names
#'         \emph{look like} thresholds, they must be a subset of \code{efficiency_thresholds}.
#'         If unnamed, its length should reasonably match either \code{length(efficiency_thresholds)}
#'         (targets-per-threshold) or \code{nrow(data)} (targets-per-DMU). In ambiguous cases a
#'         warning is issued.
#' }
#' The function throws an error on the first violation and returns \code{NULL} invisibly otherwise.
#'
#' @return Invisibly returns \code{NULL} on success.
#'
#' @seealso \code{\link[caret]{train}} for the \code{"train"} object; \code{PEAXAI_targets()}.
#'
#' @keywords internal
#' @noRd
#'

validate_parametes_PEAXAI_ranking <- function(
    data, x, y, final_model,
    efficiency_thresholds, targets, rank_basis
) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }

  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Final model ----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(final_model) || is.null(final_model) || !inherits(final_model, "train")) {
    stop("'final_model' must be a caret::train() object (class 'train').", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Efficiency thresholds ------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (rank_basis == "attainable") {
    if (missing(efficiency_thresholds) || is.null(efficiency_thresholds) ||
        !is.numeric(efficiency_thresholds) || length(efficiency_thresholds) < 1L) {
      stop("'efficiency_thresholds' must be a non-empty numeric vector.", call. = FALSE)
    }
    if (any(efficiency_thresholds <= 0 | efficiency_thresholds >= 1)) {
      stop("'efficiency_thresholds' values must lie strictly in (0,1).", call. = FALSE)
    }
  }

  # ----------------------------------------------------------------------------
  # Targets --------------------------------------------------------------------
  # ----------------------------------------------------------------------------

  # if (missing(targets) || is.null(targets)) {
  #   stop("'targets' must be provided (result of PEAXAI_targets()).", call. = FALSE)
  # }

  if (!is.null(targets)) {

    if (!is.list(targets)) {
      stop("'targets' must be a list. It is more easy pass the object provided by 'PEAXAI_targets()'", call. = FALSE)
    }

    if (length(targets) == 0L) {
      stop("'targets' list is empty.", call. = FALSE)
    }

    for (eff_i in names(targets)) {
      if (!all(names(targets[[eff_i]][["counterfactual_dataset"]]) == names(data)[c(x,y)])) {
        stop("names in 'target' are not the same than in 'data'",  call. = FALSE)
      }
    }
  }
  # } else {
  #   # stop("'targets' must be a list. It more easy pass the object provided by 'PEAXAI_targets'", call. = FALSE)
  # }

  # ----------------------------------------------------------------------------
  # Rank basis -----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.character(rank_basis) || length(rank_basis) != 1L ||
      !(rank_basis %in% c("predicted","attainable"))) {
    stop("'rank_basis' must be 'predicted' or 'attainable'.", call. = FALSE)
  }

  if (rank_basis == "attainable") {
    if (is.null(targets) || !is.list(targets)) {
      stop(paste0("If 'rank_basis' is 'attainable', it is necesary to pas 'targets' argument, more easily provided by PEAXAI_tragets()"))
    }
  }

  invisible(NULL)
}

#' @title Validate input parameters for PEAXAI peer analysis
#'
#' @description
#' Validates arguments prior to computing PEAXAI peers. It checks data containers,
#' input/output index ranges, the trained model object, efficiency thresholds,
#' and the weighting configuration for peer distance.
#'
#' @param data A \code{data.frame} or \code{matrix} with all variables (rows = DMUs).
#' @param x Numeric vector of column indices in \code{data} corresponding to input variables.
#' @param y Numeric vector of column indices in \code{data} corresponding to output variables.
#' @param final_model A trained \code{caret} model, i.e., an object of class \code{"train"}.
#' @param efficiency_thresholds Numeric vector of efficiency thresholds (each strictly in \code{(0,1)}),
#'   typically strictly increasing (e.g., \code{seq(0.75, 0.95, 0.1)}).
#' @param weighted Logical scalar; if \code{TRUE}, peer distances are weighted by
#'   \code{relative_importance}. If \code{FALSE}, \code{relative_importance} is ignored.
#' @param relative_importance When \code{weighted = TRUE}, a one-row \code{data.frame} with
#'   non-negative weights for input variables. Column names should match \code{colnames(data)[x]}.
#'   If unnamed, the number of columns must equal \code{length(x)}. Not available per-unit.
#'
#' @details
#' The function throws an informative error on the first detected violation and
#' returns \code{NULL} invisibly on success.
#'
#' @return Invisibly returns \code{NULL} on success.
#'
#' @seealso \code{\link[caret]{train}} for the \code{"train"} object.
#'
#' @keywords internal
#' @noRd
#'

validate_parametes_PEAXAI_peer <- function(
    data, x, y, final_model,
    efficiency_thresholds, weighted = FALSE, relative_importance = NULL
) {

  # ----------------------------------------------------------------------------
  # Data -----------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.data.frame(data) && !is.matrix(data)) {
    stop("'data' must be a data.frame or matrix.", call. = FALSE)
  }
  if (nrow(data) == 0L || ncol(data) == 0L) {
    stop("'data' has zero rows or zero columns.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # x and y --------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!(is.numeric(x) && is.numeric(y))) {
    stop("'x' and 'y' must be numeric index vectors.", call. = FALSE)
  }

  if (any(x %in% y)) {
    stop("'x' and 'y' must not overlap.", call. = FALSE)
  }

  p <- ncol(data)
  if (any(x < 1 | x > p) || any(y < 1 | y > p)) {
    stop("'x' and/or 'y' are out of column range.", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Final mpdel ----------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(final_model) || is.null(final_model) || !inherits(final_model, "train")) {
    stop("'final_model' must be a caret::train() object (class 'train').", call. = FALSE)
  }

  # ----------------------------------------------------------------------------
  # Efficiency thresholds ------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (missing(efficiency_thresholds) || is.null(efficiency_thresholds) ||
      !is.numeric(efficiency_thresholds) || length(efficiency_thresholds) < 1L) {
    stop("'efficiency_thresholds' must be a non-empty numeric vector.", call. = FALSE)
  }
  if (any(efficiency_thresholds <= 0 | efficiency_thresholds >= 1)) {
    stop("'efficiency_thresholds' values must lie strictly in (0,1).", call. = FALSE)
  }
  # if (is.unsorted(efficiency_thresholds, strictly = TRUE)) {
  #   stop("'efficiency_thresholds' must be strictly increasing.", call. = FALSE)
  # }


  # ----------------------------------------------------------------------------
  # Weighted -------------------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (!is.logical(weighted) || length(weighted) != 1L) {
    stop("'weighted' must be a single logical (TRUE/FALSE).", call. = FALSE)
  }

  if (isTRUE(weighted)) {
    if (is.null(relative_importance) || !is.data.frame(relative_importance)) {
      stop("'relative_importance' must be a one-row data.frame when 'weighted = TRUE'.", call. = FALSE)
    }

    # --------------------------------------------------------------------------
    # Relative importance ------------------------------------------------------
    # --------------------------------------------------------------------------
    if (nrow(relative_importance) != 1L) {
      stop("'relative_importance' must have exactly one row (global weights only).", call. = FALSE)
    }

    # Names and alignment to inputs
    input_names <- colnames(as.data.frame(data))[c(x,y)]
    ri_names    <- colnames(relative_importance)

    if (!is.null(ri_names) && any(nzchar(ri_names))) {
      # Must be a subset and aligned (we'll reorder internally if needed)
      if (!all(ri_names %in% input_names)) {
        browser()
        stop("'relative_importance' column names must match input names colnames(data)[c(x,y)].", call. = FALSE)
      }
      # Ensure we have weights for all inputs
      if (!all(input_names %in% ri_names)) {
        stop("'relative_importance' must provide weights for all inputs in colnames(data)[c(x,y)].", call. = FALSE)
      }
    } else {
      # Unnamed: must match by position
      if (ncol(relative_importance) != length(c(x,y))) {
        stop("Unnamed 'relative_importance' must have ncol == length(c(x,y)).", call. = FALSE)
      }
      # # Assign input names for clarity downstream
      # colnames(relative_importance) <- input_names
    }

    # Numeric, non-negative, finite, not all zero
    vals <- as.numeric(relative_importance[1, , drop = TRUE])
    if (any(!is.finite(vals)) || any(vals < 0)) {
      stop("'relative_importance' weights must be finite and non-negative.", call. = FALSE)
    }
    if (all(vals == 0)) {
      stop("'relative_importance' cannot be the all-zero vector.", call. = FALSE)
    }

    # (Opcional) Aviso si no suma 1
    if (abs(sum(vals) - 1) > 1e-8) {
      warning("'relative_importance' does not sum to 1; proceeding without normalization.", call. = FALSE)
    }
  } else {
    # weighted == FALSE -> ignore relative_importance silently
    # (no-op)
  }

  invisible(NULL)
}



#' @title Prepare Data and Handle Errors
#'
#' @description This function arranges the data in the required format and displays some error messages.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables in the model.
#' @param x Column indexes of input variables in \code{data}.
#' @param y Column indexes of output variables in \code{data}.
#'
#' @importFrom caret trainControl
#'
#' @return It returns a \code{matrix} in the required format and displays some error messages.

preprocessing <- function (
    data, x, y
) {

  # x and y well / bad introduced

  cols <- 1:length(data)
  if (!(all(x %in% cols))) {
    stop("x index(es) are not in data.")

    if (!(all(y %in% cols))) {
      stop("y index(es) are not in data.")
    }
  }

  # (i) data.frame, (ii) list with variables, (iii) matrix

  # data.frame format to deal with classes
  if (is.list(data) && !is.data.frame(data)) {

    # data names?
    ifelse(is.null(names(data)),
           var_names <- 1:length(data), # if not 1:x
           var_names <- names(data)
    )

    data <- matrix(unlist(data), ncol = length(var_names), byrow = F)
    colnames(data) <- var_names

  } else if (is.matrix(data) || is.data.frame(data)) {
    data <- data.frame(data)
  }

  # Classes
  varClass <- unlist(sapply(data, class))

  # Output classes
  outClass <- varClass[y] %in% c("numeric", "double", "integer")

  # Error
  if (!all(outClass)){
    stop(paste(names(data)[y][!outClass][1], "is not a numeric or integer vector"))
  }

  # Input classes
  # Ordered --> numeric
  for (i in x){
    if (is.ordered(data[, i])) {
      data[, i] <- as.numeric(data[, i])
    }
  }

  # Define classes again
  varClass <- unlist(sapply(data, class))

  inpClass <- varClass[x] %in% c("numeric", "double", "integer")

  # Error
  if (!all(inpClass)){
    stop(paste(names(data)[x][!inpClass][1], "is not a numeric, integer or ordered vector"))
  }

  data <- data[, c(x, y)]

  return(as.matrix(data))
}

#' Simulated efficiency dataset (100 DMUs)
#'
#' Dataset with 100 simulated decision-making units (DMUs) used to illustrate
#' the basic workflow of \pkg{PEAXAI} in a simple single-input/single-output
#' setting.
#'
#' Each DMU uses one input \code{x1} to produce an output \code{y}. The variable
#' \code{yD} represents the theoretical output on the deterministic frontier,
#' that is, the output level that would be observed in the absence of
#' technical inefficiency.
#'
#' @details
#' The dataset is purely simulated and is intended for examples and vignettes.
#' It contains 100 DMUs with heterogeneous input levels and corresponding
#' output levels. The observed output \code{y} can be interpreted as
#' \code{y <= yD}, where the gap between \code{yD} and \code{y} reflects
#' technical inefficiency (plus possible noise, depending on how the data
#' were generated).
#'
#' @format
#' A \code{data.frame} with 100 rows and 3 columns:
#' \describe{
#'   \item{x1}{Input of the DMU (e.g., resource use, cost or effort).}
#'   \item{y}{Observed output, potentially affected by technical inefficiency.}
#'   \item{yD}{Deterministic (theoretical) output on the efficient frontier.}
#' }
#'
#' @source
#' Simulated data generated by the authors for illustrative purposes.
#'
#' @examples
#' data(data)
#' str(data)
#' summary(data)
#'
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   ggplot2::ggplot(data, ggplot2::aes(x = x1)) +
#'     ggplot2::geom_point(ggplot2::aes(y = y), alpha = 0.6) +
#'     ggplot2::geom_line(ggplot2::aes(y = yD), color = "red") +
#'     ggplot2::labs(
#'       x = "Input x1",
#'       y = "Output",
#'       title = "Simulated DMUs and theoretical frontier"
#'     ) +
#'     ggplot2::theme_minimal()
#' }
#'
#' @docType data
#' @name data
#' @usage data(data)
#' @keywords datasets
"data"

#' @title Training Classification Models to Estimate Efficiency
#'
#' @description
#' Trains one or multiple classification algorithms to identify Pareto-efficient
#' decision-making units (DMUs). It jointly searches model hyperparameters and the
#' class-balancing level (synthetic samples via SMOTE) using k-fold cross-
#' validation or a train/validation/test split, selecting the configuration that
#' maximizes the specified metric(s). Returns, for each technique, the best fitted
#' model together with training summaries, performance metrics, and the selected
#' balancing level.
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables in the model.
#' @param x Integer vector with column indices of input variables in \code{data}.
#' @param y Integer vector with column indices of output variables in \code{data}.
#' @param z_numeric Integer vector with column indices of numeric environment variables in \code{data}. By default is \code{NULL}.
#' @param z_factor Integer vector with column indices of factor environment variables in \code{data}. By default is \code{NULL}.
#' @param B number of bootstrap replicates in Conditional DEA.
#' @param m number of units to be included in the reference set.
#' @param alpha This allow to choose the size of the Confidence Intervals computed. By defaulta alpha = FALSE. In this case no confidence interval are computed.
#' @param RTS Text string or number defining the underlying DEA technology /
#'   returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
#'     \item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
#'   }
#' @param trControl A \code{caret::trainControl}-like list that specifies the resampling
#'   strategy; recognized values for \code{$method} include \code{"cv"}, \code{"test_set"},
#'   and \code{"none"}. See \pkg{caret} documentation.
#' @param methods A \code{list} of selected machine learning models and their hyperparameters.
#' @param metric_priority A \code{string} specifying the summary metric for classification to select the optimal model. Default includes \code{"Balanced_Accuracy"} due to (normally) unbalanced data.
#' @param calibration_method Character string specifying the probability calibration
#'   method applied to the fitted classifier. Supported options are:
#'   \describe{
#'     \item{\code{"none"}}{No probability calibration is applied; raw classifier
#'       probabilities are returned.}
#'     \item{\code{"platt"}}{Platt scaling, which fits a logistic regression model
#'       mapping raw scores to calibrated probabilities. Suitable for relatively
#'       small samples and parametric probability adjustment.}
#'     \item{\code{"isotonic"}}{Isotonic regression, a non-parametric monotonic
#'       calibration method that adapts flexibly to the empirical score-probability
#'       relationship. Recommended for larger samples where sufficient calibration
#'       data are available.}
#'   }
#'   Calibration is performed using validation data whenever resampling is enabled.
#' @param hold_out Numeric proportion in (0,1) for validation split (default \code{NULL}).
#'   If \code{NULL}, training and validation use the same data.
#' @param imbalance_rate Optional target(s) for class balance via SMOTE. If \code{NULL},
#'   no synthetic balancing is performed.
#' @param seed  Integer. Seed for reproducibility.
#' @param verbose Logical; if \code{TRUE}, prints progress messages (default \code{FALSE}).
#'
#' @importFrom caret train createDataPartition confusionMatrix createFolds
#' @importFrom dplyr %>% arrange across filter select summarise group_by first semi_join all_of bind_rows
#' @importFrom pROC roc
#' @importFrom PRROC pr.curve
#' @importFrom stats binomial family median na.omit runif setNames approxfun
#' @importFrom utils combn tail
#' @importFrom rms val.prob
#' @importFrom isotone gpava
#' @importFrom np npudensbw
#'
#' @return A \code{"PEAXAI"} (list) with the best technique, best fitted models and their performance and the results by fold.
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'         weights = "dinamic"
#'      )
#'   )
#'
#'   models <- PEAXAI_fitting(
#'     data = data,
#'     x = c(1:4),
#'     y = 5,
#'     RTS = "vrs",
#'     imbalance_rate = NULL,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = c("Balanced_Accuracy", "F1", "ROC_AUC"),
#'     seed = 1,
#'     verbose = FALSE
#'   )
#' }
#'
#' @export

PEAXAI_fitting <- function (
    data, x, y, z_numeric = NULL, z_factor = NULL, RTS = "vrs",
    B = NULL, alpha = FALSE, m = NULL, imbalance_rate = NULL,
    trControl, methods, metric_priority = "Balanced_Accuracy",
    calibration_method = NULL, hold_out = NULL, seed = 314, verbose = TRUE
    ) {

  # ----------------------------------------------------------------------------
  # pre-processing -------------------------------------------------------------
  # ----------------------------------------------------------------------------

  # check if parameters are well introduced
  validate_parametes_PEAXAI_fitting(
    data = data,
    x = x,
    y = y,
    RTS = RTS,
    trControl = trControl,
    methods = methods,
    metric_priority = metric_priority,
    imbalance_rate = imbalance_rate,
    hold_out = hold_out,
    verbose = verbose
  )

  # calibration parameters
  if (is.null(calibration_method)) {
    null_calibration <- TRUE
    calibration_method <- "platt"
  } else {
    null_calibration <- FALSE
  }

  mth_cal <- calibration_method
  alpha <- 0.05

  # # reorder index 'x' and 'y' in data
  data_x <- data[,c(x), drop = FALSE]
  data_y <- data[,c(y), drop = FALSE]
  data_z_numeric <- data[, c(z_numeric), drop = FALSE]
  data_z_factor <- data[, c(z_factor), drop = FALSE]

  x <- 1:NCOL(data_x)
  y <- 1:NCOL(data_y)

  if(!is.null(z_numeric)) {
    z_numeric <- 1:NCOL(data_z_numeric)
  }

  if (!is.null(z_factor)) {
    z_factor <- 1:NCOL(data_z_factor)
  }

  y <- max(x) + y

  data <- cbind(data_x, data_y)

  # Check data
  data <- preprocessing(
    data = data,
    x = x,
    y = y
  )

  if (!is.null(z_numeric)) {
    # index
    z_numeric <- NCOL(data) + z_numeric

    # check
    ok <- all(vapply(data_z_numeric, is.numeric, logical(1)))
    if (!ok) {
      stop("All columns in z_numeric must be numeric.")
    }

    # add data
    data <- cbind(data, data_z_numeric)
  }

  if (!is.null(z_factor)) {
    # index
    z_factor <- NCOL(data) + z_factor

    # check
    ok <- all(vapply(data_z_factor, is.factor, logical(1)))
    if (!ok) {
      stop("All columns in z_factor must be factor.")
    }

    # add data
    data <- cbind(data, data_z_factor)
  }

  # save a copy before add class_efficiency
  copy_data_no_label <- data

  # copy of methods
  method_copy <- methods

  exogenous <- c(z_numeric, z_factor)

  # Similarity matrix of ALL dataset
  if (!is.null(exogenous)) {

    bandwidth <- npudensbw(dat = data[,exogenous])

  } else {
    bandwidth <- NULL
  }

  # ----------------------------------------------------------------------------
  # Step 1: Data labeling ------------------------------------------------------
  # ----------------------------------------------------------------------------
  if (isTRUE(verbose)) {
    message(
      paste0("Step 1: Labeling all observations...")
    )
  }

  data <- label_efficiency(
    data = data,
    x = x,
    y = y,
    z_numeric = z_numeric,
    z_factor = z_factor,
    RTS = RTS,
    B = B,
    m = m,
    alpha = alpha,
    bandwidth = bandwidth,
    seed = seed
  )

  if (isTRUE(verbose)) {
    info_imb_rate <- round(prop.table(table(data$class_efficiency))[1], 4) * 100
    message(
      sprintf(
        'The "efficient" class represents %.2f%% of the dataset.',
        info_imb_rate
      )
    )
  }

  # ----------------------------------------------------------------------------
  # Step 2: Create validation set ----------------------------------------------
  # ----------------------------------------------------------------------------
  # create datasets: (train + test) and (validation)
  if (is.null(hold_out)) {

    # all data is use to train and for validation
    valid_data <- data
    train_data <- data
    all_data <- data

  } else {

    if (isTRUE(verbose)) {
      message(
        sprintf("A hold-out validation set will be created using %.1f%% of the data.",
                hold_out * 100)
      )
    }

    # validation set
    valid_index <- createDataPartition(
      data$class_efficiency,
      p = hold_out,
      list = FALSE)

    # split data
    valid_data <- data[valid_index, c(x,y, z_numeric, z_factor)]

    valid_data <- cbind(
      valid_data,
      data.frame(
        class_efficiency = data[valid_index, "class_efficiency"]
      )
    )

    train_data <- data[-valid_index, c(x,y, z_numeric, z_factor)]

    if (isTRUE(verbose)) {
      message(
        paste0("Step 1.1: Labeling train observations...")
      )
    }

    # label DMUs by technology observed in train
    # valid_data <- label_efficiency(
    #   data = valid_data,
    #   REF = train_data,
    #   x = x,
    #   y = y,
    #   z_numeric = z_numeric,
    #   z_factor = z_factor,
    #   RTS = RTS,
    #   B = B,
    #   m = m,
    #   alpha = alpha,
    #   seed = seed
    # )

    train_data <- label_efficiency(
      data = train_data,
      REF = train_data,
      x = x,
      y = y,
      z_numeric = z_numeric,
      z_factor = z_factor,
      RTS = RTS,
      B = B,
      m = m,
      alpha = alpha,
      seed = seed
    )

    all_data <- data

    save_performance_validation <- vector("list", length = length(methods))
    names(save_performance_validation) <- names(methods)

  }

  # ----------------------------------------------------------------------------
  # Step 3: ML model training --------------------------------------------------
  # ----------------------------------------------------------------------------
  # the original train dataset
  real_balance <- prop.table(table(train_data$class_efficiency))[["efficient"]]
  datasets_to_train <- vector("list", length(imbalance_rate)+1)

  datasets_to_train[[1]] <- train_data
  names(datasets_to_train)[1] <- paste0(as.character(round(real_balance, 4)),"*")

  if(!is.null(imbalance_rate)) {
    names(datasets_to_train)[2:length(datasets_to_train)] <- as.character(imbalance_rate)
  }

  real_balance_stable <- paste0(as.character(round(real_balance, 4)),"*")

  RTS_available <- c("1", "vrs")

  # ----------------------------------------------------------------------------
  # Step 3.1: Train performance with different hyperparameters -----------------
  # ----------------------------------------------------------------------------
  # save model
  best_models <- vector("list", length(methods))
  names(best_models) <- names(methods)

  # save performance
  best_performance  <- vector("list", length(methods))
  names(best_performance) <- names(methods)

  # more information about fitting (grid)
  performance_train_all_by_dataset <- vector("list", length(datasets_to_train))
  names(performance_train_all_by_dataset) <- names(datasets_to_train)

  # only the best by imbalance and ML method
  best_performance_train_all_by_dataset <- vector("list", length(datasets_to_train))
  names(best_performance_train_all_by_dataset) <- names(datasets_to_train)

  # metrics evaluation
  metric <- metric_priority[1]

  set.seed(seed)
  # create k-folds. Same folds for every ML method
  folds <- createFolds(
    datasets_to_train[[1]]$class_efficiency,
    k = trControl[["number"]]
  )

  # performance by fold
  performance_by_fold <- vector("list", length(folds))
  names(performance_by_fold) <- names(folds)

  # save test set (data + indices) for each fold
  test_sets_by_fold <- vector("list", length(folds))
  names(test_sets_by_fold) <- names(folds)

  test_idx_by_fold <- vector("list", length(folds))
  names(test_idx_by_fold) <- names(folds)

  # cross validation
  if (isTRUE(verbose)) {
    message("Step 2: Fitting models using cross-validation...")
  }

  for(fold_i in names(folds)) {

    if (isTRUE(verbose)) {
      message(
        paste0(fold_i, " of ", length(names(folds)))
      )
    }

    test_set_idx <- unlist(folds[fold_i])
    train_set_idx <- unlist(folds[names(folds) != fold_i])

    test_set  <- train_data[test_set_idx, ]
    train_set <- train_data[train_set_idx, ]

    # store test info per fold
    test_sets_by_fold[[fold_i]] <- test_set
    test_idx_by_fold[[fold_i]]  <- test_set_idx

    # delete old label
    train_set <- train_set %>% select(-"class_efficiency")

    # re-label only on train
    train_set <- label_efficiency(
      data = train_set,
      REF = train_set,
      x = x,
      y = y,
      z_numeric = z_numeric,
      z_factor = z_factor,
      RTS = RTS,
      B = B,
      m = round(nrow(train_set)^(2/3)),
      bandwidth = NULL,
      alpha = alpha,
      seed = seed
    )

    # --------------------------------------------------------------------------
    # Step 3.1: Addressing imbalance rate (if there is convexity) ---------------
    # --------------------------------------------------------------------------
    # datasets_to_train <- list(train_set)
    # names(datasets_to_train)[1] <- real_balance_stable

    if (!is.null(imbalance_rate)) {

      if (as.character(RTS) %in% RTS_available) {

        if (is.null(z_numeric) & is.null(z_numeric)) {

          train_data_SMOTE <- SMOTE_data(
            data = train_set,
            x = x,
            y = y,
            RTS = RTS,
            balance_data = imbalance_rate,
            seed = seed
          )

        } else {

          # create new datasets addressing imbalance
          train_data_SMOTE <- SMOTE_Z_data(
            data = train_set,
            x = x,
            y = y,
            z_numeric = z_numeric,
            z_factor = z_factor,
            balance_data = imbalance_rate,
            RTS = RTS,
            B = B,
            m = m,
            alpha = alpha,
            bandwidth = NULL,
            seed = seed
          )

        }

        # save datasets in this fold_i
        datasets_to_train_fold_i <- vector("list", length = 1)
        datasets_to_train_fold_i[[1]] <- train_set
        names(datasets_to_train_fold_i)[1] <- real_balance_stable

        datasets_to_train_fold_i <- append(train_data_SMOTE, datasets_to_train_fold_i, after = 0)

      } else {
        message("Without the convexity assumption, SMOTE units cannot be created on the efficient frontier.")
      }

      # end balancing fold_i

    } else {
      datasets_to_train_fold_i <- vector("list", length = 1)
      datasets_to_train_fold_i[[1]] <- train_set
      names(datasets_to_train_fold_i)[1] <- real_balance_stable
    }

    # Train and get performance
    performance_train_all_by_method <- vector("list", length(methods))
    names(performance_train_all_by_method) <- names(methods)

    best_performance_train_all_by_method <- vector("list", length(methods))
    names(best_performance_train_all_by_method) <- names(methods)

    performance_by_fold_by_grid <- vector("list", length(methods))
    names(performance_by_fold_by_grid) <- names(methods)

    for (method_i in names(methods)) {

      if (isTRUE(verbose)) message(paste0("Training ", method_i, " method."))

      for (datasets_to_train_i in names(datasets_to_train_fold_i)) {

        if(method_i == "glm") {

          methods[[method_i]][["tuneGrid"]] <- data.frame(1)
        }

        for (grid_i in 1:nrow(methods[[method_i]][["tuneGrid"]])) {

          new_parameters <- methods[[method_i]]

          if (is.data.frame(new_parameters[["tuneGrid"]])) {
            new_parameters[["tuneGrid"]] <- new_parameters[["tuneGrid"]][grid_i, ]
          } else {
            new_parameters[["tuneGrid"]] <- new_parameters[["tuneGrid"]][grid_i]
            new_parameters[["tuneGrid"]] <- as.data.frame(new_parameters[["tuneGrid"]])
            names(new_parameters[["tuneGrid"]]) <- names(methods[[method_i]][["tuneGrid"]])
          }

          # train model (inner CV handled in train_PEAXAI through trControl)
          model_fit <- train_PEAXAI(
            data = datasets_to_train_fold_i[[datasets_to_train_i]],
            method = method_i,
            parameters = new_parameters,
            trControl = trControl,
            metric_priority = metric_priority,
            seed = seed
          )

          # --------------------------------------------------------------------
          # performance --------------------------------------------------------
          # --------------------------------------------------------------------
          levls <- c("efficient", "not_efficient")

          # raw probabilities
          y_hat_prob <- predict(model_fit, newdata = test_set, type = "prob")[, 1]
          if (all(is.na(y_hat_prob))) {
            if (isTRUE(verbose)) {
              warning("Predictions returned NA for method ", method_i, ". Generating NA performance for this configuration.")
            }

            performance_metrics_names <- c(
              "Accuracy", "Kappa", "Recall", "Specificity",
              "Precision", "F1", "Balanced_Accuracy",
              "G_mean", "ROC_AUC", "PR_AUC",
              "Cross_Entropy", "Cross_Entropy_Efficient_class",
              "Cross_Entropy_not_Efficient_class"
            )

            performance_na <- as.data.frame(as.list(stats::setNames(rep(NA_real_, length(performance_metrics_names)), performance_metrics_names)))

            performance <- cbind(
              data.frame(method = method_i),
              data.frame(Fold = fold_i),
              new_parameters[["tuneGrid"]],
              performance_na
            )

            performance_by_fold_by_grid[[method_i]][[datasets_to_train_i]][[grid_i]] <- performance
            next
          }

          # classifications
          y_hat <- ifelse(y_hat_prob > 0.5, "efficient", "not_efficient")
          y_hat <- factor(y_hat, levels = levls)

          # reference
          y_obs <- factor(test_set[, "class_efficiency"], levels = levls)

          # # save calibration test per fold/method/balance/grid
          # calibration <- cbind(as.data.frame(y_obs), y_hat_prob)
          # names(calibration) <- c("obs", "efficient")
          #
          # # check if Calibration is needed
          # y_obs_label <- ifelse(y_obs == "efficient", 1,0)
          #
          # eps <- 1e-5
          # y_hat_prob <- pmin(pmax(y_hat_prob, eps), 1 - eps)
          #
          # calibration$DMU <- test_set_idx
          #
          # # store calibration for this fold
          # calibrating_datasets_by_method[[method_i]][[datasets_to_train_i]][[fold_i]][[grid_i]] <- calibration

          # confusion matrix
          cm <- confusionMatrix(
            data = y_hat,
            reference = y_obs,
            mode = "everything",
            positive = "efficient"
          )

          if (is.na(cm$byClass[["F1"]])) {
            cm$byClass[["F1"]] <- 0
          }

          if (is.na(cm$byClass[["Precision"]])) {
            cm$byClass[["Precision"]] <- 0
          }

          # ROC/PR (robust when 1 class)
          if (length(unique(y_obs)) == 1) {

            roc_auc_val <- 0
            pr_auc_val  <- 0

          } else {

            roc_obj <- roc(
              response  = y_obs,
              predictor = y_hat_prob,
              levels    = rev(levls),
              direction = "<",
              quiet     = TRUE
            )

            pr_obj <- pr.curve(
              scores.class0 = y_hat_prob[y_obs == "not_efficient"],
              scores.class1 = y_hat_prob[y_obs == "efficient"],
              curve = TRUE
            )

            roc_auc_val <- as.numeric(roc_obj$auc)
            pr_auc_val  <- unname(pr_obj$auc.integral)
          }

          # gmean
          gmean <- sqrt(cm$byClass["Recall"]*cm$byClass["Specificity"])

          # --------------------------------------------------------------------
          # cross-entropy ------------------------------------------------------
          # --------------------------------------------------------------------
          # cross start
          y_bin <- ifelse(y_obs == "efficient", 1, 0)
          p <- y_hat_prob
          eps <- 1e-15
          p_clipped <- pmin(pmax(p, eps), 1 - eps)

          cross_entropy <- -mean(
            y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
          )

          # cross-entropy EFFICIENT
          idx_eff <- which(y_bin == 1)
          y_bin_eff <- y_bin[idx_eff]
          p <- y_hat_prob[idx_eff]
          eps <- 1e-15
          p_clipped <- pmin(pmax(p, eps), 1 - eps)

          cross_entropy_eff <- -mean(
            y_bin_eff * log(p_clipped) + (1 - y_bin_eff) * log(1 - p_clipped)
          )

          # cross-entropy NOT EFFICIENT
          idx_not_eff <- which(y_bin == 0)
          y_bin_not_eff <- y_bin[idx_not_eff]
          p <- y_hat_prob[idx_not_eff]
          eps <- 1e-15
          p_clipped <- pmin(pmax(p, eps), 1 - eps)

          cross_entropy_not_eff <- -mean(
            y_bin_not_eff * log(p_clipped) + (1 - y_bin_not_eff) * log(1 - p_clipped)
          )


          performance <- c(
            cm$overall[c("Accuracy", "Kappa")],
            cm$byClass[c("Recall", "Specificity",
                         "Precision", "F1",
                         "Balanced Accuracy")],
            "G_mean" = unname(gmean),
            "ROC_AUC" = roc_auc_val,
            "PR_AUC" = pr_auc_val,
            "Cross_Entropy" = cross_entropy,
            "Cross_Entropy_Efficient_class" = cross_entropy_eff,
            "Cross_Entropy_not_Efficient_class" = cross_entropy_not_eff
          )

          names(performance)[7] <- "Balanced_Accuracy"
          performance_metrics_names <- names(performance)

          performance <- cbind(
            data.frame(method = method_i),
            data.frame(Fold = fold_i),
            new_parameters[["tuneGrid"]],
            as.data.frame(as.list(performance))
          )

          performance_by_fold_by_grid[[method_i]][[datasets_to_train_i]][[grid_i]] <- performance

        } # end grid loop

      } # end datasets loop

    } # end methods loop

    performance_by_fold[[fold_i]] <- performance_by_fold_by_grid

  } # end folds loop

  # method
  names_methods <- names(methods)
  results_fit <- vector("list", length(methods))

  for (method_i in names(methods)) {

    performace_by_imbalance <- NULL

    for (balance_i in names(datasets_to_train)) {

      for (grid_i in 1:NROW(methods[[method_i]][["tuneGrid"]])) {

        performace_tunGrid <- methods[[method_i]][["tuneGrid"]]

        for (fold_i in names(performance_by_fold)) {

          # warning(performance_by_fold[[fold_i]][[method_i]][[balance_i]][[grid_i]])
          performance_i <- performance_by_fold[[fold_i]][[method_i]][[balance_i]][[grid_i]]
          performance_i <- performance_i[performance_metrics_names]

          results_fit[[method_i]][[balance_i]][[fold_i]] <- performance_i

        }

        if(!is.data.frame(performace_tunGrid)) {
          performace_tunGrid <- as.data.frame(performace_tunGrid)
          names(performace_tunGrid) <- names(methods[[method_i]][["tuneGrid"]])
        }

        performace_tunGrid <- performace_tunGrid[grid_i,]

        performance_fold <- bind_rows(results_fit[[method_i]][[balance_i]])

        # me gustaría que si hay NAs en algún fold, sea todo NA la salida.
        if (any(is.na(performance_fold[performance_metrics_names]))) {

          # Generar NA para todas las métricas y sus desviaciones estándar
          mean_fold <- as.data.frame(as.list(stats::setNames(rep(NA_real_, length(performance_metrics_names)), performance_metrics_names)))
          sd_fold <- as.data.frame(as.list(stats::setNames(rep(NA_real_, length(performance_metrics_names)), paste0(performance_metrics_names, "SD"))))

        } else {

          mean_fold <-  performance_fold %>%
            summarise(
              # mean
              across(all_of(performance_metrics_names), ~mean(.x, na.rm = FALSE))
            )

          sd_fold <- performance_fold %>%
            summarise(
              # SD
              across(all_of(performance_metrics_names), ~sd(.x, na.rm = FALSE),
                     .names = "{col}SD")
            )
        }

        if (is.numeric(performace_tunGrid)) {
          performace_tunGrid <- as.data.frame(performace_tunGrid)
          names(performace_tunGrid) <- names(methods[[method_i]][["tuneGrid"]])
        }
        performance <- cbind(performace_tunGrid, mean_fold, sd_fold)

        performance_train_all_by_method[[method_i]][[balance_i]][[grid_i]] <- performance

      } # end by  grid

      performance_train_all_by_method[[method_i]][[balance_i]]
      performance <- bind_rows(performance_train_all_by_method[[method_i]][[balance_i]])

      # performance_train_all_by_method[[method_i]][[balance_i]] <- performance
      df_imbalance <- as.data.frame(balance_i)
      names(df_imbalance) <- "Imbalance_rate"

      if (method_i == "glm") {

        performance <- performance[2:ncol(performance)]
        performance <- cbind(df_imbalance, performance)
      } else {
        names_methods_grid <- names(methods[[method_i]][["tuneGrid"]])
        df_parameters <- performance[, names_methods_grid]
        if (is.vector(df_parameters)) {
          df_parameters <- as.data.frame(df_parameters)
          names(df_parameters) <- names_methods_grid
        }
        names_metrics <- setdiff(names(performance), names_methods_grid)
        df_metrics <- performance[, names_metrics]
        performance <- cbind(df_imbalance, df_parameters, round(df_metrics, 2))
      }

      performace_by_imbalance <- rbind(performace_by_imbalance, performance)

    } # end by balance

    performance_train_all_by_method[[method_i]] <- performace_by_imbalance

  }

  # ----------------------------------------------------------------------------
  # Step 3.3: Train the best model and assing its performance ------------------
  # ----------------------------------------------------------------------------
  # train the best models without cross-validation
  save_performance_train <- vector("list", length = length(methods))
  names(save_performance_train) <- names(methods)

  save_best_model_fit <- vector("list", length = length(methods))
  names(save_best_model_fit) <- names(methods)

  save_calibration_model <- vector("list", length(methods))
  names(save_calibration_model) <- names(methods)

  for (method_i in names(methods)) {

    # first, select the best imbalance rate
    # dataframe with results of CV of method_i

    best_results_train <- performance_train_all_by_method[[method_i]]

    best_results_train <- as.data.frame(best_results_train)

    best_results_train$Imbalance_rate <- as.numeric(gsub("\\*", "", best_results_train$Imbalance_rate))
    best_results_train$diff_imbalance <- abs(best_results_train$Imbalance_rate - round(real_balance, 4))

    # how similar is the dataset to original dataset
    best_results_train$diff_imbalance <- 1 - best_results_train$diff_imbalance

    # add * in the orginal dataset name
    best_results_train$Imbalance_rate[best_results_train$diff_imbalance == 1] <- paste0(best_results_train$Imbalance_rate[best_results_train$diff_imbalance == 1], "*")

    # add difference imbalance
    metric_priority <- c(metric_priority, "diff_imbalance")

    best_results_train <- best_results_train[do.call(
      order, c(best_results_train[metric_priority], list(decreasing = TRUE))), ]

    best_balance_i <- best_results_train[1, "Imbalance_rate"]

    # copy of methods
    best_methods <- methods

    # only the best hyperparameters
    names_tuneGrid <- names(best_methods[[method_i]][["tuneGrid"]])

    if (method_i != "glm") {
      best_methods[[method_i]][["tuneGrid"]] <- best_results_train[1, names_tuneGrid]

      # get idx of hyparameters
      tg <- method_copy[[method_i]][["tuneGrid"]]
      br <- best_results_train[1, names_tuneGrid, drop = FALSE]

      idx <- which(
        apply(tg[, names_tuneGrid, drop = FALSE], 1,
              function(r) all(r == unlist(br)))
      )

    } else {
      idx <- 1
    }

    # train the best model TRAIN
    # NO VALIDATION SET

    if (best_balance_i != real_balance_stable) {
      # generate the best balanced class distribution

      if (is.null(z_numeric) & is.null(z_numeric)) {

        # create new datasets addressing imbalance
        train_data_SMOTE <- SMOTE_data(
          data = train_data,
          x = x,
          y = y,
          RTS = RTS,
          balance_data = as.numeric(best_balance_i),
          seed = seed
        )

      } else {

        # create new datasets addressing imbalance
        train_data_SMOTE <- SMOTE_Z_data(
          data = train_data,
          x = x,
          y = y,
          z_numeric = z_numeric,
          z_factor = z_factor,
          balance_data = as.numeric(best_balance_i),
          RTS = RTS,
          B = B,
          m = m,
          alpha = alpha,
          bandwidth = bandwidth,
          seed = seed
        )

      }

      datasets_to_train[[best_balance_i]] <- train_data_SMOTE[[best_balance_i]]

    }

    # All data used, performance is obteined of cv results
    model_fit <- train_PEAXAI(
      data = datasets_to_train[[best_balance_i]],
      method = method_i,
      parameters = best_methods[[method_i]],
      trControl = trainControl(method = "none", classProbs = TRUE),
      metric_priority = metric_priority[1],
      seed = seed
    )

    performance <- best_results_train[1,]

#     # save calibration
#     calibration_dataset <- NULL
#     for(fold_i in names(folds)) {
#
#       calibration_dataset <- rbind(calibration_dataset, calibrating_datasets_by_method[[method_i]][[as.character(best_balance_i)]][[fold_i]][[idx]])
#     }
#
#     calibration_dataset$obs <- ifelse(calibration_dataset$obs == "efficient", 1,0)
#
#     res <- val.prob(
#       p = calibration_dataset$efficient,
#       y = calibration_dataset$obs)
#
#     # check if it is necessary calibrator model
#     if (res["U:p"] < alpha | res["S:p"] < alpha | abs(res["Intercept"]) > 0.2 | res["Slope"] - 1 > 0.2) {
#
#       # function to apply calibration
#       predict_calibrated <- function(newdata, model_fit, calibration_model, mth_cal) {
#
#         s_new <- predict(model_fit, newdata = newdata, type = "prob")[, "efficient"]
#
#         if (mth_cal == "isotonic") {
#           # calibration_model is a function: iso_fun
#           return(calibration_model(s_new))
#         }
#
#         # platt (glm)
#         return(
#           predict(calibration_model, newdata = data.frame(s = s_new), type = "response")
#         )
#       }
#
#       if (mth_cal == "isotonic") {
#
#         # 1. Variable respuesta binaria (1 = efficient)
#         y_prob <- calibration_dataset$obs
#
#         # 2. Probabilidades "raw" del modelo en el conjunto de calibración
#         s <- calibration_dataset$efficient
#
#         # 3. Ajuste isotónico (monótono creciente)
#         iso_fit <- gpava(z = s, y = y_prob)
#
#         # 4. Creamos una función continua para cualquier valor de probabilidad
#         iso_fun <- approxfun(iso_fit$x, iso_fit$yf, rule = 2)
#
#         calibration_model <- iso_fun
#
#       } else {
#
#         # platt
#         # 1. Variable respuesta binaria (1 = efficient)
#         y_prob <- calibration_dataset$obs
#
#         # 2. Probabilidades "raw" del modelo en el conjunto de calibración
#         s <- calibration_dataset$efficient
#
#         # Platt
#         platt_mod <- glm(y_prob ~ s, family = binomial)
#         summary(platt_mod)
#
#         calibration_model <- platt_mod
#
#       }
# browser()
#       # check is calibration works
#       # check calibration
#       calibration_dataset_cal <- calibration_dataset
#
#       calibration_dataset_cal$efficient <- predict_calibrated(
#         newdata = data[calibration_dataset_cal$DMU,c(x,y,z_numeric,z_factor)],
#         model_fit = model_fit,
#         calibration_model = calibration_model,
#         mth_cal = mth_cal)
#
#       res_2 <- val.prob(
#         p = calibration_dataset_cal$efficient,
#         y = calibration_dataset_cal$obs)
#
#       # score_cal <- function(r) {
#       #
#       #   # # penalizes deviations from (Intercept = 0, Slope = 1) and calibration errors
#       #   # abs(r[["Intercept"]]) +
#       #   #   abs(r[["Slope"]] - 1) +
#       #   #   # r[["Emax"]] +
#       #   #   r[["Eavg"]]
#       #   1 - r[["C (ROC)"]]
#       #
#       # }
#       #
#       # accept <- score_cal(res_2) < score_cal(res)
#       # accept <- TRUE
#       #
#       # if (accept == FALSE) {
#       #   # ----------------------------------------------------------------------
#       #   # calibration doesn't work ---------------------------------------------
#       #   # ----------------------------------------------------------------------
#       #   calibration_model <- NULL
#       # }
#
#     } else {
#       # ------------------------------------------------------------------------
#       # NO calibration ---------------------------------------------------------
#       # ------------------------------------------------------------------------
#       calibration_model <- NULL
#     }

    # save results
    save_performance_train[[method_i]] <- performance
    save_best_model_fit[[method_i]] <- model_fit
    # save_calibration_model[[method_i]] <- calibration_model

    # WITH VALIDATION SET
    # It is necessary to check the performance of the best model without cv on validation set
    # after that, it is necessary to train again WITH ALL data and the best hyperparameters
    if (!is.null(hold_out)) {

      # --------------------------------------------------------------------
      # performance --------------------------------------------------------
      # --------------------------------------------------------------------
      calibration_model <- NULL
      # check performance on valid data
      if (!is.null(calibration_model)) {

        # y_hat <- predict_calibrated(
        #   newdata = valid_data,
        #   model_fit = model_fit,
        #   calibration_model = calibration_model,
        #   mth_cal = mth_cal
        # )

      } else {

        y_hat <- predict(model_fit, newdata = valid_data, type = "prob")[["efficient"]]

      }

      # y_hat <- predict(model_fit, newdata = valid_data, type = "prob")[["efficient"]]
      y_hat_prob <- y_hat
      y_obs <- valid_data$class_efficiency
      y_obs_label <- ifelse(valid_data$class_efficiency == "efficient", 1, 0)

      # res <- val.prob(
      #   p = y_hat_prob,
      #   y = y_obs_label)

      if (any(is.na(y_hat))) {

        out_names <- c("Accuracy", "Kappa",
                       "Recall", "Specificity", "Precision", "F1", "Balanced_Accuracy", "G_mean",
                       "ROC-AUC", "PR-AUC",
                       "Cross_Entropy", "Cross_Entropy_Efficient")

        performance <- setNames(rep(NA, length(out_names)), out_names)

      } else {

        # levels
        levls <- c("efficient", "not_efficient")

        # predictions
        y_hat_prob <- y_hat
        y_hat <- ifelse(y_hat > 0.5, "efficient", "not_efficient")
        y_hat <- factor(y_hat, levels = levls)

        # reference
        y_obs <- factor(y_obs, levels = levls)

        # confusion matrix
        cm <- confusionMatrix(
          data = y_hat,
          reference = y_obs,
          mode = "everything",
          positive = "efficient"
        )

        # G-mean
        gmean <- sqrt(cm$byClass["Recall"]*cm$byClass["Specificity"])

        # ROC-AUC
        roc_obj <- roc(
          response = y_obs,
          predictor = y_hat_prob,
          levels = rev(levls),
          direction = "<",
          quiet = TRUE)

        # PR-AUC
        pr_obj <- pr.curve(
          scores.class0 = y_hat_prob[y_obs == "not_efficient"],
          scores.class1 = y_hat_prob[y_obs == "efficient"],
          curve = TRUE
        )

        # cross_entropy
        y_bin <- ifelse(y_obs == "efficient", 1, 0)
        p <- y_hat_prob

        # no log(0)
        eps <- 1e-15
        p_clipped <- pmin(pmax(p, eps), 1 - eps)

        cross_entropy <- -mean(
          y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
        )

        # cross entropy efficient class
        p <- y_hat_prob[y_bin == 1]
        y_bin <- y_bin[y_bin == 1]

        # no log(0)
        eps <- 1e-15
        p_clipped <- pmin(pmax(p, eps), 1 - eps)

        cross_entropy_efficient <- -mean(
          y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
        )

        #
        # cross entropy NOT efficient class
        # cross_entropy
        y_bin <- ifelse(y_obs == "efficient", 1, 0)
        p <- y_hat_prob

        p <- y_hat_prob[y_bin == 0]
        y_bin <- y_bin[y_bin == 0]

        # no log(0)
        eps <- 1e-15
        p_clipped <- pmin(pmax(p, eps), 1 - eps)

        cross_entropy_not_efficient <- -mean(
          y_bin * log(p_clipped) + (1 - y_bin) * log(1 - p_clipped)
        )

        performance <- c(
          cm$overall[c("Accuracy", "Kappa")],
          cm$byClass[c("Recall", "Specificity",
                       "Precision", "F1")],
          "Balanced_Accuracy" = cm$byClass[["Balanced Accuracy"]],
          "G_mean" = unname(gmean),
          "ROC_AUC" = roc_obj$auc,
          "PR_AUC" = unname(pr_obj$auc.integral),
          "Cross_Entropy" = cross_entropy,
          "Cross_Entropy_Efficient" = cross_entropy_efficient,
          "Cross_Entropy_not_Efficient" = cross_entropy_not_efficient
        )

        # save a copy
        same_performance <- performance

        best_imbalance_rate <- as.data.frame(best_balance_i)
        names(best_imbalance_rate) <- "Imbalance_rate"

        best_hyperparameter <- best_results_train[1, names_tuneGrid]

        if (method_i == "glm") {

          parameters <- methods[["glm"]]

          if (parameters[["weights"]][1] == "dinamic") {
            w0 <- nrow(datasets_to_train[[as.character(best_balance_i)]]) / (2 * length(which(datasets_to_train[[as.character(best_balance_i)]]$class_efficiency == "not_efficient")))
            w1 <- nrow(datasets_to_train[[as.character(best_balance_i)]]) / (2 * length(which(datasets_to_train[[as.character(best_balance_i)]]$class_efficiency == "efficient")))
          } else if (is.data.frame(parameters[["weights"]])) {
            w0 <- parameters[["weights"]][["w0"]]
            w1 <- parameters[["weights"]][["w1"]]
          } else {
            w0 <- 1
            w1 <- 1
          }

          best_hyperparameter <- data.frame(w0, w1)

        }

        performance <- t(performance)

        performance <- cbind(
          best_imbalance_rate,
          best_hyperparameter,
          performance)

      }

      save_performance_validation[[method_i]] <- performance

      # train the FINAL model train+hold_out
      if (best_balance_i != real_balance_stable) {
        # generate the best balanced class distribution

        if (is.null(z_numeric) & is.null(z_numeric)) {

          # create new datasets addressing imbalance
          all_data_SMOTE <- SMOTE_data(
            data = all_data,
            x = x,
            y = y,
            RTS = RTS,
            balance_data = as.numeric(best_balance_i),
            seed = seed
          )

          all_data_SMOTE <- all_data_SMOTE[[1]]

        } else {

          # create new datasets addressing imbalance
          all_data_SMOTE <- SMOTE_Z_data(
            data = all_data,
            x = x,
            y = y,
            z_numeric = z_numeric,
            z_factor = z_factor,
            balance_data = as.numeric(best_balance_i),
            RTS = RTS,
            B = B,
            m = m,
            alpha = alpha,
            bandwidth = bandwidth,
            seed = seed
          )

          all_data_SMOTE <- all_data_SMOTE[[1]]

        }

      } else {
        all_data_SMOTE <- all_data
      }

      model_fit <- train_PEAXAI(
        data = all_data_SMOTE,
        method = method_i,
        parameters = best_methods[[method_i]],
        trControl = trainControl(method = "none", classProbs = TRUE),
        metric_priority = metric_priority[1],
        seed = seed
      )

      save_best_model_fit[[method_i]] <- model_fit

    } # end hold out

  } # end method

  # show results without hold out
  if(is.null(hold_out)) {

    if(null_calibration) {
      save_calibration_model <- NULL
    }

    output_PEAXAI_models <- list(
      best_model_fit = save_best_model_fit,
      performance_train = save_performance_train,
      performance_train_all = performance_train_all_by_method
      # calibration_dataset = save_calibrations_dataset,
      # calibration_model = save_calibration_model
    )

    return(output_PEAXAI_models)

  } else {
    # validation output

    output_PEAXAI_models <- list(
      best_model_fit = save_best_model_fit,
      performace_hold_out = save_performance_validation,
      performance_train = save_performance_train,
      performance_train_all = performance_train_all_by_method
      # calibration_dataset = save_calibrations_dataset,
      # calibration_model = save_calibration_model
    )

    return(output_PEAXAI_models)
  }
browser()
}

#' @title Create New SMOTE Units to Balance Data combinations of m + s
#'
#' @description This function creates new DMUs to address data imbalances. If the majority class is efficient, it generates new inefficient DMUs by worsering the observed units. Conversely, if the majority class is inefficient, it projects inefficient DMUs to the frontier. Finally, a random selection if performed to keep a proportion of 0.65 for the majority class and 0.35 for the minority class.
#'
#' @param data A \code{data.frame} containing the variables used in the model.
#' @param x Column indexes of the input variables in the \code{data}.
#' @param y Column indexes of the output variables in the \code{data}.
#' @param RTS Text string or number defining the underlying DEA technology /
#'   returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
#'   \describe{
#'     \item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
#'     \item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
#'     \item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
#'     \item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
#'     \item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
#'     \item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
#'   }
#' @param balance_data Indicate level of efficient units to achive and the number of efficient and not efficient units.
#' @param seed  Integer. Seed for reproducibility.
#'
#' @return It returns a \code{data.frame} with the newly created set of DMUs incorporated.
#' @export

SMOTE_data <- function (
    data, x, y, RTS = "vrs", balance_data, seed
) {

  # first, determine the efficient facets
  facets <- convex_facets(
    data = data,
    x = x,
    y = y,
    RTS = RTS
  )

  # second, populate the efficient facets
  balance_datasets <- get_SMOTE_DMUs(
    data = data,
    facets = facets,
    x = x,
    y = y,
    balance_data = balance_data,
    seed = seed,
  )

  return(balance_datasets)
}

#' @title Projection-Based Efficiency counterfactuals
#'
#' @description
#' Computes efficiency projections for each observation based on a trained
#' classifier from \pkg{caret} that provides class probabilities via
#' \code{predict(type = "prob")}. For each probability threshold, the function
#' finds the direction and magnitude of change in input–output space required
#' for a unit to reach a specified efficiency level, following a directional
#' distance approach.
#'
#' @param data A \code{data.frame} or \code{matrix} containing input and output variables.
#' @param x A numeric vector indicating the column indexes of input variables in \code{data}.
#' @param y A numeric vector indicating the column indexes of output variables in \code{data}.
#' @param final_model A fitted \pkg{caret} model of class \code{"train"} that supports
#'   \code{predict(type = "prob")} and returns a probability column for the efficient class.
#' @param calibration_model Optional probability-calibration model applied to the raw predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
#' If provided, calibrated probabilities are used for ranking and threshold-based decisions.
#' Set to \code{NULL} to use uncalibrated predictions.
#' @param sign If it is TRUE, only
#' @param efficiency_thresholds A numeric vector of probability levels in (0,1)
#'   that define the efficiency classes (e.g., \code{c(0.75, 0.9, 0.95)}).
#' @param directional_vector A \code{list} with the required information to
#'   construct the directional vector, including:
#'   \itemize{
#'     \item \code{relative_importance}: Numeric vector of variable importances that sum to 1.
#'     \item \code{baseline}: \code{"mean"}, \code{"median"}, \code{"self"} or \code{"ones"}.
#'   }
#' @param n_expand Numeric. Number of expansion steps used to enlarge the initial
#'   search range for \eqn{\beta}.
#' @param n_grid Integer. Number of grid points evaluated during each iteration
#'   to refine the cutoff value of \eqn{\beta}.
#' @param max_y Numeric. Upper-limit multiplier for output expansion in the search
#'   procedure (default = 2).
#' @param min_x Numeric. Lower-limit multiplier for input contraction in the search
#'   procedure (default = 1).
#'
#' @details
#' For each observation and for each probability level in \code{efficiency_thresholds},
#' the function searches for the smallest directional distance \eqn{\beta} such that
#' the predicted probability of belonging to the efficient class reaches the target.
#'
#' @return
#' A named \code{list} with one element per threshold. Each element contains:
#' \itemize{
#'   \item \code{data}: A \code{data.frame} of projected input–output values at that threshold.
#'   \item \code{beta}: A two-column \code{data.frame} with the optimal \eqn{\beta}
#'         and the corresponding predicted probability.
#' }
#'
#' @seealso
#' \code{\link{find_beta_maxmin}} for initializing search bounds;
#' \code{\link[caret]{train}} for model training.
#'
#' @examples
#' \donttest{
#'   data("firms", package = "PEAXAI")
#'
#'   data <- subset(
#'     firms,
#'     autonomous_community == "Comunidad Valenciana"
#'   )
#'
#'   x <- 1:4
#'   y <- 5
#'   RTS <- "vrs"
#'   imbalance_rate <- NULL
#'
#'   trControl <- list(
#'     method = "cv",
#'     number = 3
#'   )
#'
#'   # glm method
#'   methods <- list(
#'     "glm" = list(
#'       weights = "dinamic"
#'      )
#'   )
#'
#'   metric_priority <- c("Balanced_Accuracy", "F1", "ROC_AUC")
#'
#'   models <- PEAXAI_fitting(
#'     data = data, x = x, y = y, RTS = RTS,
#'     imbalance_rate = imbalance_rate,
#'     methods = methods,
#'     trControl = trControl,
#'     metric_priority = metric_priority,
#'     verbose = FALSE,
#'     seed = 1
#'   )
#'
#'   final_model <- models[["best_model_fit"]][["glm"]]
#'
#'   importance_method <- list(name = "PI", n.repetitions = 5)
#'
#'   relative_importance <- PEAXAI_global_importance(
#'     final_model = final_model,
#'     x = x,
#'     y = y,
#'     explain_data = data,
#'     reference_data = final_model$trainingData,
#'     importance_method = importance_method,
#'     seed = 1
#'   )
#'
#'   efficiency_thresholds <- seq(0.75, 0.95, 0.1)
#'
#'   directional_vector <- list(
#'     relative_importance = relative_importance,
#'     baseline  = "mean"
#'   )
#'
#'   targets <- PEAXAI_counterfactuals(
#'     data = data,
#'     x = x, y = y,
#'     final_model = final_model,
#'     efficiency_thresholds = efficiency_thresholds,
#'     directional_vector = directional_vector,
#'     n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
#'   )
#' }
#'
#' @export

PEAXAI_counterfactuals <- function (
    data, x, y, final_model, calibration_model = NULL,
    sign = FALSE, efficiency_thresholds, directional_vector,
    n_expand, n_grid, max_y = 2, min_x = 1
) {

  # validate_parametes_PEAXAI_targets(
  #   data, x, y, final_model,
  #   efficiency_thresholds, directional_vector,
  #   n_expand, n_grid, max_y, min_x
  # )

  data <- as.data.frame(data)

  # reorder index 'x' and 'y' in data
  data <- data[, c(x,y)]
  x <- 1:(ncol(data) - length(y))
  y <- (length(x) + 1):ncol(data)

  names_data <- names(data[,c(x,y)])

  # min and max values
  min_x_possible <- apply(as.matrix(data[,x]), 2, min)

  max_y_possible <- apply(as.matrix(data[,y]), 2, max)

  # number of decimals to round
  precision_prob <- 5

  # ----------------------------------------------------------------------------
  # Build vector G (directional vector) ----------------------------------------
  # ----------------------------------------------------------------------------
  is_global <- is.null(nrow(directional_vector[["relative_importance"]])) || nrow(directional_vector[["relative_importance"]]) == 1

  if (is_global) {

    # # relative importance
    # score_imp_x <- as.numeric(directional_vector[["relative_importance"]][x])
    # score_imp_y <- as.numeric(directional_vector[["relative_importance"]][y])

    # baseline
    if (directional_vector[["baseline"]] == "mean") {

      # new
      L2_nrom_mean <- sqrt(sum(colMeans(data[, c(x,y)])^2))

      # normalize
      v <- colMeans(data[, c(x,y)])

      v_w <- v * directional_vector[["relative_importance"]]

      # noralize
      L2_nrom_vector <- sqrt(sum(v_w[, c(x,y)]^2))
      v_final <- (v_w / L2_nrom_vector) * L2_nrom_mean

      # vector_gx <- as.data.frame(v_final[,x])
      # names(vector_gx) <- names_data[x]
      # vector_gx <- -vector_gx
      # vector_gy <- as.data.frame(v_final[,y])
      # names(vector_gy) <- names_data[y]
      # end new

      # baseline_x <- vector_gx
      # baseline_y <- vector_gy
      # baseline_x <- as.data.frame(t(apply(as.matrix(data[,x]), 2, mean)))
      # names(baseline_x) <- names_data[x]
      # baseline_y <- as.data.frame(t(apply(as.matrix(data[,y]), 2, mean)))
      # names(baseline_y) <- names_data[y]

    } else if (directional_vector[["baseline"]] == "median") {
stop("Not available.")
      # new
      L2_nrom_mean <- sqrt(sum(colMeans(data[, c(x,y)])^2))

      # normalize
      v <- colMeans(data[, c(x,y)])

      v_w <- v * directional_vector[["relative_importance"]]

      # noralize
      L2_nrom_vector <- sqrt(sum(v_w[, c(x,y)]^2))
      v_final <- (v_w / L2_nrom_vector) * L2_nrom_mean

      # # directional vector
      # vector_gx <- as.data.frame(t(v_final[,x]))
      # names(vector_gx) <- names_data[x]
      # vector_gx <- -vector_gx
      # vector_gy <- as.data.frame(t(v_final[,y]))
      # names(vector_gy) <- names_data[y]
      # end new

      # baseline_x <- as.data.frame(t(apply(as.matrix(data[,x]), 2, median)))
      # names(baseline_x) <- names_data[x]
      # baseline_y <- as.data.frame(t(apply(as.matrix(data[,y]), 2, median)))
      # names(baseline_y) <- names_data[y]

    } else if (directional_vector[["baseline"]] == "self") {
      stop("Not available.")
      baseline_x <- (data[,x])
      names(baseline_x) <- names_data[x]
      baseline_y <- (data[,y])
      names(baseline_y) <- names_data[y]

    } else if (directional_vector[["baseline"]] == "ones") {
      stop("Not available.")
      baseline_x <- as.data.frame(t(rep(1, NCOL(data[,x]))))
      names(baseline_x) <- names_data[x]
      baseline_y <- as.data.frame(t(rep(1, NCOL(data[,y]))))
      names(baseline_y) <- names_data[y]

    }

    # # directional vector
    # vector_gx <- as.data.frame(-(score_imp_x * baseline_x))
    # names(vector_gx) <- names_data[x]
    #
    # vector_gy <- as.data.frame(score_imp_y * baseline_y)
    # names(vector_gy) <- names_data[y]

    # directional vector
    if (!is.null(nrow(v_final)) && nrow(v_final) > 1) {
      v_final <- t(v_final)
    }

    if (is.null(dim(v_final))) {
      v_final <- as.data.frame(t(v_final))
    } else {
      v_final <- as.data.frame(v_final)
    }

    # directional vector
    vector_gx <- v_final[, x, drop = FALSE]
    names(vector_gx) <- names_data[x]
    vector_gx <- -vector_gx

    vector_gy <- v_final[, y, drop = FALSE]
    names(vector_gy) <- names_data[y]

  } else {

    if (!is.null(nrow(directional_vector[["relative_importance"]])) && nrow(directional_vector[["relative_importance"]]) != nrow(data)) {
      stop("Error: The number of rows in relative_importance must be 1 (global) or equal to the number of DMUs in data (local).")
    }

    if (isTRUE(sign)) {

      mat <- directional_vector[["relative_importance"]]

      # Asegurar matriz numérica
      mat <- as.matrix(mat)
      storage.mode(mat) <- "numeric"

      # ----------------------------
      # Columnas x:
      # positivos -> 0
      # negativos -> se quedan igual
      # ----------------------------
      tmp_x <- mat[, x, drop = FALSE]
      tmp_x[!is.na(tmp_x) & tmp_x > 0] <- 0
      mat[, x] <- tmp_x

      # ----------------------------
      # Columnas y:
      # positivos -> 0
      # negativos -> positivos
      # ----------------------------
      tmp_y <- mat[, y, drop = FALSE]
      tmp_y[!is.na(tmp_y) & tmp_y > 0] <- 0
      tmp_y[!is.na(tmp_y) & tmp_y < 0] <- abs(tmp_y[!is.na(tmp_y) & tmp_y < 0])
      mat[, y] <- tmp_y

      # ----------------------------
      # Normalizar por fila
      # ----------------------------
      row_totals <- rowSums(abs(mat), na.rm = TRUE)

      mat_rescaled <- mat
      idx <- row_totals > 0

      mat_rescaled[idx, ] <- mat[idx, , drop = FALSE] / row_totals[idx]

      directional_vector[["relative_importance"]] <- mat_rescaled
    }

    # normalize
    if (directional_vector[["baseline"]] == "mean") {

      L2_nrom_mean <- sqrt(sum(colMeans(data[, c(x,y)])^2))

      # normalize
      v <- (data[, c(x,y)]/sqrt(rowSums(data[, c(x,y)]^2)))*L2_nrom_mean

      v_w <- v * directional_vector[["relative_importance"]]

      # normalize
      L2_nrom_vector <- sqrt(rowSums(v_w[, c(x,y)]^2))
      v_final <- (v_w / L2_nrom_vector) * L2_nrom_mean


    } else if (directional_vector[["baseline"]] == "self") {

      L2_nrom_self <- sqrt(sum(data[, c(x,y)]^2))

      # normalize
      w <- directional_vector[["relative_importance"]]
      w <- w * data[, c(x,y)]
      L2_nrom_w <- sqrt(rowSums(w^2))

      w_nrom <- w/L2_nrom_w
      sqrt(rowSums(w_nrom^2))

      v_final <- w_nrom*L2_nrom_self

      # # directional vector
      # vector_gx <- as.data.frame(v_final[,x])
      # names(vector_gx) <- names_data[x]
      # vector_gy <- as.data.frame(v_final[,y])
      # names(vector_gy) <- names_data[y]

    } else if (directional_vector[["baseline"]] == "ones") {

        L2_nrom_one <- sqrt(sum(rep(1, length(c(x,y)))^2))

        # normalize
        w <- directional_vector[["relative_importance"]]
        L2_nrom_w <- sqrt(rowSums(directional_vector[["relative_importance"]]^2))

        w_nrom <- w/L2_nrom_w
        sqrt(rowSums(w_nrom^2))

        v_final <- w_nrom*L2_nrom_one

        # # directional vector
        # vector_gx <- as.data.frame(v_final[,x])
        # names(vector_gx) <- names_data[x]
        #
        # vector_gy <- as.data.frame(v_final[,y])
        # names(vector_gy) <- names_data[y]

    }

    # directional vector
    vector_gx <- as.data.frame(v_final[, x, drop = FALSE])
    names(vector_gx) <- names_data[x]
    vector_gx[is.na(vector_gx)] <- 0

    if (sign == FALSE) {
      vector_gx <- -vector_gx
    }

    vector_gy <- as.data.frame(v_final[, y, drop = FALSE])
    names(vector_gy) <- names_data[y]
    vector_gy[is.na(vector_gy)] <- 0

    if (sign == FALSE) {
      vector_gy <- abs(vector_gy)
    }

  } # end local directional vector


  # ----------------------------------------------------------------------------
  # Determining the max beta  --------------------------------------------------
  # ----------------------------------------------------------------------------

  # find the first approximation of max beta for max(efficiency_thresholds)
  find_beta_maxmin <- find_beta_maxmin(
    data = data,
    x = x,
    y = y,
    final_model = final_model,
    calibration_model = calibration_model,
    efficiency_thresholds = efficiency_thresholds,
    n_expand = n_expand,
    vector_gx = vector_gx,
    vector_gy = vector_gy,
    max_y = max_y,
    min_x = min_x
  )

  result_thresholds <- vector("list", length = length(efficiency_thresholds))
  names(result_thresholds) <- as.character(efficiency_thresholds)

  # for each threshold
  for (thr in efficiency_thresholds) {

    message(paste0("In progress: ", thr))

    data <- data[, c(x,y)]

    # save data points from scenario
    data_scenario <- as.data.frame(
      matrix(
        data = NA,
        ncol = length(c(x,y)),
        nrow = nrow(data)
      )
    )
    names(data_scenario) <- names(data)

    betas <- as.data.frame(matrix(
      data = NA,
      ncol = 2,
      nrow = nrow(data)
    ))

    variables <- c(x, y)

    # loop for each observation
    for (i in 1:nrow(data)) {

      # inicial prediction
      prediction_0 <- PEAXAI_predict(
        data = data[i,variables],
        x = x,
        y = y,
        final_model = final_model,
        calibration_model = calibration_model
      )

      # the DMU is more efficient then threshold?
      if (prediction_0 > thr) {
        betas[i, 1] <- 0
        betas[i, 2] <- thr

        data_scenario[i,] <- data[i,]

      } else {

        # Inicializar el rango inicial de 'y'
        range_beta <- as.matrix(
          seq(from = find_beta_maxmin[i, "min"], # find_beta_maxmin[i, "min"]
              to = find_beta_maxmin[i, "max"],
              length.out = n_grid))

        # Crear la matriz para aplicar predict()
        changes <- as.data.frame(matrix(
          data = NA,
          ncol = length(variables),
          nrow = length(range_beta)
        ))

        # Nombrar las columnas como en data original
        names(changes) <- names(data)

        if(directional_vector[["baseline"]] != "self" & is_global) {

          change_x <- matrix(
            data = rep(as.numeric(vector_gx), each = nrow(changes)),
            nrow = nrow(changes),
            ncol = length(vector_gx)
          )

          change_y <- matrix(
            data = rep(as.numeric(vector_gy), each = nrow(changes)),
            nrow = nrow(changes),
            ncol = length(vector_gy)
          )

        } else {

          change_x <- matrix(
            data = unlist(vector_gx[i,]),
            nrow = nrow(changes),
            ncol = length(vector_gx),
            byrow = TRUE
          )

          change_y <- matrix(
            data = unlist(vector_gy[i,]),
            nrow = nrow(changes),
            ncol = length(vector_gy)
          )

        }

        found_cut_off <- FALSE
        iter_count <- 0

        while (!found_cut_off) {

#           iter_count <- iter_count + 1
#
#           # matrix to apply changes
#           matrix_eff <- as.data.frame(matrix(
#             data = NA,
#             ncol = length(variables),
#             nrow = length(range_beta)
#           ))
#           names(matrix_eff) <- names(data)
#
#           # Asignar valores para 'x' y 'y'
#           matrix_eff[, x] <- data[i,x]
#           matrix_eff[, x] <- sweep(change_x, 1, range_beta, "*") + matrix_eff[, x]
#
#           matrix_eff[, y] <- data[i, y]
#           matrix_eff[, y] <- sweep(change_y, 1, range_beta, "*") + matrix_eff[, y]
#
#           # know if there are not possible values
#           mx <- as.matrix(matrix_eff[, x, drop = FALSE])
#           min_x_possible_vector <- as.numeric(min_x_possible*min_x)
#
#           my <- as.matrix(matrix_eff[, y, drop = FALSE])
#           max_y_possible_vector <- as.numeric(max_y_possible*max_y)
#
#           # sums how many TRUE lines are violating the restriction
#           viol <- rowSums(mx < rep(min_x_possible_vector, each = nrow(mx))) > 0
#           idx_viol_x <- which(viol)
#           viol <- rowSums(my > rep(max_y_possible_vector, each = nrow(my))) > 0
#           idx_viol_y <- which(viol)
#
#           viol <- rep(FALSE, nrow(mx))
#           viol[idx_viol_x] <- TRUE
#           viol[idx_viol_y] <- TRUE
#
#           keep <- !viol
#           if (any(!keep)) {
#             matrix_eff <- matrix_eff[keep, , drop = FALSE]
#             range_beta <- range_beta[keep, , drop = FALSE]
#           }
# browser()
#           # probability for each row
#           eff_vector <- apply(matrix_eff, 1, function(row) {
#
#             row_df <- as.data.frame(t(row))
#             colnames(row_df) <- names(data)
#
#             # pred <- unlist(predict(final_model, row_df, type = "prob")[1])
#             pred <- PEAXAI_predict(
#               data = matrix_eff,
#               x = x,
#               y = y,
#               final_model = final_model,
#               calibration_model = calibration_model
#             )
#
#             return(pred)
#
#           })
          #
          # # # Ensures that each position is at least the maximum value observed up to that point
          # eff_vector <- cummax(eff_vector)

          iter_count <- iter_count + 1
          # 1. Asegurar que range_beta es un vector simple (evita problemas de dimensiones)
          rb_vec <- as.numeric(range_beta)
          n_current <- length(rb_vec)
          # 2. Operaciones matriciales puras para los cambios
          # Multiplicamos la matriz de direcciones por el vector de betas
          matrix_eff_x <- change_x * rb_vec +
            matrix(as.numeric(data[i, x]), nrow = n_current, ncol = length(x), byrow = TRUE)

          matrix_eff_y <- change_y * rb_vec +
            matrix(as.numeric(data[i, y]), nrow = n_current, ncol = length(y), byrow = TRUE)
          # 3. Construcción de un data.frame completamente limpio ("matrix-free")
          # El lapply asegura que toda la "basura" de las matrices (nmatrix.1) desaparezca
          matrix_eff <- as.data.frame(cbind(matrix_eff_x, matrix_eff_y))
          matrix_eff[] <- lapply(matrix_eff, as.numeric)
          names(matrix_eff) <- names(data)[c(x, y)]

          # 4. Restricciones de dominio
          min_x_possible_vector <- as.numeric(min_x_possible * min_x)
          max_y_possible_vector <- as.numeric(max_y_possible * max_y)
          viol_x <- rowSums(sweep(matrix_eff_x, 2, min_x_possible_vector, "<")) > 0
          viol_y <- rowSums(sweep(matrix_eff_y, 2, max_y_possible_vector, ">")) > 0

          keep <- !(viol_x | viol_y)
          if (any(!keep)) {
            matrix_eff <- matrix_eff[keep, , drop = FALSE]
            rb_vec <- rb_vec[keep]
            range_beta <- as.matrix(rb_vec) # Mantener el formato matriz (n_grid x 1) por seguridad
          }
          if (nrow(matrix_eff) == 0) {
            # Si todos los puntos se salen del dominio, vector vacío
            eff_vector <- numeric(0)
          } else {
            # 5. Predicción vectorizada ultrarrápida
            eff_vector <- PEAXAI_predict(
              data = matrix_eff,
              x = x,
              y = y,
              final_model = final_model,
              calibration_model = calibration_model
            )
          }

          # Ensures that each position is at least the maximum value observed up to that point
          eff_vector <- cummax(eff_vector)

          # no changes case: min == max?
          if(round(eff_vector[1], precision_prob) == round(eff_vector[length(eff_vector)], precision_prob)) {

            data_scenario[i, x] <- data[i,x]
            data_scenario[i, y] <- data[i,y]

            betas[i, 1] <- 0
            betas[i, 2] <- eff_vector[1]
            break
          }

          # some problems?
          if (length(eff_vector) == 0 | is.null(eff_vector)) {

            data_scenario[i, x] <- rep(NA, ncol(matrix_eff[,x]))
            data_scenario[i, y] <- rep(NA, ncol(matrix_eff[,y]))

            betas[i, 1] <- NA
            betas[i, 2] <- NA
            break

          } else {
            # If there are not problems...

            # Did any prediction equal the target threshold?
            if (any(round(eff_vector, precision_prob) == round(thr, precision_prob))) {

              # Take the first matching index
              idx <- which(round(eff_vector, precision_prob) == round(thr, precision_prob))[1]
              found_cut_off <- TRUE

              # Store the 'x' and 'y' values that match the threshold
              data_scenario[i, x] <- matrix_eff[idx, x]
              data_scenario[i, y] <- matrix_eff[idx, y]

              betas[i, 1] <- range_beta[idx]
              betas[i, 2] <- thr
              break

            } else {

              # Find the interval where the target threshold lies
              pos <- which(head(eff_vector, -1) < thr & tail(eff_vector, -1) > thr)

              if (length(pos) > 1) {
                pos <- pos[1]
              } else if (length(pos) == 0) {
                idx_max <- which(eff_vector == max(eff_vector))
                pos <- idx_max[1]
              }

              if (is.na(pos)) {
                browser()
                break
              }

              # if position is in the maximum beta
              if (pos == length(eff_vector)) {

                # no more probability to be efficient
                # save best results
                data_scenario[i, x] <- matrix_eff[pos, x]
                data_scenario[i, y] <- matrix_eff[pos, y]

                betas[i, 1] <- range_beta[pos]

                pred <- as.data.frame(data_scenario[i,variables])
                names(pred) <- names(data)

                betas[i, 2] <- eff_vector[pos]
                break

              } else if (pos < 1) {
                browser()
                # if position is before the minimum beta
                pos <- 1

              }

              # Refine the 'y' interval between indices pos and pos + 1
              range_beta <- seq(from = range_beta[pos], to = range_beta[pos + 1], length.out = n_grid)

            }

          }

          if (range_beta[n_grid] - range_beta[1] <= 0.000001) {

            data_scenario[i, x] <- matrix_eff[(n_grid/2), x]
            data_scenario[i, y] <- matrix_eff[(n_grid/2), y]

            betas[i, 1] <- range_beta[(n_grid/2)]
            betas[i, 2] <- eff_vector[(n_grid/2)]

            # print("end while by dif")
            found_cut_off <- TRUE

          }

          if (iter_count == 20) {

            data_scenario[i, x] <- matrix_eff[(n_grid/2), x]
            data_scenario[i, y] <- matrix_eff[(n_grid/2), y]

            betas[i, 1] <- range_beta[(n_grid/2)]
            betas[i, 2] <- NA

            # print("end while by iter")
            found_cut_off <- TRUE

          }

        } # end while

      }

    } # end for

    names(betas) <- c("betas", "probability")

    result_thresholds[[as.character(thr)]][["counterfactual_dataset"]] <- data_scenario
    result_thresholds[[as.character(thr)]][["inefficiencies"]] <- betas
  }

  return(result_thresholds)

}

#' @title Search Range for Directional Efficiency Parameter (\eqn{\beta})
#'
#' @description
#' Estimates, for each observation, the minimum and maximum feasible values of the
#' directional distance parameter \eqn{\beta} used in projection-based efficiency
#' analysis. This function is an internal step of \code{\link{PEAXAI_counterfactuals}},
#' providing the initial search bounds for the iterative determination of efficiency targets.
#'
#' @param data A \code{data.frame} or \code{matrix} containing input and output variables.
#' @param x A numeric vector with the column indexes of input variables in \code{data}.
#' @param y A numeric vector with the column indexes of output variables in \code{data}.
#' @param final_model A fitted \pkg{caret} model of class \code{"train"} that supports
#'   \code{predict(type = "prob")} and returns a probability column for the efficient class.
#' @param calibration_model A model to calibrate.
#' @param efficiency_thresholds A numeric vector of probability levels in (0,1).
#'   Its minimum and maximum values delimit the target interval used to bracket \eqn{\beta}.
#' @param n_expand Integer. Increment step size applied to \eqn{\beta} at each iteration.
#' @param vector_gx A numeric vector or \code{data.frame} with directional changes for inputs
#'   (typically negative direction), usually built inside \code{PEAXAI_targets}.
#' @param vector_gy A numeric vector or \code{data.frame} with directional changes for outputs
#'   (positive direction).
#' @param max_y Numeric. Upper-limit multiplier for output expansion relative to observed maxima.
#' @param min_x Numeric. Lower-limit multiplier for input contraction relative to observed minima.
#'
#' @details
#' For each DMU, the function expands outputs and contracts inputs along the specified
#' direction until the predicted probability of efficiency (from \code{final_model})
#' reaches the maximum in \code{efficiency_thresholds} or feasible domain limits.
#' The resulting interval \eqn{[\beta_{\min}, \beta_{\max}]} is then used by
#' \code{\link{PEAXAI_counterfactuals}} to refine projections via grid search.
#'
#' @return
#' A \code{data.frame} with two numeric columns:
#' \describe{
#'   \item{\code{min}}{Minimum feasible value of \eqn{\beta} for each observation.}
#'   \item{\code{max}}{Maximum feasible value of \eqn{\beta} for each observation.}
#' }
#'
#' @seealso
#' \code{\link{PEAXAI_counterfactuals}} (efficiency projections based on \eqn{\beta});
#' \code{\link[caret]{train}} (model training with class probabilities).
#'
#' @export
#'

find_beta_maxmin <- function(
  data, x, y, final_model, calibration_model,
  efficiency_thresholds, n_expand, vector_gx,
  vector_gy, max_y, min_x
) {

  betas <- as.data.frame(matrix(
    data = NA,
    ncol = 2,
    nrow = nrow(data)
  ))

  variables <- c(x,y)

  max_efficiency_threshold <- max(efficiency_thresholds)
  min_efficiency_threshold <- min(efficiency_thresholds)

  # for each DMU, it will be calculated the max beta possible
  for (i in 1:nrow(data)) {

    # prediction_0 <- predict(final_model, data[i,variables], type = "prob")[1]
    prediction_0 <- PEAXAI_predict(
      data = data[i,variables],
      x = x,
      y = y,
      final_model = final_model,
      calibration_model = calibration_model
    )

    # the DMU is more efficient then threshold?
    if (prediction_0 > max_efficiency_threshold) {
      betas[i, 1] <- 0
      betas[i, 2] <- 0

    } else {

      control_threshold <- FALSE
      control_threshold_min <- FALSE

      names_data <- names(data[, c(x,y)])

      # control. Not min or max than observed resources
      min_x_i <- apply(as.matrix(data[,x]), 2, min) * min_x
      min_x_i <- as.data.frame(t(as.matrix(min_x_i)))
      names(min_x_i) <- names_data[x]

      max_y_i <- apply(as.matrix(data[,y]), 2, max) * max_y
      max_y_i <- as.data.frame(t(as.matrix(max_y_i)))
      names(max_y_i) <- names_data[y]

      manifold_check <- FALSE

      beta_j <- 0

      betas[i,1] <- NA

      prediction_j_max <- prediction_0

      while(control_threshold == FALSE) {

        beta_j <- beta_j + n_expand

        if (nrow(vector_gx) > 1) {
          new_x <- data[i,x] + beta_j * vector_gx[i,]
          new_y <- data[i,y] + beta_j * vector_gy[i,]
        } else {
          new_x <- data[i,x] + beta_j * vector_gx
          new_y <- data[i,y] + beta_j * vector_gy
        }

        new_point <- cbind(new_x, new_y)
        names(new_point) <- names(data[,variables])

        # prediction_j <- predict(final_model, new_point, type = "prob")[1]

        # ### new
        # library(dbscan)
        #
        # # domain data
        # manifold <- as.matrix(data[,c(x,y)])
        #
        # # 3. Combinar para calcular
        # data_plus_new_unit <- rbind(manifold, new_point)
        #
        # # 4. Calcular LOF (k debe ser similar al número de vecinos que esperarías)
        # lof_scores <- lof(data_plus_new_unit, minPts = length(c(x,y))*2)
        # if (tail(lof_scores, 1) > lof_scores[i] & lof_scores[i] > 2) {
        #   manifold_check <- TRUE
        # }
        # ###

        # ###
        # # domain data
        # manifold <- as.matrix(data[,c(x,y)])
        # # daisy() maneja automáticamente variables numéricas y categóricas
        # distances <- daisy(rbind(manifold, new_point), metric = "gower")
        #
        # # Convertimos a matriz y extraemos la última fila (nuestra distancia con el resto)
        # dist_matrix <- as.matrix(distances)
        # dist_cf <- dist_matrix[nrow(dist_matrix), 1:(nrow(dist_matrix)-1)]
        #
        # # El vecino más cercano según Gower
        # min_gower <- min(dist_cf)
        #
        # if (min_gower > 0.15) {
        #   manifold_check <- TRUE
        # }
        # ###

        prediction_j <- PEAXAI_predict(
          data = new_point,
          x = x,
          y = y,
          final_model = final_model,
          calibration_model = calibration_model
        )

        # Check if the probability has decreased. The probability function should be monotonic.
        if (prediction_j < prediction_j_max) {
          prediction_j <- prediction_j_max
        } else {
          prediction_j_max <- prediction_j
        }

        if (control_threshold_min == FALSE & prediction_j > min_efficiency_threshold) {

          betas[i, 1] <- beta_j - n_expand
          control_threshold_min <- TRUE

        }

        if(prediction_j >= max_efficiency_threshold) {

          # break the while loop
          betas[i, 2] <- beta_j
          control_threshold <- TRUE

        } else if (any(new_point[x] < min_x_i) || any(new_point[y] > max_y_i) || manifold_check == TRUE) {
          # break the while loop
          betas[i, 2] <- beta_j
          control_threshold <- TRUE
        }

      }

      if (is.na(betas[i,1])) {
        betas[i, 1] <- 0
      }

    }

  }

  names(betas) <- c("min", "max")

  return(betas)
}

#' @title Predict Probability of Efficiency Using a Fitted Model
#'
#' @description
#' Predicts probabilities for new decision-making units (DMUs) using a fitted
#' \pkg{caret} classification model. If \code{calibration_model} is provided,
#' the raw classifier probabilities are post-processed to obtain calibrated
#' probability estimates (e.g., Platt scaling via logistic regression or
#' isotonic calibration via a monotone mapping).
#'
#' @param data A \code{data.frame} or \code{matrix} containing the variables used
#'   for prediction.
#' @param x Integer vector indicating the column indices of input variables in
#'   \code{data}.
#' @param y Integer vector indicating the column indices of output variables in
#'   \code{data}.
#' @param final_model A fitted \pkg{caret} model provided by PEAXAI_fitting().
#' @param calibration_model Optional calibration object returned by \code{PEAXAI_fitting()}.
#'   If \code{NULL}, raw probabilities are returned. If \code{calibration_model$method == "glm.fit"},
#'   Platt scaling is applied by predicting \code{type = "response"} from a logistic regression
#'   using the raw score \code{s}. Otherwise, an isotonic calibration function is retrieved via
#'   \code{calibration_model[[final_model$method]]} and applied to the raw probabilities.
#'
#' @return A numeric vector of predicted probabilities (raw or calibrated), one per
#'   row of \code{data}.
#'
#' @export

PEAXAI_predict <- function (
    data, x, y, final_model, calibration_model = NULL
) {

  data <- as.data.frame(data)

  # reorder index 'x' and 'y' in data
  data <- data[, c(x,y)]
  x <- 1:(ncol(data) - length(y))
  y <- (length(x) + 1):ncol(data)

  names_data <- names(data[,c(x,y)])

  name_model <- final_model$method

  if(is.null(calibration_model)) {

    # --------------------------------------------------------------------------
    # No calibration -----------------------------------------------------------
    # --------------------------------------------------------------------------
    predictions <- predict(final_model, newdata = data, type = "prob")[,1]

  } else {

    # --------------------------------------------------------------------------
    # Calibration --------------------------------------------------------------
    # --------------------------------------------------------------------------

    # Original probabilities
    s_new <- predict(final_model,
                     newdata = data,
                     type = "prob")[,1]

    if (calibration_model$method == "glm.fit") {

      # platt
      predictions <- predict(calibration_model,
                             newdata = data.frame(s = s_new),
                             type = "response")
    } else {

      # isotonic
      predictions <- calibration_model[[name_model]](s_new)

    }

  }

  return(predictions)
}

---
title: "PEAXAI: Probabilistic Efficiency Analysis using Explainable Artificial Intelligence"
author: "Institute 'Center of Operations Research', Miguel Hernández University of Elche"
output:
  rmarkdown::html_vignette:
    toc: true
    number_sections: false
vignette: >
  %\VignetteIndexEntry{PEAXAI: Probabilistic Efficiency Analysis using Explainable Artificial Intelligence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

<style>
  body { text-align: justify; }
</style>

```{r setup, include=FALSE}
# Enable a "fast mode" during R CMD check to keep CRAN timings under control
is_check <- ("CheckExEnv" %in% search()) || any(c("_R_CHECK_TIMINGS_","_R_CHECK_LICENSE_") %in% names(Sys.getenv()))

# Global knitr options for compact output and consistent figures
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE, message = FALSE,
  fig.width = 7, fig.height = 4.8, dpi = 96
)

# Seed for reproducibility
set.seed(1)
```

This vignette introduces the main functions of the `PEAXAI` package.

`PEAXAI` provides a methodology that integrates Data Envelopment Analysis (DEA) with Machine Learning (ML) algorithms to train a classifier that estimates the probability that each decision-making unit (DMU) is (in)efficient.

Once the model is trained, the package offers several functions for post-hoc analysis, including:

- Estimating feature importance,
- Generating counterfactual scenarios for each DMU — that is, determining the minimum changes required for a DMU to be considered efficient along a predefined direction vector and based on a user-defined efficiency threshold,
- Producing efficiency rankings (either based on the predicted probabilities or on attainable improvements), and
- Identifying peer units given a specific efficiency threshold.

These tools facilitate benchmarking and policy simulations in a transparent and explainable framework.

## Example Dataset: Spanish Food Industry Firms

To illustrate the functionality of the `PEAXAI` package, we use a dataset of 917 companies in the food industry sector operating in Spain. The data were retrieved from the SABI (Sistema de Análisis de Balances Ibéricos) database for the year 2023 and include only firms with more than 50 employees.

Each row corresponds to a single firm and includes both financial and operational variables relevant for productivity and efficiency analysis.

The output variable used to measure performance is:

- `operating_income`: Operating income in millions of euros.

The input variables include:

- `total_assets`: Total assets (in millions of euros).
- `employees`: Number of employees.
- `fixed_assets`: Tangible fixed assets (in millions of euros).
- `personnel_expenses`: Personnel-related costs (in millions of euros).

Additionally, the variable `autonomous_community` indicates the geographical location of each firm within one of Spain’s 17 autonomous communities. This allows the analysis to reflect regional institutional and market heterogeneity.

The dataset exhibits a wide dispersion across firms, including both medium-sized and large enterprises. This heterogeneity poses a realistic challenge for evaluating efficiency and provides a rich testing ground for model explainability.


## Load the package and example dataset
For illustration purposes, this vignette focuses on a subset of the data: firms located in the Autonomous Community of Valencia (Comunidad Valenciana).

```{r}
library(PEAXAI)
data("firms", package = "PEAXAI")
```

For illustration, we focus on firms located in the Comunidad Valenciana (Autonomous Community of Valencia).
Filter the dataset to include only firms from the Comunidad Valenciana and remove the `autonomous_community` column.
```{r}
data <- subset(firms, autonomous_community == "Comunidad Valenciana")[, -ncol(firms)]
rm(firms)
```

Quick structure and summary.
```{r}
str(data)
summary(data)
```

## Specify parameters for PEAXAI model fitting
Determine inputs (x) and outputs (y) indices.
```{r}
x <- c(1:4)
y <- c(5)
```

Determine the technology assumptions.
```{r}
RTS <- "vrs"
```

To address class imbalance, we apply SMOTE based on the best-practice frontier (assuming a convex production technology). First, we identify the combinations of DMUs that define the frontier. Then, we generate one of two types of synthetic DMUs from these combinations: efficient units (on the frontier) and near-frontier inefficient units (inside the frontier’s envelope).

When the efficient class is the minority, efficient-class SMOTE on the best-practice frontier helps the classifier learn the frontier more accurately. Optionally, if the observed imbalance still falls short of the target ratio, we also generate near-frontier inefficient synthetic units derived from the frontier geometry to sharpen the separation between efficient and inefficient regions.
```{r}
imbalance_rate <- seq(0.05, 0.4, 0.05) 
```

Define machine learning methods and their tuning parameters. Here we configure a neural network (nnet) with a grid of hyperparameters
```{r}
methods <- list(
  "nnet" = list(
    tuneGrid = expand.grid(
      size = c(1, 5, 10, 20),
      decay = 10^seq(-5, -1, by = 1)
    ),
    maxit = 100,
    preProcess = c("center", "scale"),
    # # --- arguments nnet ---
    entropy = TRUE,
    skip = TRUE,
    maxit = 1000,
    MaxNWts = 100000,
    trace = FALSE,
    weights = NULL
  )
)
```

Parameters for controlling model training (k-fold cross-validation)
```{r}
trControl <- list(
  method = "cv",
  number = 5
)
```

Classification metrics to optimize during training (priority order in case of ties)
```{r}
metric_priority <- c("Balanced_Accuracy", "F1", "ROC_AUC")
```

Define hold-out sample to evaluate model performance on unseen data. These samples are excluded from the cross-validation and hyperparameter tuning process. By default, is `NULL`.
```{r}
hold_out <- NULL
```

Set random seed for reproducibility (ensures consistent results across runs).
```{r}
seed <- 1
```

## Apply PEAXAI: from efficiency labeling to ML-based classification

We now apply the full PEAXAI pipeline:

1. Estimate efficiency labels using Additive DEA,
2. Balance the data using SMOTE units,
3. Train the machine learning classifier with cross-validation,
4. Select the best model according to multiple performance metrics.
 
```{r echo=TRUE, message=FALSE, warning=FALSE}
models <- PEAXAI_fitting(
  data = data,
  x = x,
  y = y,
  RTS = RTS,
  imbalance_rate = imbalance_rate,
  methods = methods,
  trControl = trControl,
  metric_priority = metric_priority,
  hold_out = hold_out,
  verbose = TRUE,
  seed = seed
)
```

```{r echo=FALSE}
models$best_model_fit
```

```{r echo=FALSE}
models$performance_train
```


## XAI method for global importance
This example uses SHAP (SHapley Additive exPlanations) computed with the fastshap package to estimate global feature importance with `PEAXAI_global_importance`. We approximate Shapley values via Monte Carlo sampling (`nsim`) and obtain a global importance score by aggregating absolute SHAP values across observations (and then normalizing to relative importance). For alternative XAI methods and options, see the package documentation.
```{r}
importance_method <- list(
    name = "SHAP",
    nsim = 200
)
```

## Apply XAI method on the model fine-tuned.
If the method needs a reference distribution or to train a surrogate, use `background` to choose that dataset.
`target` selects the dataset on which explanations are computed.

If `target` = `train`, relative importances are computed on the training distribution and thus reflect what the model actually learned during fitting.

If `target` = `real`, relative importances are computed on a dataset intended to approximate the real-world (deployment) distribution, yielding importances that better represent the underlying problem as observed in practice.

```{r}
relative_importance <- PEAXAI_global_importance(
  final_model = models[["best_model_fit"]][["nnet"]],
  x = x,
  y = y,
  explain_data = data,
  reference_data = data,
  importance_method = importance_method
)
```

```{r echo=FALSE}
relative_importance
```

## Selected efficiency thresholds
We evaluate results at three efficiency cutoffs: 0.75, 0.85, and 0.95.
```{r}
efficiency_thresholds <- seq(0.75, 0.95, 0.1) 
```

## Directional vector for target setting
We define a global, model-aware direction for improving inputs/outputs.

`relative_importance` guides the direction using the model’s global importance.

`scope` = `global` applies a single direction to all DMUs.

The argument `relative_importance` is the direction to make the counterfactual analysis. The directional vector can be custom or provided by PEAXAI_global_importance funtion.

For example, if the directional vector is custom:
If it not possible (or we do not want) to change a specific input (like `employees`), we need to write:
```{r}
relative_importance_custom <- t(matrix(
  data = c(0.2, 0, 0.2, 0.2, 0.4),
))
relative_importance_custom <- as.data.frame(relative_importance_custom)
names(relative_importance_custom) <- names(data)[c(x,y)]
```

```{r echo=FALSE}
relative_importance_custom
```

But, if we do not know which direction define, we can use the relative importace by the model fitted:
```{r}
directional_vector <- list(
  relative_importance = relative_importance,
  baseline  = "mean"        
)
```

## Compute targets at selected thresholds
Given the thresholds and directional vector, we compute feasible improvement targets for each DMU.
Key controls:

`n_expand` = 0.5: expansion factor for the search neighborhood.

`n_grid` = 50: grid resolution for exploring adjustments.

`max_y` = 2, `min_x` = 1: bounds that cap output expansion and input contraction.

```{r}
targets <- PEAXAI_counterfactuals(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  efficiency_thresholds = efficiency_thresholds,
  directional_vector = directional_vector,
  n_expand = 0.25,
  n_grid = 50,
  max_y = 1,
  min_x = 1
)
```

```{r}
head(targets[["0.85"]][["counterfactual_dataset"]], 10)
```

```{r}
head(targets[["0.85"]][["inefficiencies"]], 10)
```

## Efficiency rankings
We compute efficiency rankings at each threshold using two bases:

`Predicted`: ranks by the model’s predicted probability of being efficient (`rank_basis` = `predicted`).

`Attainable`: ranks by attainable improvements/targets implied by the model (`rank_basis` = `attainable`).

```{r}
ranking <- PEAXAI_ranking(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  rank_basis = "predicted"
)

ranking2 <- PEAXAI_ranking(
    data = data,
    x = x,
    y = y,
    final_model = models[["best_model_fit"]][["nnet"]],
    efficiency_thresholds = efficiency_thresholds,
    targets = targets,
    rank_basis = "attainable"
  )
```

```{r}
head(round(ranking, 4), 50)
```



```{r}
head(round(ranking2[["0.85"]], 4), 50)
```

## Peers by efficiency thresholds
For each threshold, we identify peer DMUs that serve as reference comparators.

`weighted` = FALSE treats peers uniformly; set TRUE to weight importance given `relative_importance`.
```{r}
peers <- PEAXAI_peer(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  targets = targets,
  efficiency_thresholds = efficiency_thresholds,
  weighted = FALSE,
  relative_importance = relative_importance
)
```

```{r}
head(peers, 50)
```

---
title: "PEAXAI: Probabilistic Efficiency Analysis using Explainable Artificial Intelligence"
author: "Institute 'Center of Operations Research', Miguel Hernández University of Elche"
output:
  rmarkdown::html_vignette:
    toc: true
    number_sections: false
vignette: >
  %\VignetteIndexEntry{PEAXAI: Probabilistic Efficiency Analysis using Explainable Artificial Intelligence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

<style>
  body { text-align: justify; }
</style>

```{r setup, include=FALSE}
# Enable a "fast mode" during R CMD check to keep CRAN timings under control
is_check <- ("CheckExEnv" %in% search()) || any(c("_R_CHECK_TIMINGS_","_R_CHECK_LICENSE_") %in% names(Sys.getenv()))

# Global knitr options for compact output and consistent figures
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE, message = FALSE,
  fig.width = 7, fig.height = 4.8, dpi = 96
)

# Seed for reproducibility
set.seed(1)
```

This vignette introduces the main functions of the `PEAXAI` package.

`PEAXAI` provides a methodology that integrates Data Envelopment Analysis (DEA) with Machine Learning (ML) algorithms to train a classifier that estimates the probability that each decision-making unit (DMU) is (in)efficient.

Once the model is trained, the package offers several functions for post-hoc analysis, including:

- Estimating feature importance,
- Generating counterfactual scenarios for each DMU — that is, determining the minimum changes required for a DMU to be considered efficient along a predefined direction vector and based on a user-defined efficiency threshold,
- Producing efficiency rankings (either based on the predicted probabilities or on attainable improvements), and
- Identifying peer units given a specific efficiency threshold.

These tools facilitate benchmarking and policy simulations in a transparent and explainable framework.

## Example Dataset: Spanish Food Industry Firms

To illustrate the functionality of the `PEAXAI` package, we use a dataset of 917 companies in the food industry sector operating in Spain. The data were retrieved from the SABI (Sistema de Análisis de Balances Ibéricos) database for the year 2023 and include only firms with more than 50 employees.

Each row corresponds to a single firm and includes both financial and operational variables relevant for productivity and efficiency analysis.

The output variable used to measure performance is:

- `operating_income`: Operating income in millions of euros.

The input variables include:

- `total_assets`: Total assets (in millions of euros).
- `employees`: Number of employees.
- `fixed_assets`: Tangible fixed assets (in millions of euros).
- `personnel_expenses`: Personnel-related costs (in millions of euros).

Additionally, the variable `autonomous_community` indicates the geographical location of each firm within one of Spain’s 17 autonomous communities and 2 autonomous cities. This allows the analysis to reflect regional institutional and market heterogeneity.

The dataset exhibits a wide dispersion across firms, including both medium-sized and large enterprises. This heterogeneity poses a realistic challenge for evaluating efficiency and provides a rich testing ground for model explainability.


## Load the package and example dataset
```{r}
library(PEAXAI)
data("firms", package = "PEAXAI")
```

For illustration purposes, this vignette focuses on a subset of the data: the 97 firms located in the Autonomous Community of Valencia (Comunidad Valenciana). We drop the `autonomous_community` column, so that the remaining columns are exactly the four inputs and the output.
```{r}
data <- subset(
  firms,
  autonomous_community == "Comunidad Valenciana",
  select = -autonomous_community
)
rm(firms)
```

Quick structure and summary.
```{r}
str(data)
summary(data)
```

## Specify parameters for PEAXAI model fitting
Determine inputs (x) and outputs (y) indices.
```{r}
x <- c(1:4)
y <- c(5)
```

Determine the technology assumptions.
```{r}
RTS <- "vrs"
```

To address class imbalance, we apply SMOTE based on the best-practice frontier (assuming a convex production technology). First, we identify the combinations of DMUs that define the frontier. Then, we generate one of two types of synthetic DMUs from these combinations: efficient units (on the frontier) and near-frontier inefficient units (inside the frontier’s envelope).

When the efficient class is the minority, efficient-class SMOTE on the best-practice frontier helps the classifier learn the frontier more accurately. Optionally, if the observed imbalance still falls short of the target ratio, we also generate near-frontier inefficient synthetic units derived from the frontier geometry to sharpen the separation between efficient and inefficient regions.
```{r}
imbalance_rate <- seq(0.05, 0.4, 0.05) 
```

Define machine learning methods and their tuning parameters. Here we configure a neural network (nnet) with a grid of hyperparameters
```{r}
methods <- list(
  "nnet" = list(
    tuneGrid = expand.grid(
      size = c(1, 5, 10, 20),
      decay = 10^seq(-5, -1, by = 1)
    ),
    preProcess = c("center", "scale"),
    # --- arguments passed on to nnet ---
    skip = TRUE,
    maxit = 100,
    MaxNWts = 100000,
    trace = FALSE
  )
)
```

Parameters for controlling model training (k-fold cross-validation)
```{r}
trControl <- list(
  method = "cv",
  number = 5
)
```

Classification metrics to optimize during training (priority order in case of ties)
```{r}
metric_priority <- c("Balanced_Accuracy", "F1", "ROC_AUC")
```

Define hold-out sample to evaluate model performance on unseen data. These samples are excluded from the cross-validation and hyperparameter tuning process. By default, is `NULL`.
```{r}
hold_out <- NULL
```

Set random seed for reproducibility (ensures consistent results across runs).
```{r}
seed <- 1
```

## Apply PEAXAI: from efficiency labeling to ML-based classification

We now apply the full PEAXAI pipeline:

1. Estimate efficiency labels using Additive DEA,
2. Balance the data using SMOTE units,
3. Train the machine learning classifier with cross-validation,
4. Select the best model according to multiple performance metrics.
 
```{r echo=TRUE, message=FALSE, warning=FALSE}
models <- PEAXAI_fitting(
  data = data,
  x = x,
  y = y,
  RTS = RTS,
  imbalance_rate = imbalance_rate,
  methods = methods,
  trControl = trControl,
  metric_priority = metric_priority,
  hold_out = hold_out,
  verbose = TRUE,
  seed = seed
)
```

```{r echo=FALSE}
models$best_model_fit
```

```{r echo=FALSE}
models$performance_train
```


## XAI method for global importance
This example uses SHAP (SHapley Additive exPlanations) computed with the kernelshap package to estimate global feature importance with `PEAXAI_global_importance`. We approximate Shapley values via a background sample (`bg_n`) and obtain a global importance score by aggregating absolute SHAP values across observations (and then normalizing to relative importance). When `bg_n` is not specified, a background sample of 200 observations is used. For alternative XAI methods and options, see the package documentation.
```{r}
importance_method <- list(
    name = "SHAP"
)
```

## Apply XAI method on the model fine-tuned.
Two datasets are supplied.

`explain_data` holds the DMUs on which the explanations are computed. Passing the observed sample, as we do here, yields importances that represent the underlying problem as it is observed in practice.

`reference_data` is the distribution the method perturbs against, that is, the background sample for SHAP or the reference distribution for a surrogate model. Passing `final_model$trainingData` reflects what the model actually learned during fitting, including the synthetic DMUs generated by SMOTE; passing the observed sample restricts the background to real DMUs.

Both datasets must carry the variables the model was trained on under their exact names.

```{r}
relative_importance <- PEAXAI_global_importance(
  final_model = models[["best_model_fit"]][["nnet"]],
  x = x,
  y = y,
  explain_data = data,
  reference_data = data,
  importance_method = importance_method
)
```

```{r echo=FALSE}
relative_importance
```

## Selected efficiency thresholds
We evaluate results at three efficiency cutoffs: 0.75, 0.85, and 0.95.
```{r}
efficiency_thresholds <- seq(0.75, 0.95, 0.1) 
```

## Directional vector for target setting
We define a global, model-aware direction for improving inputs/outputs.

`relative_importance` is the direction along which the counterfactual analysis moves each DMU. It can be supplied by the user or taken from `PEAXAI_global_importance()`, which returns a single row and therefore applies one direction to every DMU. A matrix with one row per DMU, such as the one returned by `PEAXAI_local_importance()`, applies a DMU-specific direction instead.

`baseline` sets the unit of that direction: `"mean"`, `"median"`, `"self"` or `"ones"`.

For example, if the directional vector is custom:
If it not possible (or we do not want) to change a specific input (like `employees`), we need to write:
```{r}
relative_importance_custom <- t(matrix(
  data = c(0.2, 0, 0.2, 0.2, 0.4),
))
relative_importance_custom <- as.data.frame(relative_importance_custom)
names(relative_importance_custom) <- names(data)[c(x,y)]
```

```{r echo=FALSE}
relative_importance_custom
```

But, if we do not know which direction define, we can use the relative importace by the model fitted:
```{r}
directional_vector <- list(
  relative_importance = relative_importance,
  baseline  = "mean"        
)
```

## Compute targets at selected thresholds
Given the thresholds and directional vector, we compute feasible improvement targets for each DMU.
Key controls:

`n_expand` = 0.25: expansion factor for the search neighborhood.

`n_grid` = 50: grid resolution for exploring adjustments.

`max_y` = 1, `min_x` = 1: bounds that cap output expansion and input contraction.

```{r}
targets <- PEAXAI_counterfactuals(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  efficiency_thresholds = efficiency_thresholds,
  directional_vector = directional_vector,
  n_expand = 0.25,
  n_grid = 50,
  max_y = 1,
  min_x = 1
)
```

```{r}
head(targets[["0.85"]][["counterfactual_dataset"]], 10)
```

```{r}
head(targets[["0.85"]][["inefficiencies"]], 10)
```

## Efficiency rankings
We compute efficiency rankings at each threshold using two bases:

`Predicted`: ranks by the model’s predicted probability of being efficient (`rank_basis` = `predicted`).

`Attainable`: ranks by attainable improvements/targets implied by the model (`rank_basis` = `attainable`).

```{r}
ranking <- PEAXAI_ranking(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  rank_basis = "predicted"
)

ranking2 <- PEAXAI_ranking(
    data = data,
    x = x,
    y = y,
    final_model = models[["best_model_fit"]][["nnet"]],
    efficiency_thresholds = efficiency_thresholds,
    targets = targets,
    rank_basis = "attainable"
  )
```

```{r}
head(round(ranking, 4), 50)
```



```{r}
head(round(ranking2[["0.85"]], 4), 50)
```

## Peers by efficiency thresholds
For each threshold, we identify peer DMUs that serve as reference comparators.

`weighted` = FALSE treats peers uniformly; set TRUE to weight importance given `relative_importance`.
```{r}
peers <- PEAXAI_peer(
  data = data,
  x = x,
  y = y,
  final_model = models[["best_model_fit"]][["nnet"]],
  targets = targets,
  efficiency_thresholds = efficiency_thresholds,
  weighted = FALSE,
  relative_importance = relative_importance
)
```

```{r}
head(peers, 50)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label_efficiency.R
\name{label_efficiency}
\alias{label_efficiency}
\title{Data preprocessing and efficiency labeling with Additive DEA}
\usage{
label_efficiency(
  data,
  REF = data,
  x,
  y,
  z_numeric = NULL,
  z_factor = NULL,
  RTS = "vrs",
  B = 1,
  alpha = FALSE,
  m = NULL,
  bandwidth = NULL,
  seed
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing all variables.}

\item{REF}{Optional reference set of inputs that defines the technology
(defaults to the columns indicated by \code{x} in \code{data}). Must have
the same number of rows as \code{data}.}

\item{x}{Integer vector with column indices of input variables in \code{data}.}

\item{y}{Integer vector with column indices of output variables in \code{data}.}

\item{z_numeric}{Integer vector with column indices of continuous/numeric exogenous variables in \code{data}. By default is \code{NULL}.}

\item{z_factor}{Integer vector with column indices of discrete/factor exogenous variables in \code{data}. By default is \code{NULL}.}

\item{RTS}{Character or integer specifying the DEA technology / returns-to-scale
assumption (default: \code{"vrs"}). Accepted values:
\describe{
\item{\code{0} / \code{"fdh"}}{Free disposability hull (no convexity).}
\item{\code{1} / \code{"vrs"}}{Variable returns to scale (convexity + free disposability).}
\item{\code{2} / \code{"drs"}}{Decreasing returns to scale (convexity, down-scaling, free disposability).}
\item{\code{3} / \code{"crs"}}{Constant returns to scale (convexity + free disposability).}
\item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling only, convexity + free disposability).}
\item{\code{5} / \code{"add"}}{Additivity (integer up/down scaling) with free disposability.}
}}

\item{B}{Integer. The number of bootstrap replicates (iterations) in the conditional DEA framework.}

\item{alpha}{Numeric. The nominal size of the confidence interval (e.g., \code{0.05}). If \code{FALSE}, no confidence intervals are computed.}

\item{m}{Integer. The number of units to be drawn to form the reference set in each bootstrap replicate of the conditional DEA.}

\item{bandwidth}{Optionally, the bandwidth parameters for the unconditional kernel density estimator used in the conditional DEA framework. Typical estimation is done using \code{\link[np]{npudensbw}} which supports mixed continuous and discrete data types.}

\item{seed}{Integer or \code{NULL}. Seed for pseudo-random number generator ensuring reproducibility in the bootstrap sampling process.}
}
\value{
A \code{data.frame} equal to \code{data} (retaining all input \code{x} and
output \code{y} columns) plus a new factor column \code{class_efficiency}
with levels \code{c("not_efficient","efficient")}.
}
\description{
Labels each DMU (Decision Making Unit) as efficient or not using the
Additive DEA model, optionally after basic data preprocessing. It supports both
standard unconditional DEA and conditional DEA frameworks (when exogenous variables
are provided). The resulting factor \code{class_efficiency} has levels
\code{c("not_efficient","efficient")}, where \code{"efficient"} is the positive
class for downstream modeling.
}
\details{
Internally relies on \code{\link[Benchmarking]{dea.add}} to compute Additive DEA
scores and derive the binary efficiency label. When exogenous variables \code{z_numeric}
or \code{z_factor} are provided, a conditional DEA is performed by sampling a reference
set of size \code{m} internally \code{B} times, based on distances computed via kernel
density estimation.
}
\examples{
\donttest{
# Example (assuming columns 1:2 are inputs and 3 is output):
# out <- label_efficiency(data = df, x = 1:2, y = 3, RTS = "vrs")
# table(out$class_efficiency)
}

}
\seealso{
\code{\link[Benchmarking]{dea.add}}, \code{\link[rcDEA]{conditional_DEA}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_ranking.R
\name{PEAXAI_ranking}
\alias{PEAXAI_ranking}
\title{Generate Efficiency Rankings Based on Probabilistic Classification}
\usage{
PEAXAI_ranking(
  data,
  x,
  y,
  final_model,
  calibration_model = NULL,
  efficiency_thresholds,
  targets = NULL,
  rank_basis
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the input and output variables.}

\item{x}{Integer vector specifying the column indices of input variables in \code{data}.}

\item{y}{Integer vector specifying the column indices of output variables in \code{data}.}

\item{final_model}{A fitted classification model used to estimate efficiency probabilities.
Supported types are:
\itemize{
\item \code{"train"}: an object fitted with \pkg{caret}.
\item \code{"glm"}: a binomial logistic regression model.
}}

\item{calibration_model}{Optional probability-calibration model applied to the raw
predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
If provided, calibrated probabilities are used for ranking and threshold-based decisions.
Set to \code{NULL} to use uncalibrated predictions.}

\item{efficiency_thresholds}{Numeric vector defining one or more efficiency probability
thresholds to determine the attainable frontier or peer set.}

\item{targets}{A named list containing, for each efficiency threshold, the corresponding
attainable targets and estimated \eqn{\beta} values (e.g., obtained from counterfactual analysis).
Each element should be a list with a component named \code{"beta"}.}

\item{rank_basis}{Character string specifying the ranking criterion. Options are:
\itemize{
\item \code{"predicted"}: order units by predicted efficiency probability.
\item \code{"attainable"}: order by attainable probability, then by \eqn{\beta},
and finally by predicted probability (see Details).
}}
}
\value{
\itemize{
\item If \code{rank_basis = "predicted"}: a \code{data.frame} sorted by predicted efficiency probability.
\item If \code{rank_basis = "attainable"}: a named list of \code{data.frame}s, one per efficiency threshold,
each sorted according to the hierarchical ranking scheme described above.
}
}
\description{
Produces efficiency rankings of decision-making units (DMUs) according to the probabilities
estimated by a fitted classification model. Two ranking modes are supported:
\itemize{
\item \code{"predicted"}: ranks DMUs solely by their predicted probability of being efficient.
\item \code{"attainable"}: ranks DMUs hierarchically according to:
(1) the attainable (target) efficiency probability,
(2) the size of the improvement parameter \eqn{\beta} (smaller is better),
and (3) the predicted efficiency probability (higher is better).
}
This allows to integrate both predictive and counterfactual (attainable) information
into the efficiency ranking.
}
\details{
The attainable-based ranking combines predictive efficiency with the modeled potential
for improvement (\eqn{\beta}) and the probability of reaching a target frontier level.
This approach yields a more nuanced and interpretable prioritization of DMUs, reflecting
both their current and achievable performance under the estimated model.

When \code{rank_basis = "attainable"}, ties in attainable probability are broken first
by the magnitude of \eqn{\beta} (ascending), and then by the predicted probability
(descending).
}
\examples{
\donttest{
  data("firms", package = "PEAXAI")

  data <- subset(
    firms,
    autonomous_community == "Comunidad Valenciana"
  )

  x <- 1:4
  y <- 5
  RTS <- "vrs"
  imbalance_rate <- NULL

  trControl <- list(
    method = "cv",
    number = 3
  )

  # glm method
  methods <- list(
    "glm" = list(
      weights = "dinamic"
     )
  )

  metric_priority <- c("Balanced_Accuracy", "ROC_AUC")

  models <- PEAXAI_fitting(
    data = data, x = x, y = y, RTS = RTS,
    imbalance_rate = imbalance_rate,
    methods = methods,
    trControl = trControl,
    metric_priority = metric_priority,
    verbose = FALSE,
    seed = 1
  )

  final_model <- models[["best_model_fit"]][["glm"]]

  importance_method <- list(name = "PI", n.repetitions = 5)

  relative_importance <- PEAXAI_global_importance(
    final_model = final_model,
    x = x,
    y = y,
    explain_data = data,
    reference_data = final_model$trainingData,
    importance_method = importance_method,
    seed = 1
  )

  efficiency_thresholds <- seq(0.75, 0.95, 0.1)

  directional_vector <- list(
    relative_importance = relative_importance,
    baseline  = "mean"
  )

  targets <- PEAXAI_counterfactuals(
    data = data,
    x = x, y = y,
    final_model = final_model,
    efficiency_thresholds = efficiency_thresholds,
    directional_vector = directional_vector,
    n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
  )

  ranking <- PEAXAI_ranking(data = data, x = x, y = y,
  final_model = final_model, rank_basis = "predicted")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{preprocessing}
\alias{preprocessing}
\title{Prepare Data and Handle Errors}
\usage{
preprocessing(data, x, y)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the variables in the model.}

\item{x}{Column indexes of input variables in \code{data}.}

\item{y}{Column indexes of output variables in \code{data}.}
}
\value{
It returns a \code{matrix} in the required format and displays some error messages.
}
\description{
This function arranges the data in the required format and displays some error messages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_counterfactuals.R
\name{find_beta_maxmin}
\alias{find_beta_maxmin}
\title{Search Range for Directional Efficiency Parameter (\eqn{\beta})}
\usage{
find_beta_maxmin(
  data,
  x,
  y,
  final_model,
  calibration_model,
  efficiency_thresholds,
  n_expand,
  vector_gx,
  vector_gy,
  max_y,
  min_x
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing input and output variables.}

\item{x}{A numeric vector with the column indexes of input variables in \code{data}.}

\item{y}{A numeric vector with the column indexes of output variables in \code{data}.}

\item{final_model}{A fitted \pkg{caret} model of class \code{"train"} that supports
\code{predict(type = "prob")} and returns a probability column for the efficient class.}

\item{calibration_model}{A model to calibrate.}

\item{efficiency_thresholds}{A numeric vector of probability levels in (0,1).
Its minimum and maximum values delimit the target interval used to bracket \eqn{\beta}.}

\item{n_expand}{Integer. Increment step size applied to \eqn{\beta} at each iteration.}

\item{vector_gx}{A numeric vector or \code{data.frame} with directional changes for inputs
(typically negative direction), usually built inside \code{PEAXAI_targets}.}

\item{vector_gy}{A numeric vector or \code{data.frame} with directional changes for outputs
(positive direction).}

\item{max_y}{Numeric. Upper-limit multiplier for output expansion relative to observed maxima.}

\item{min_x}{Numeric. Lower-limit multiplier for input contraction relative to observed minima.}
}
\value{
A \code{data.frame} with two numeric columns:
\describe{
\item{\code{min}}{Minimum feasible value of \eqn{\beta} for each observation.}
\item{\code{max}}{Maximum feasible value of \eqn{\beta} for each observation.}
}
}
\description{
Estimates, for each observation, the minimum and maximum feasible values of the
directional distance parameter \eqn{\beta} used in projection-based efficiency
analysis. This function is an internal step of \code{\link{PEAXAI_counterfactuals}},
providing the initial search bounds for the iterative determination of efficiency targets.
}
\details{
For each DMU, the function expands outputs and contracts inputs along the specified
direction until the predicted probability of efficiency (from \code{final_model})
reaches the maximum in \code{efficiency_thresholds} or feasible domain limits.
The resulting interval \eqn{[\beta_{\min}, \beta_{\max}]} is then used by
\code{\link{PEAXAI_counterfactuals}} to refine projections via grid search.
}
\seealso{
\code{\link{PEAXAI_counterfactuals}} (efficiency projections based on \eqn{\beta});
\code{\link[caret]{train}} (model training with class probabilities).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_predict.R
\name{PEAXAI_predict}
\alias{PEAXAI_predict}
\title{Predict Probability of Efficiency Using a Fitted Model}
\usage{
PEAXAI_predict(data, x, y, final_model, calibration_model = NULL)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the variables used
for prediction.}

\item{x}{Integer vector indicating the column indices of input variables in
\code{data}.}

\item{y}{Integer vector indicating the column indices of output variables in
\code{data}.}

\item{final_model}{A fitted \pkg{caret} model provided by PEAXAI_fitting().}

\item{calibration_model}{Optional calibration object returned by \code{PEAXAI_fitting()}.
If \code{NULL}, raw probabilities are returned. If \code{calibration_model$method == "glm.fit"},
Platt scaling is applied by predicting \code{type = "response"} from a logistic regression
using the raw score \code{s}. Otherwise, an isotonic calibration function is retrieved via
\code{calibration_model[[final_model$method]]} and applied to the raw probabilities.}
}
\value{
A numeric vector of predicted probabilities (raw or calibrated), one per
row of \code{data}.
}
\description{
Predicts probabilities for new decision-making units (DMUs) using a fitted
\pkg{caret} classification model. If \code{calibration_model} is provided,
the raw classifier probabilities are post-processed to obtain calibrated
probability estimates (e.g., Platt scaling via logistic regression or
isotonic calibration via a monotone mapping).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-firms.R
\docType{data}
\name{firms}
\alias{firms}
\title{Spanish Food Industry Firms Dataset}
\format{
A \code{data.frame} with 917 rows and 6 columns:
\describe{
\item{total_assets}{Total assets (millions of euros).}
\item{employees}{Number of employees.}
\item{fixed_assets}{Tangible fixed assets (millions of euros).}
\item{personnel_expenses}{Personnel expenses (millions of euros).}
\item{operating_income}{Operating income (millions of euros).}
\item{autonomous_community}{Autonomous community where the firm operates.}
}
}
\source{
SABI (Sistema de Análisis de Balances Ibéricos) database, 2023.
Firms with more than 50 employees in the Spanish food industry.
}
\usage{
data(firms)
}
\description{
Dataset containing information on food industry companies located in Spain,
used to illustrate efficiency analysis within the \pkg{PEAXAI} package.
The dataset reflects the institutional and market heterogeneity that shapes firm-level efficiency
across Spain’s 17 autonomous communities.
}
\details{
The dataset includes 917 food industry firms with more than 50 employees, collected from the
\emph{SABI} database for the year 2023. Each observation corresponds to a single company.
Variables reflect both operational and financial dimensions relevant for productivity and efficiency assessment.

The output variable is:
\itemize{
\item \code{operating_income} — Operating income (in millions of euros), measuring revenues generated
from core business activities.
}

The input variables are:
\itemize{
\item \code{total_assets} — Total assets (millions of euros), representing resources employed.
\item \code{employees} — Number of employees, indicating workforce size (only firms with more than 50 workers are included).
\item \code{fixed_assets} — Tangible fixed assets (millions of euros), such as buildings and machinery.
\item \code{personnel_expenses} — Personnel expenses (millions of euros), including salaries, benefits, and training.
}

The variable \code{autonomous_community} identifies the territorial location of each firm within Spain.

The sample displays substantial dispersion across variables, encompassing both small and large firms.
This heterogeneity affects measures of central tendency—mean and median values differ considerably—thus providing a realistic challenge for efficiency and explainability analyses.
}
\examples{
data(firms)
str(firms)
summary(firms)

if (requireNamespace("ggplot2", quietly = TRUE)) {
  ggplot2::ggplot(firms, ggplot2::aes(x = employees, y = operating_income)) +
    ggplot2::geom_point(alpha = 0.6) +
    ggplot2::labs(
      x = "Number of employees",
      y = "Operating income (millions of euros)",
      title = "Spanish Food Industry Firms (2023)"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      plot.title = ggplot2::element_text(face = "bold"),
      axis.line = ggplot2::element_line(color = "black"),
      axis.ticks = ggplot2::element_line(color = "black"),
      panel.grid.minor = ggplot2::element_blank()
    )
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convex_facets.R
\name{convex_facets}
\alias{convex_facets}
\title{Identify Maximal Facets of the Convex Frontier}
\usage{
convex_facets(data, x, y, RTS = "vrs")
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the variables used in the model.}

\item{x}{Integer vector indicating the column indexes of the input variables in \code{data}.}

\item{y}{Integer vector indicating the column indexes of the output variables in \code{data}.}

\item{RTS}{Text string or number defining the underlying DEA technology /
returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
\describe{
\item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
\item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
\item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
\item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
\item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
\item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
}}
}
\value{
A \code{matrix} where each row represents a maximal facet, and the values are the row indices of the efficient DMUs in \code{data} that form that facet. Returns an empty \code{matrix} if there are no efficient units.
}
\description{
Identifies the maximal subsets of efficient Decision-Making Units (DMUs)
that lie on the same face (facet) of the efficient frontier. It relies on
the concept of "maximal friends" to determine which efficient units can
be grouped together to form convex combinations under the chosen returns-to-scale assumption.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-100-DMUs.R
\docType{data}
\name{data}
\alias{data}
\title{Simulated efficiency dataset (100 DMUs)}
\format{
A \code{data.frame} with 100 rows and 3 columns:
\describe{
\item{x1}{Input of the DMU (e.g., resource use, cost or effort).}
\item{y}{Observed output, potentially affected by technical inefficiency.}
\item{yD}{Deterministic (theoretical) output on the efficient frontier.}
}
}
\source{
Simulated data generated by the authors for illustrative purposes.
}
\usage{
data(data)
}
\description{
Dataset with 100 simulated decision-making units (DMUs) used to illustrate
the basic workflow of \pkg{PEAXAI} in a simple single-input/single-output
setting.
}
\details{
Each DMU uses one input \code{x1} to produce an output \code{y}. The variable
\code{yD} represents the theoretical output on the deterministic frontier,
that is, the output level that would be observed in the absence of
technical inefficiency.

The dataset is purely simulated and is intended for examples and vignettes.
It contains 100 DMUs with heterogeneous input levels and corresponding
output levels. The observed output \code{y} can be interpreted as
\code{y <= yD}, where the gap between \code{yD} and \code{y} reflects
technical inefficiency (plus possible noise, depending on how the data
were generated).
}
\examples{
data(data)
str(data)
summary(data)

if (requireNamespace("ggplot2", quietly = TRUE)) {
  ggplot2::ggplot(data, ggplot2::aes(x = x1)) +
    ggplot2::geom_point(ggplot2::aes(y = y), alpha = 0.6) +
    ggplot2::geom_line(ggplot2::aes(y = yD), color = "red") +
    ggplot2::labs(
      x = "Input x1",
      y = "Output",
      title = "Simulated DMUs and theoretical frontier"
    ) +
    ggplot2::theme_minimal()
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_counterfactuals.R
\name{PEAXAI_counterfactuals}
\alias{PEAXAI_counterfactuals}
\title{Projection-Based Efficiency counterfactuals}
\usage{
PEAXAI_counterfactuals(
  data,
  x,
  y,
  final_model,
  calibration_model = NULL,
  sign = FALSE,
  efficiency_thresholds,
  directional_vector,
  n_expand,
  n_grid,
  max_y = 2,
  min_x = 1
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing input and output variables.}

\item{x}{A numeric vector indicating the column indexes of input variables in \code{data}.}

\item{y}{A numeric vector indicating the column indexes of output variables in \code{data}.}

\item{final_model}{A fitted \pkg{caret} model of class \code{"train"} that supports
\code{predict(type = "prob")} and returns a probability column for the efficient class.}

\item{calibration_model}{Optional probability-calibration model applied to the raw predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
If provided, calibrated probabilities are used for ranking and threshold-based decisions.
Set to \code{NULL} to use uncalibrated predictions.}

\item{sign}{If it is TRUE, only}

\item{efficiency_thresholds}{A numeric vector of probability levels in (0,1)
that define the efficiency classes (e.g., \code{c(0.75, 0.9, 0.95)}).}

\item{directional_vector}{A \code{list} with the required information to
construct the directional vector, including:
\itemize{
\item \code{relative_importance}: Numeric vector of variable importances that sum to 1.
\item \code{baseline}: \code{"mean"}, \code{"median"}, \code{"self"} or \code{"ones"}.
}}

\item{n_expand}{Numeric. Number of expansion steps used to enlarge the initial
search range for \eqn{\beta}.}

\item{n_grid}{Integer. Number of grid points evaluated during each iteration
to refine the cutoff value of \eqn{\beta}.}

\item{max_y}{Numeric. Upper-limit multiplier for output expansion in the search
procedure (default = 2).}

\item{min_x}{Numeric. Lower-limit multiplier for input contraction in the search
procedure (default = 1).}
}
\value{
A named \code{list} with one element per threshold. Each element contains:
\itemize{
\item \code{data}: A \code{data.frame} of projected input–output values at that threshold.
\item \code{beta}: A two-column \code{data.frame} with the optimal \eqn{\beta}
and the corresponding predicted probability.
}
}
\description{
Computes efficiency projections for each observation based on a trained
classifier from \pkg{caret} that provides class probabilities via
\code{predict(type = "prob")}. For each probability threshold, the function
finds the direction and magnitude of change in input–output space required
for a unit to reach a specified efficiency level, following a directional
distance approach.
}
\details{
For each observation and for each probability level in \code{efficiency_thresholds},
the function searches for the smallest directional distance \eqn{\beta} such that
the predicted probability of belonging to the efficient class reaches the target.
}
\examples{
\donttest{
  data("firms", package = "PEAXAI")

  data <- subset(
    firms,
    autonomous_community == "Comunidad Valenciana"
  )

  x <- 1:4
  y <- 5
  RTS <- "vrs"
  imbalance_rate <- NULL

  trControl <- list(
    method = "cv",
    number = 3
  )

  # glm method
  methods <- list(
    "glm" = list(
      weights = "dinamic"
     )
  )

  metric_priority <- c("Balanced_Accuracy", "F1", "ROC_AUC")

  models <- PEAXAI_fitting(
    data = data, x = x, y = y, RTS = RTS,
    imbalance_rate = imbalance_rate,
    methods = methods,
    trControl = trControl,
    metric_priority = metric_priority,
    verbose = FALSE,
    seed = 1
  )

  final_model <- models[["best_model_fit"]][["glm"]]

  importance_method <- list(name = "PI", n.repetitions = 5)

  relative_importance <- PEAXAI_global_importance(
    final_model = final_model,
    x = x,
    y = y,
    explain_data = data,
    reference_data = final_model$trainingData,
    importance_method = importance_method,
    seed = 1
  )

  efficiency_thresholds <- seq(0.75, 0.95, 0.1)

  directional_vector <- list(
    relative_importance = relative_importance,
    baseline  = "mean"
  )

  targets <- PEAXAI_counterfactuals(
    data = data,
    x = x, y = y,
    final_model = final_model,
    efficiency_thresholds = efficiency_thresholds,
    directional_vector = directional_vector,
    n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
  )
}

}
\seealso{
\code{\link{find_beta_maxmin}} for initializing search bounds;
\code{\link[caret]{train}} for model training.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SMOTE_data.R
\name{SMOTE_data}
\alias{SMOTE_data}
\title{Create New SMOTE Units to Balance Data combinations of m + s}
\usage{
SMOTE_data(data, x, y, RTS = "vrs", balance_data, seed)
}
\arguments{
\item{data}{A \code{data.frame} containing the variables used in the model.}

\item{x}{Column indexes of the input variables in the \code{data}.}

\item{y}{Column indexes of the output variables in the \code{data}.}

\item{RTS}{Text string or number defining the underlying DEA technology /
returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
\describe{
\item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
\item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
\item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
\item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
\item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
\item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
}}

\item{balance_data}{Indicate level of efficient units to achive and the number of efficient and not efficient units.}

\item{seed}{Integer. Seed for reproducibility.}
}
\value{
It returns a \code{data.frame} with the newly created set of DMUs incorporated.
}
\description{
This function creates new DMUs to address data imbalances. If the majority class is efficient, it generates new inefficient DMUs by worsering the observed units. Conversely, if the majority class is inefficient, it projects inefficient DMUs to the frontier. Finally, a random selection if performed to keep a proportion of 0.65 for the majority class and 0.35 for the minority class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_peer.R
\name{PEAXAI_peer}
\alias{PEAXAI_peer}
\title{Identify Benchmark Peers Based on Estimated Efficiency Probabilities}
\usage{
PEAXAI_peer(
  data,
  x,
  y,
  final_model,
  calibration_model = NULL,
  efficiency_thresholds,
  targets,
  weighted = FALSE,
  relative_importance = NULL
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing input and output variables used in the efficiency model.}

\item{x}{Integer vector indicating the column indices of input variables in \code{data}.}

\item{y}{Integer vector indicating the column indices of output variables in \code{data}.}

\item{final_model}{A fitted classification model used to estimate efficiency probabilities. Supported classes: \code{"train"} (from \pkg{caret}) or \code{"glm"} (binomial).}

\item{calibration_model}{Optional probability-calibration model applied to the raw
predicted probabilities from \code{final_model} (e.g., Platt scaling or isotonic regression).
If provided, calibrated probabilities are used for ranking and threshold-based decisions.
Set to \code{NULL} to use uncalibrated predictions.}

\item{efficiency_thresholds}{Numeric vector indicating the minimum probability values required to consider a DMU as efficient.}

\item{targets}{A named list containing, for each efficiency threshold, the corresponding}

\item{weighted}{Logical. If \code{TRUE}, peers are selected using weighted Euclidean distances based on variable importance. If \code{FALSE} (default), unweighted distances are used.}

\item{relative_importance}{Optional named numeric vector indicating the relative importance of each input/output variable (used when \code{weighted = TRUE}).}
}
\value{
A named list of matrices. Each element corresponds to an efficiency threshold and contains, for each DMU, the index of the closest efficient peer.
If \code{weighted = FALSE}, the list contains unweighted peers. If \code{weighted = TRUE}, the list contains weighted peers.
}
\description{
Identifies peer units (i.e., reference benchmarks) for each decision-making unit (DMU) based on predicted probabilities of technical efficiency.
Given a fitted classification model that estimates the probability of being efficient, the function selects, for each DMU, its nearest efficient peer
according to Euclidean or weighted distances. Multiple efficiency thresholds can be specified to assess different levels of benchmarking stringency.
}
\details{
This function enables probabilistic peer identification under uncertainty, supporting flexible definitions of efficiency based on thresholds over estimated probabilities.
When \code{weighted = TRUE}, variable weights (e.g., derived from feature importance) modulate the peer selection process, allowing for context-aware benchmarking.
}
\examples{
\donttest{
  data("firms", package = "PEAXAI")

  data <- subset(
    firms,
    autonomous_community == "Comunidad Valenciana"
  )

  x <- 1:4
  y <- 5
  RTS <- "vrs"
  imbalance_rate <- NULL

  trControl <- list(
    method = "cv",
    number = 3
  )

  # glm method
  methods <- list(
    "glm" = list(
      weights = "dinamic"
     )
   )

  metric_priority <- c("Balanced_Accuracy", "ROC_AUC")

  models <- PEAXAI_fitting(
    data = data, x = x, y = y, RTS = RTS,
    imbalance_rate = imbalance_rate,
    methods = methods,
    trControl = trControl,
    metric_priority = metric_priority,
    verbose = FALSE,
    seed = 1
  )

  final_model <- models[["best_model_fit"]][["glm"]]

  importance_method <- list(name = "PI", n.repetitions = 5)

  relative_importance <- PEAXAI_global_importance(
    final_model = final_model,
    x = x,
    y = y,
    explain_data = data,
    reference_data = final_model$trainingData,
    importance_method = importance_method,
    seed = 1
  )

  efficiency_thresholds <- seq(0.75, 0.95, 0.1)

  directional_vector <- list(
   relative_importance = relative_importance,
   baseline  = "mean"
  )

  targets <- PEAXAI_counterfactuals(
    data = data,
    x = x, y = y,
    final_model = final_model,
    efficiency_thresholds = efficiency_thresholds,
    directional_vector = directional_vector,
    n_expand = 0.5, n_grid = 50, max_y = 2, min_x = 1
  )

  peers <- PEAXAI_peer(data = data, x = x, y = y, final_model = final_model,
  efficiency_thresholds = efficiency_thresholds, targets = targets, weighted = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_importance_features.R
\name{PEAXAI_global_importance}
\alias{PEAXAI_global_importance}
\title{Global feature importance for efficiency classifiers}
\usage{
PEAXAI_global_importance(
  final_model,
  x,
  y,
  explain_data,
  reference_data,
  importance_method,
  seed = 314
)
}
\arguments{
\item{final_model}{A fitted model. If it is a base-\code{glm} binomial, probabilities
are obtained with \code{type = "response"}; otherwise the function expects
\code{predict(type = "prob")} with a column named \code{"efficient"}.}

\item{x}{Integer or character vector with the columns used as \strong{inputs} (predictors).}

\item{y}{Integer or character vector with the columns used as \strong{outputs} (targets used
to define \code{class_efficiency} in training; not included in \code{X} when explaining).}

\item{explain_data}{A \code{data.frame} with the original observed DMUs passed
by the user to \code{PEAXAI_fitting()}. This dataset represents the real,
non-augmented observations on which explanations can be computed.}

\item{reference_data}{A \code{data.frame} with the training dataset used to fit
the optimal model returned by \code{PEAXAI_fitting()}. Depending on the class-
balancing procedure, this dataset may contain only real observed DMUs or both
real and synthetic DMUs.}

\item{importance_method}{A named list (or data.frame-like) with the backend and its args:
\describe{
\item{\code{name}}{One of \code{"SA"}, \code{"SHAP"}, \code{"PI"}.}
\item{\code{method}}{(SA) One of \code{"1D-SA"}, \code{"sens"}, \code{"DSA"}, \code{"MSA"}, \code{"CSA"}, \code{"GSA"}.}
\item{\code{measures}}{(SA) e.g. \code{"AAD"}, \code{"gradient"}, \code{"variance"}, \code{"range"}.}
\item{\code{levels}}{(SA) Discretization levels used by \code{rminer::Importance}.}
\item{\code{baseline}}{(SA) Baseline value for SA, if applicable.}
\item{\code{bg_n}}{(SHAP) Background sample size for \code{kernelshap::kernelshap} (default: 200).}
\item{\code{n.repetitions}}{(PI) Number of permutations per feature for \code{iml::FeatureImp}.}
}}

\item{seed}{Integer. Seed for reproducibility.}
}
\value{
A named numeric vector (or 1-row data.frame) of normalized importances,
with names matching the predictor columns; the values sum to 1.
}
\description{
Computes \strong{global feature importance} for a fitted classification model that
separates Pareto-efficient DMUs, using one of three XAI backends:
\itemize{
\item \code{"SA"} — Sensitivity Analysis via \pkg{rminer}.
\item \code{"SHAP"} — Model-agnostic SHAP approximations via \pkg{kernelshap}.
\item \code{"PI"} — Permutation Importance via \pkg{iml}.
}
You can evaluate the model on either the training domain (\code{background = "train"})
or the real-world domain (\code{background = "real"}) and compute importance on a
chosen \code{target} set (\code{"train"} or \code{"real"}). Importances are
returned normalized to sum to 1.
}
\details{
Internally, the function builds background/target sets with \code{xai_prepare_sets()}.
For \code{glm} models, the positive class is assumed to be the \strong{second level}
(\code{"efficient"}) and probabilities are extracted with \code{type = "response"}.
For other models (e.g., \pkg{caret}), \code{predict(type = "prob")[, "efficient"]} is used.
}
\examples{
\donttest{
  data("firms", package = "PEAXAI")

  data <- subset(
    firms,
    autonomous_community == "Comunidad Valenciana"
  )

  x <- 1:4
  y <- 5
  RTS <- "vrs"
  imbalance_rate <- NULL

  trControl <- list(
    method = "cv",
    number = 3
  )

  # glm method
  methods <- list(
    "glm" = list(
      weights = "dinamic"
     )
   )

  metric_priority <- c("Balanced_Accuracy", "ROC_AUC")

  models <- PEAXAI_fitting(
    data = data, x = x, y = y, RTS = RTS,
    imbalance_rate = imbalance_rate,
    methods = methods,
    trControl = trControl,
    metric_priority = metric_priority,
    seed = 1,
    verbose = FALSE
  )

  final_model <- models[["best_model_fit"]][["glm"]]

  importance_method <- list(name = "PI", n.repetitions = 5)

  relative_importance <- PEAXAI_global_importance(
    final_model = final_model,
    x = x,
    y = y,
    explain_data = data,
    reference_data = final_model$trainingData,
    importance_method = importance_method,
    seed = 1
  )

  imp <- PEAXAI_global_importance(
    final_model = final_model,
    x = x,
    y = y,
    explain_data = data,
    reference_data = final_model$trainingData,
    importance_method = importance_method,
    seed = 1
  )

  head(imp)
}

}
\seealso{
\code{\link[kernelshap]{kernelshap}}, \code{\link[iml]{FeatureImp}},
\code{\link[rminer]{Importance}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PEAXAI_fitting.R
\name{PEAXAI_fitting}
\alias{PEAXAI_fitting}
\title{Training Classification Models to Estimate Efficiency}
\usage{
PEAXAI_fitting(
  data,
  x,
  y,
  z_numeric = NULL,
  z_factor = NULL,
  RTS = "vrs",
  B = NULL,
  alpha = FALSE,
  m = NULL,
  imbalance_rate = NULL,
  trControl,
  methods,
  metric_priority = "Balanced_Accuracy",
  calibration_method = NULL,
  hold_out = NULL,
  seed = 314,
  verbose = TRUE
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the variables in the model.}

\item{x}{Integer vector with column indices of input variables in \code{data}.}

\item{y}{Integer vector with column indices of output variables in \code{data}.}

\item{z_numeric}{Integer vector with column indices of numeric environment variables in \code{data}. By default is \code{NULL}.}

\item{z_factor}{Integer vector with column indices of factor environment variables in \code{data}. By default is \code{NULL}.}

\item{RTS}{Text string or number defining the underlying DEA technology /
returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
\describe{
\item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
\item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
\item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
\item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
\item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
\item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
}}

\item{B}{number of bootstrap replicates in Conditional DEA.}

\item{alpha}{This allow to choose the size of the Confidence Intervals computed. By defaulta alpha = FALSE. In this case no confidence interval are computed.}

\item{m}{number of units to be included in the reference set.}

\item{imbalance_rate}{Optional target(s) for class balance via SMOTE. If \code{NULL},
no synthetic balancing is performed.}

\item{trControl}{A \code{caret::trainControl}-like list that specifies the resampling
strategy; recognized values for \code{$method} include \code{"cv"}, \code{"test_set"},
and \code{"none"}. See \pkg{caret} documentation.}

\item{methods}{A \code{list} of selected machine learning models and their hyperparameters.}

\item{metric_priority}{A \code{string} specifying the summary metric for classification to select the optimal model. Default includes \code{"Balanced_Accuracy"} due to (normally) unbalanced data.}

\item{calibration_method}{Character string specifying the probability calibration
method applied to the fitted classifier. Supported options are:
\describe{
\item{\code{"none"}}{No probability calibration is applied; raw classifier
probabilities are returned.}
\item{\code{"platt"}}{Platt scaling, which fits a logistic regression model
mapping raw scores to calibrated probabilities. Suitable for relatively
small samples and parametric probability adjustment.}
\item{\code{"isotonic"}}{Isotonic regression, a non-parametric monotonic
calibration method that adapts flexibly to the empirical score-probability
relationship. Recommended for larger samples where sufficient calibration
data are available.}
}
Calibration is performed using validation data whenever resampling is enabled.}

\item{hold_out}{Numeric proportion in (0,1) for validation split (default \code{NULL}).
If \code{NULL}, training and validation use the same data.}

\item{seed}{Integer. Seed for reproducibility.}

\item{verbose}{Logical; if \code{TRUE}, prints progress messages (default \code{FALSE}).}
}
\value{
A \code{"PEAXAI"} (list) with the best technique, best fitted models and their performance and the results by fold.
}
\description{
Trains one or multiple classification algorithms to identify Pareto-efficient
decision-making units (DMUs). It jointly searches model hyperparameters and the
class-balancing level (synthetic samples via SMOTE) using k-fold cross-
validation or a train/validation/test split, selecting the configuration that
maximizes the specified metric(s). Returns, for each technique, the best fitted
model together with training summaries, performance metrics, and the selected
balancing level.
}
\examples{
\donttest{
  data("firms", package = "PEAXAI")

  data <- subset(
    firms,
    autonomous_community == "Comunidad Valenciana"
  )

  trControl <- list(
    method = "cv",
    number = 3
  )

  # glm method
  methods <- list(
    "glm" = list(
        weights = "dinamic"
     )
  )

  models <- PEAXAI_fitting(
    data = data,
    x = c(1:4),
    y = 5,
    RTS = "vrs",
    imbalance_rate = NULL,
    methods = methods,
    trControl = trControl,
    metric_priority = c("Balanced_Accuracy", "F1", "ROC_AUC"),
    seed = 1,
    verbose = FALSE
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SMOTE_Z_data.R
\name{SMOTE_Z_data}
\alias{SMOTE_Z_data}
\title{Create New SMOTE Units to Balance Data combinations of m + s}
\usage{
SMOTE_Z_data(
  data,
  x,
  y,
  z_numeric,
  z_factor,
  balance_data,
  RTS = "vrs",
  B,
  m,
  alpha,
  bandwidth = NULL,
  seed
)
}
\arguments{
\item{data}{A \code{data.frame} containing the variables used in the model.}

\item{x}{Column indexes of the input variables in the \code{data}.}

\item{y}{Column indexes of the output variables in the \code{data}.}

\item{z_numeric}{Integer vector with column indices of numeric environment variables in \code{data}. By default is \code{NULL}.}

\item{z_factor}{Integer vector with column indices of factor environment variables in \code{data}. By default is \code{NULL}.}

\item{balance_data}{Indicate level of efficient units to achive and the number of efficient and not efficient units.}

\item{RTS}{Text string or number defining the underlying DEA technology /
returns-to-scale assumption (default: \code{"vrs"}). Accepted values:
\describe{
\item{\code{0} / \code{"fdh"}}{Free disposability hull, no convexity assumption.}
\item{\code{1} / \code{"vrs"}}{Variable returns to scale, convexity and free disposability.}
\item{\code{2} / \code{"drs"}}{Decreasing returns to scale, convexity, down-scaling and free disposability.}
\item{\code{3} / \code{"crs"}}{Constant returns to scale, convexity and free disposability.}
\item{\code{4} / \code{"irs"}}{Increasing returns to scale (up-scaling, not down-scaling), convexity and free disposability.}
\item{\code{5} / \code{"add"}}{Additivity (scaling up and down, but only with integers), and free disposability.}
}}

\item{B}{number of bootstrap replicates in Conditional DEA.}

\item{m}{number of units to be included in the reference set.}

\item{alpha}{This allow to choose the size of the Confidence Intervals computed. By defaulta alpha = FALSE. In this case no confidence interval are computed.}

\item{bandwidth}{the bandwidth parameters for the unconditional kernel density estimator used in the conditional DEA framework. It is typically obtained using \code{\link[np]{npudensbw}} and supports mixed data types, including continuous variables and discrete unordered or ordered factors. Bandwidths can be selected using normal reference rules, likelihood cross-validation, or least-squares cross-validation following Li and Racine (2003). If \code{NULL}, the bandwidth is estimated internally.}

\item{seed}{Integer. Seed for reproducibility.}
}
\value{
It returns a \code{data.frame} with the newly created set of DMUs incorporated.
}
\description{
This function creates new DMUs to address data imbalances. If the majority class is efficient, it generates new inefficient DMUs by worsering the observed units. Conversely, if the majority class is inefficient, it projects inefficient DMUs to the frontier. Finally, a random selection if performed to keep a proportion of 0.65 for the majority class and 0.35 for the minority class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/training.R
\name{train_PEAXAI}
\alias{train_PEAXAI}
\title{Training a Classification Machine Learning Model}
\usage{
train_PEAXAI(data, method, parameters, trControl, metric_priority, seed)
}
\arguments{
\item{data}{A \code{data.frame} or \code{matrix} containing the variables in the model.}

\item{method}{Parameters for controlling the training process (from the \code{'caret'} package).}

\item{parameters}{A \code{list} of selected machine learning models and their hyperparameters.}

\item{trControl}{A \code{list} of selected machine learning learning.}

\item{metric_priority}{A \code{string} specifying the summary metric for classification to select the optimal model. Default includes \code{"Balanced_Accuracy"} due to (normally) unbalanced data.}

\item{seed}{Integer. Seed for reproducibility.}
}
\value{
It returns a \code{list} with the chosen model.
}
\description{
This function trains a set of models and selects best hyperparameters for each of them.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xai_prepare_sets.R
\name{xai_prepare_sets}
\alias{xai_prepare_sets}
\title{Prepare Training and Target Datasets from a caret Model}
\usage{
xai_prepare_sets(
  data,
  x,
  y,
  final_model,
  background,
  target,
  type,
  threshold,
  levels_order
)
}
\arguments{
\item{data}{A \code{data.frame} containing the original dataset used to train the model. Only needed when using \code{"real"} as background or target.}

\item{x}{Not currently used. Reserved for future input variable selection.}

\item{y}{Not currently used. Reserved for future output variable specification.}

\item{final_model}{A trained model object of class \code{"train"} from the \pkg{caret} package.}

\item{background}{A character string, either \code{"train"} or \code{"real"}, specifying the background dataset used for explainability.}

\item{target}{A character string, either \code{"train"} or \code{"real"}, specifying the target dataset to be explained.}

\item{type}{Not currently used. Reserved for future prediction types.}

\item{threshold}{Not currently used. Reserved for future thresholding logic.}

\item{levels_order}{A character vector specifying the levels of the response factor, typically \code{c("not_efficient", "efficient")}. Not currently used, but can help in reordering or relabeling.}
}
\value{
A \code{list} with two elements:
\describe{
\item{\code{train_data}}{A \code{data.frame} representing the background dataset, with the class column renamed to \code{"class_efficiency"} and positioned last.}
\item{\code{target_data}}{A \code{data.frame} representing the target dataset, formatted in the same way.}
}
}
\description{
Extracts and formats the training and/or target datasets from a machine learning model trained with \code{caret::train},
allowing for distinction between using the full training data or only the original subset used for modeling.
It standardizes the class column to be named \code{"class_efficiency"} and positions it as the last column.
}
